'''
licenses.py: license descriptions.

 ╭───────────────────── Notice ── Notice ── Notice ────────────────────╮
 |     THIS FILE IS AUTOMATICALLY GENERATED. DO NOT EDIT THIS FILE.    |
 |  For more info, see the script ../dev/scripts/create-licenses-file  |
 ╰───────────────────── Notice ── Notice ── Notice ────────────────────╯
'''

from collections import namedtuple
from commonpy.data_structures import CaseFoldDict


# The following named tuple is used in one of the hash tables below. Note that
# callers do not need to import this named tuple definition; they can access
# access the fields directly on the objects retrieved from the table.

License = namedtuple('License', 'title description url')


# The following hash table has all SPDX license identifiers as keys, and our
# simple License named tuple as values. The URLs are the main URLs that SDPX
# gives as the reference URLs. The descriptions come from a different source,
# choosealicense.com. That source does not have a description for all SPDX
# licenses, which is why not all the entries below have a description.

LICENSES = CaseFoldDict({
  '0BSD': License(title='''BSD Zero Clause License''', description='''The BSD Zero Clause license goes further than the BSD 2-Clause license to allow you unlimited freedom with the software without requirements to include the copyright notice, license text, or disclaimer in either source or binary forms.''', url='https://spdx.org/licenses/0BSD'),
  'AAL': License(title='''Attribution Assurance License''', description='''''', url='https://spdx.org/licenses/AAL'),
  'Abstyles': License(title='''Abstyles License''', description='''''', url='https://spdx.org/licenses/Abstyles'),
  'AdaCore-doc': License(title='''AdaCore Doc License''', description='''''', url='https://spdx.org/licenses/AdaCore-doc'),
  'Adobe-2006': License(title='''Adobe Systems Incorporated Source Code License Agreement''', description='''''', url='https://spdx.org/licenses/Adobe-2006'),
  'Adobe-Glyph': License(title='''Adobe Glyph List License''', description='''''', url='https://spdx.org/licenses/Adobe-Glyph'),
  'ADSL': License(title='''Amazon Digital Services License''', description='''''', url='https://spdx.org/licenses/ADSL'),
  'AFL-1.1': License(title='''Academic Free License v1.1''', description='''''', url='https://spdx.org/licenses/AFL-1.1'),
  'AFL-1.2': License(title='''Academic Free License v1.2''', description='''''', url='https://spdx.org/licenses/AFL-1.2'),
  'AFL-2.0': License(title='''Academic Free License v2.0''', description='''''', url='https://spdx.org/licenses/AFL-2.0'),
  'AFL-2.1': License(title='''Academic Free License v2.1''', description='''''', url='https://spdx.org/licenses/AFL-2.1'),
  'AFL-3.0': License(title='''Academic Free License v3.0''', description='''The Academic Free License is a variant of the Open Software License that does not require that the source code of derivative works be disclosed. It contains explicit copyright and patent grants and reserves trademark rights in the author.''', url='https://spdx.org/licenses/AFL-3.0'),
  'Afmparse': License(title='''Afmparse License''', description='''''', url='https://spdx.org/licenses/Afmparse'),
  'AGPL-1.0': License(title='''Affero General Public License v1.0''', description='''''', url='https://spdx.org/licenses/AGPL-1.0'),
  'AGPL-1.0-only': License(title='''Affero General Public License v1.0 only''', description='''''', url='https://spdx.org/licenses/AGPL-1.0-only'),
  'AGPL-1.0-or-later': License(title='''Affero General Public License v1.0 or later''', description='''''', url='https://spdx.org/licenses/AGPL-1.0-or-later'),
  'AGPL-3.0': License(title='''GNU Affero General Public License v3.0''', description='''Permissions of this strongest copyleft license are conditioned on making available complete source code of licensed works and modifications, which include larger works using a licensed work, under the same license. Copyright and license notices must be preserved. Contributors provide an express grant of patent rights. When a modified version is used to provide a service over a network, the complete source code of the modified version must be made available.''', url='https://spdx.org/licenses/AGPL-3.0'),
  'AGPL-3.0-only': License(title='''GNU Affero General Public License v3.0 only''', description='''''', url='https://spdx.org/licenses/AGPL-3.0-only'),
  'AGPL-3.0-or-later': License(title='''GNU Affero General Public License v3.0 or later''', description='''''', url='https://spdx.org/licenses/AGPL-3.0-or-later'),
  'Aladdin': License(title='''Aladdin Free Public License''', description='''''', url='https://spdx.org/licenses/Aladdin'),
  'AMDPLPA': License(title='''AMD's plpa_map.c License''', description='''''', url='https://spdx.org/licenses/AMDPLPA'),
  'AML': License(title='''Apple MIT License''', description='''''', url='https://spdx.org/licenses/AML'),
  'AMPAS': License(title='''Academy of Motion Picture Arts and Sciences BSD''', description='''''', url='https://spdx.org/licenses/AMPAS'),
  'ANTLR-PD': License(title='''ANTLR Software Rights Notice''', description='''''', url='https://spdx.org/licenses/ANTLR-PD'),
  'ANTLR-PD-fallback': License(title='''ANTLR Software Rights Notice with license fallback''', description='''''', url='https://spdx.org/licenses/ANTLR-PD-fallback'),
  'Apache-1.0': License(title='''Apache License 1.0''', description='''''', url='https://spdx.org/licenses/Apache-1.0'),
  'Apache-1.1': License(title='''Apache License 1.1''', description='''''', url='https://spdx.org/licenses/Apache-1.1'),
  'Apache-2.0': License(title='''Apache License 2.0''', description='''A permissive license whose main conditions require preservation of copyright and license notices. Contributors provide an express grant of patent rights. Licensed works, modifications, and larger works may be distributed under different terms and without source code.''', url='https://spdx.org/licenses/Apache-2.0'),
  'APAFML': License(title='''Adobe Postscript AFM License''', description='''''', url='https://spdx.org/licenses/APAFML'),
  'APL-1.0': License(title='''Adaptive Public License 1.0''', description='''''', url='https://spdx.org/licenses/APL-1.0'),
  'App-s2p': License(title='''App::s2p License''', description='''''', url='https://spdx.org/licenses/App-s2p'),
  'APSL-1.0': License(title='''Apple Public Source License 1.0''', description='''''', url='https://spdx.org/licenses/APSL-1.0'),
  'APSL-1.1': License(title='''Apple Public Source License 1.1''', description='''''', url='https://spdx.org/licenses/APSL-1.1'),
  'APSL-1.2': License(title='''Apple Public Source License 1.2''', description='''''', url='https://spdx.org/licenses/APSL-1.2'),
  'APSL-2.0': License(title='''Apple Public Source License 2.0''', description='''''', url='https://spdx.org/licenses/APSL-2.0'),
  'Arphic-1999': License(title='''Arphic Public License''', description='''''', url='https://spdx.org/licenses/Arphic-1999'),
  'Artistic-1.0': License(title='''Artistic License 1.0''', description='''''', url='https://spdx.org/licenses/Artistic-1.0'),
  'Artistic-1.0-cl8': License(title='''Artistic License 1.0 w/clause 8''', description='''''', url='https://spdx.org/licenses/Artistic-1.0-cl8'),
  'Artistic-1.0-Perl': License(title='''Artistic License 1.0 (Perl)''', description='''''', url='https://spdx.org/licenses/Artistic-1.0-Perl'),
  'Artistic-2.0': License(title='''Artistic License 2.0''', description='''Heavily favored by the Perl community, the Artistic license requires that modified versions of the software do not prevent users from running the standard version.''', url='https://spdx.org/licenses/Artistic-2.0'),
  'Baekmuk': License(title='''Baekmuk License''', description='''''', url='https://spdx.org/licenses/Baekmuk'),
  'Bahyph': License(title='''Bahyph License''', description='''''', url='https://spdx.org/licenses/Bahyph'),
  'Barr': License(title='''Barr License''', description='''''', url='https://spdx.org/licenses/Barr'),
  'Beerware': License(title='''Beerware License''', description='''''', url='https://spdx.org/licenses/Beerware'),
  'Bitstream-Charter': License(title='''Bitstream Charter Font License''', description='''''', url='https://spdx.org/licenses/Bitstream-Charter'),
  'Bitstream-Vera': License(title='''Bitstream Vera Font License''', description='''''', url='https://spdx.org/licenses/Bitstream-Vera'),
  'BitTorrent-1.0': License(title='''BitTorrent Open Source License v1.0''', description='''''', url='https://spdx.org/licenses/BitTorrent-1.0'),
  'BitTorrent-1.1': License(title='''BitTorrent Open Source License v1.1''', description='''''', url='https://spdx.org/licenses/BitTorrent-1.1'),
  'blessing': License(title='''SQLite Blessing''', description='''''', url='https://spdx.org/licenses/blessing'),
  'BlueOak-1.0.0': License(title='''Blue Oak Model License 1.0.0''', description='''''', url='https://spdx.org/licenses/BlueOak-1.0.0'),
  'Borceux': License(title='''Borceux license''', description='''''', url='https://spdx.org/licenses/Borceux'),
  'Brian-Gladman-3-Clause': License(title='''Brian Gladman 3-Clause License''', description='''''', url='https://spdx.org/licenses/Brian-Gladman-3-Clause'),
  'BSD-1-Clause': License(title='''BSD 1-Clause License''', description='''''', url='https://spdx.org/licenses/BSD-1-Clause'),
  'BSD-2-Clause': License(title='''BSD 2-Clause "Simplified" License''', description='''A permissive license that comes in two variants, the BSD 2-Clause and BSD 3-Clause. Both have very minute differences to the MIT license.''', url='https://spdx.org/licenses/BSD-2-Clause'),
  'BSD-2-Clause-FreeBSD': License(title='''BSD 2-Clause FreeBSD License''', description='''''', url='https://spdx.org/licenses/BSD-2-Clause-FreeBSD'),
  'BSD-2-Clause-NetBSD': License(title='''BSD 2-Clause NetBSD License''', description='''''', url='https://spdx.org/licenses/BSD-2-Clause-NetBSD'),
  'BSD-2-Clause-Patent': License(title='''BSD-2-Clause Plus Patent License''', description='''''', url='https://spdx.org/licenses/BSD-2-Clause-Patent'),
  'BSD-2-Clause-Views': License(title='''BSD 2-Clause with views sentence''', description='''''', url='https://spdx.org/licenses/BSD-2-Clause-Views'),
  'BSD-3-Clause': License(title='''BSD 3-Clause "New" or "Revised" License''', description='''A permissive license similar to the BSD 2-Clause License, but with a 3rd clause that prohibits others from using the name of the copyright holder or its contributors to promote derived products without written consent.''', url='https://spdx.org/licenses/BSD-3-Clause'),
  'BSD-3-Clause-Attribution': License(title='''BSD with attribution''', description='''''', url='https://spdx.org/licenses/BSD-3-Clause-Attribution'),
  'BSD-3-Clause-Clear': License(title='''BSD 3-Clause Clear License''', description='''A variant of the BSD 3-Clause License that explicitly does not grant any patent rights.''', url='https://spdx.org/licenses/BSD-3-Clause-Clear'),
  'BSD-3-Clause-LBNL': License(title='''Lawrence Berkeley National Labs BSD variant license''', description='''''', url='https://spdx.org/licenses/BSD-3-Clause-LBNL'),
  'BSD-3-Clause-Modification': License(title='''BSD 3-Clause Modification''', description='''''', url='https://spdx.org/licenses/BSD-3-Clause-Modification'),
  'BSD-3-Clause-No-Military-License': License(title='''BSD 3-Clause No Military License''', description='''''', url='https://spdx.org/licenses/BSD-3-Clause-No-Military-License'),
  'BSD-3-Clause-No-Nuclear-License': License(title='''BSD 3-Clause No Nuclear License''', description='''''', url='https://spdx.org/licenses/BSD-3-Clause-No-Nuclear-License'),
  'BSD-3-Clause-No-Nuclear-License-2014': License(title='''BSD 3-Clause No Nuclear License 2014''', description='''''', url='https://spdx.org/licenses/BSD-3-Clause-No-Nuclear-License-2014'),
  'BSD-3-Clause-No-Nuclear-Warranty': License(title='''BSD 3-Clause No Nuclear Warranty''', description='''''', url='https://spdx.org/licenses/BSD-3-Clause-No-Nuclear-Warranty'),
  'BSD-3-Clause-Open-MPI': License(title='''BSD 3-Clause Open MPI variant''', description='''''', url='https://spdx.org/licenses/BSD-3-Clause-Open-MPI'),
  'BSD-4-Clause': License(title='''BSD 4-Clause "Original" or "Old" License''', description='''A permissive license similar to the BSD 3-Clause License, but with an "advertising clause" that requires an acknowledgment of the original source in all advertising material.''', url='https://spdx.org/licenses/BSD-4-Clause'),
  'BSD-4-Clause-Shortened': License(title='''BSD 4 Clause Shortened''', description='''''', url='https://spdx.org/licenses/BSD-4-Clause-Shortened'),
  'BSD-4-Clause-UC': License(title='''BSD-4-Clause (University of California-Specific)''', description='''''', url='https://spdx.org/licenses/BSD-4-Clause-UC'),
  'BSD-4.3RENO': License(title='''BSD 4.3 RENO License''', description='''''', url='https://spdx.org/licenses/BSD-4.3RENO'),
  'BSD-4.3TAHOE': License(title='''BSD 4.3 TAHOE License''', description='''''', url='https://spdx.org/licenses/BSD-4.3TAHOE'),
  'BSD-Advertising-Acknowledgement': License(title='''BSD Advertising Acknowledgement License''', description='''''', url='https://spdx.org/licenses/BSD-Advertising-Acknowledgement'),
  'BSD-Attribution-HPND-disclaimer': License(title='''BSD with Attribution and HPND disclaimer''', description='''''', url='https://spdx.org/licenses/BSD-Attribution-HPND-disclaimer'),
  'BSD-Protection': License(title='''BSD Protection License''', description='''''', url='https://spdx.org/licenses/BSD-Protection'),
  'BSD-Source-Code': License(title='''BSD Source Code Attribution''', description='''''', url='https://spdx.org/licenses/BSD-Source-Code'),
  'BSL-1.0': License(title='''Boost Software License 1.0''', description='''A simple permissive license only requiring preservation of copyright and license notices for source (and not binary) distribution. Licensed works, modifications, and larger works may be distributed under different terms and without source code.''', url='https://spdx.org/licenses/BSL-1.0'),
  'BUSL-1.1': License(title='''Business Source License 1.1''', description='''''', url='https://spdx.org/licenses/BUSL-1.1'),
  'bzip2-1.0.5': License(title='''bzip2 and libbzip2 License v1.0.5''', description='''''', url='https://spdx.org/licenses/bzip2-1.0.5'),
  'bzip2-1.0.6': License(title='''bzip2 and libbzip2 License v1.0.6''', description='''''', url='https://spdx.org/licenses/bzip2-1.0.6'),
  'C-UDA-1.0': License(title='''Computational Use of Data Agreement v1.0''', description='''''', url='https://spdx.org/licenses/C-UDA-1.0'),
  'CAL-1.0': License(title='''Cryptographic Autonomy License 1.0''', description='''''', url='https://spdx.org/licenses/CAL-1.0'),
  'CAL-1.0-Combined-Work-Exception': License(title='''Cryptographic Autonomy License 1.0 (Combined Work Exception)''', description='''''', url='https://spdx.org/licenses/CAL-1.0-Combined-Work-Exception'),
  'Caldera': License(title='''Caldera License''', description='''''', url='https://spdx.org/licenses/Caldera'),
  'CATOSL-1.1': License(title='''Computer Associates Trusted Open Source License 1.1''', description='''''', url='https://spdx.org/licenses/CATOSL-1.1'),
  'CC-BY-1.0': License(title='''Creative Commons Attribution 1.0 Generic''', description='''''', url='https://spdx.org/licenses/CC-BY-1.0'),
  'CC-BY-2.0': License(title='''Creative Commons Attribution 2.0 Generic''', description='''''', url='https://spdx.org/licenses/CC-BY-2.0'),
  'CC-BY-2.5': License(title='''Creative Commons Attribution 2.5 Generic''', description='''''', url='https://spdx.org/licenses/CC-BY-2.5'),
  'CC-BY-2.5-AU': License(title='''Creative Commons Attribution 2.5 Australia''', description='''''', url='https://spdx.org/licenses/CC-BY-2.5-AU'),
  'CC-BY-3.0': License(title='''Creative Commons Attribution 3.0 Unported''', description='''''', url='https://spdx.org/licenses/CC-BY-3.0'),
  'CC-BY-3.0-AT': License(title='''Creative Commons Attribution 3.0 Austria''', description='''''', url='https://spdx.org/licenses/CC-BY-3.0-AT'),
  'CC-BY-3.0-DE': License(title='''Creative Commons Attribution 3.0 Germany''', description='''''', url='https://spdx.org/licenses/CC-BY-3.0-DE'),
  'CC-BY-3.0-IGO': License(title='''Creative Commons Attribution 3.0 IGO''', description='''''', url='https://spdx.org/licenses/CC-BY-3.0-IGO'),
  'CC-BY-3.0-NL': License(title='''Creative Commons Attribution 3.0 Netherlands''', description='''''', url='https://spdx.org/licenses/CC-BY-3.0-NL'),
  'CC-BY-3.0-US': License(title='''Creative Commons Attribution 3.0 United States''', description='''''', url='https://spdx.org/licenses/CC-BY-3.0-US'),
  'CC-BY-4.0': License(title='''Creative Commons Attribution 4.0 International''', description='''Permits almost any use subject to providing credit and license notice. Frequently used for media assets and educational materials. The most common license for Open Access scientific publications. Not recommended for software.''', url='https://spdx.org/licenses/CC-BY-4.0'),
  'CC-BY-NC-1.0': License(title='''Creative Commons Attribution Non Commercial 1.0 Generic''', description='''''', url='https://spdx.org/licenses/CC-BY-NC-1.0'),
  'CC-BY-NC-2.0': License(title='''Creative Commons Attribution Non Commercial 2.0 Generic''', description='''''', url='https://spdx.org/licenses/CC-BY-NC-2.0'),
  'CC-BY-NC-2.5': License(title='''Creative Commons Attribution Non Commercial 2.5 Generic''', description='''''', url='https://spdx.org/licenses/CC-BY-NC-2.5'),
  'CC-BY-NC-3.0': License(title='''Creative Commons Attribution Non Commercial 3.0 Unported''', description='''''', url='https://spdx.org/licenses/CC-BY-NC-3.0'),
  'CC-BY-NC-3.0-DE': License(title='''Creative Commons Attribution Non Commercial 3.0 Germany''', description='''''', url='https://spdx.org/licenses/CC-BY-NC-3.0-DE'),
  'CC-BY-NC-4.0': License(title='''Creative Commons Attribution Non Commercial 4.0 International''', description='''''', url='https://spdx.org/licenses/CC-BY-NC-4.0'),
  'CC-BY-NC-ND-1.0': License(title='''Creative Commons Attribution Non Commercial No Derivatives 1.0 Generic''', description='''''', url='https://spdx.org/licenses/CC-BY-NC-ND-1.0'),
  'CC-BY-NC-ND-2.0': License(title='''Creative Commons Attribution Non Commercial No Derivatives 2.0 Generic''', description='''''', url='https://spdx.org/licenses/CC-BY-NC-ND-2.0'),
  'CC-BY-NC-ND-2.5': License(title='''Creative Commons Attribution Non Commercial No Derivatives 2.5 Generic''', description='''''', url='https://spdx.org/licenses/CC-BY-NC-ND-2.5'),
  'CC-BY-NC-ND-3.0': License(title='''Creative Commons Attribution Non Commercial No Derivatives 3.0 Unported''', description='''''', url='https://spdx.org/licenses/CC-BY-NC-ND-3.0'),
  'CC-BY-NC-ND-3.0-DE': License(title='''Creative Commons Attribution Non Commercial No Derivatives 3.0 Germany''', description='''''', url='https://spdx.org/licenses/CC-BY-NC-ND-3.0-DE'),
  'CC-BY-NC-ND-3.0-IGO': License(title='''Creative Commons Attribution Non Commercial No Derivatives 3.0 IGO''', description='''''', url='https://spdx.org/licenses/CC-BY-NC-ND-3.0-IGO'),
  'CC-BY-NC-ND-4.0': License(title='''Creative Commons Attribution Non Commercial No Derivatives 4.0 International''', description='''''', url='https://spdx.org/licenses/CC-BY-NC-ND-4.0'),
  'CC-BY-NC-SA-1.0': License(title='''Creative Commons Attribution Non Commercial Share Alike 1.0 Generic''', description='''''', url='https://spdx.org/licenses/CC-BY-NC-SA-1.0'),
  'CC-BY-NC-SA-2.0': License(title='''Creative Commons Attribution Non Commercial Share Alike 2.0 Generic''', description='''''', url='https://spdx.org/licenses/CC-BY-NC-SA-2.0'),
  'CC-BY-NC-SA-2.0-DE': License(title='''Creative Commons Attribution Non Commercial Share Alike 2.0 Germany''', description='''''', url='https://spdx.org/licenses/CC-BY-NC-SA-2.0-DE'),
  'CC-BY-NC-SA-2.0-FR': License(title='''Creative Commons Attribution-NonCommercial-ShareAlike 2.0 France''', description='''''', url='https://spdx.org/licenses/CC-BY-NC-SA-2.0-FR'),
  'CC-BY-NC-SA-2.0-UK': License(title='''Creative Commons Attribution Non Commercial Share Alike 2.0 England and Wales''', description='''''', url='https://spdx.org/licenses/CC-BY-NC-SA-2.0-UK'),
  'CC-BY-NC-SA-2.5': License(title='''Creative Commons Attribution Non Commercial Share Alike 2.5 Generic''', description='''''', url='https://spdx.org/licenses/CC-BY-NC-SA-2.5'),
  'CC-BY-NC-SA-3.0': License(title='''Creative Commons Attribution Non Commercial Share Alike 3.0 Unported''', description='''''', url='https://spdx.org/licenses/CC-BY-NC-SA-3.0'),
  'CC-BY-NC-SA-3.0-DE': License(title='''Creative Commons Attribution Non Commercial Share Alike 3.0 Germany''', description='''''', url='https://spdx.org/licenses/CC-BY-NC-SA-3.0-DE'),
  'CC-BY-NC-SA-3.0-IGO': License(title='''Creative Commons Attribution Non Commercial Share Alike 3.0 IGO''', description='''''', url='https://spdx.org/licenses/CC-BY-NC-SA-3.0-IGO'),
  'CC-BY-NC-SA-4.0': License(title='''Creative Commons Attribution Non Commercial Share Alike 4.0 International''', description='''''', url='https://spdx.org/licenses/CC-BY-NC-SA-4.0'),
  'CC-BY-ND-1.0': License(title='''Creative Commons Attribution No Derivatives 1.0 Generic''', description='''''', url='https://spdx.org/licenses/CC-BY-ND-1.0'),
  'CC-BY-ND-2.0': License(title='''Creative Commons Attribution No Derivatives 2.0 Generic''', description='''''', url='https://spdx.org/licenses/CC-BY-ND-2.0'),
  'CC-BY-ND-2.5': License(title='''Creative Commons Attribution No Derivatives 2.5 Generic''', description='''''', url='https://spdx.org/licenses/CC-BY-ND-2.5'),
  'CC-BY-ND-3.0': License(title='''Creative Commons Attribution No Derivatives 3.0 Unported''', description='''''', url='https://spdx.org/licenses/CC-BY-ND-3.0'),
  'CC-BY-ND-3.0-DE': License(title='''Creative Commons Attribution No Derivatives 3.0 Germany''', description='''''', url='https://spdx.org/licenses/CC-BY-ND-3.0-DE'),
  'CC-BY-ND-4.0': License(title='''Creative Commons Attribution No Derivatives 4.0 International''', description='''''', url='https://spdx.org/licenses/CC-BY-ND-4.0'),
  'CC-BY-SA-1.0': License(title='''Creative Commons Attribution Share Alike 1.0 Generic''', description='''''', url='https://spdx.org/licenses/CC-BY-SA-1.0'),
  'CC-BY-SA-2.0': License(title='''Creative Commons Attribution Share Alike 2.0 Generic''', description='''''', url='https://spdx.org/licenses/CC-BY-SA-2.0'),
  'CC-BY-SA-2.0-UK': License(title='''Creative Commons Attribution Share Alike 2.0 England and Wales''', description='''''', url='https://spdx.org/licenses/CC-BY-SA-2.0-UK'),
  'CC-BY-SA-2.1-JP': License(title='''Creative Commons Attribution Share Alike 2.1 Japan''', description='''''', url='https://spdx.org/licenses/CC-BY-SA-2.1-JP'),
  'CC-BY-SA-2.5': License(title='''Creative Commons Attribution Share Alike 2.5 Generic''', description='''''', url='https://spdx.org/licenses/CC-BY-SA-2.5'),
  'CC-BY-SA-3.0': License(title='''Creative Commons Attribution Share Alike 3.0 Unported''', description='''''', url='https://spdx.org/licenses/CC-BY-SA-3.0'),
  'CC-BY-SA-3.0-AT': License(title='''Creative Commons Attribution Share Alike 3.0 Austria''', description='''''', url='https://spdx.org/licenses/CC-BY-SA-3.0-AT'),
  'CC-BY-SA-3.0-DE': License(title='''Creative Commons Attribution Share Alike 3.0 Germany''', description='''''', url='https://spdx.org/licenses/CC-BY-SA-3.0-DE'),
  'CC-BY-SA-4.0': License(title='''Creative Commons Attribution Share Alike 4.0 International''', description='''Similar to CC-BY-4.0 but requires derivatives be distributed under the same or a similar, compatible license. Frequently used for media assets and educational materials. A previous version is the default license for Wikipedia and other Wikimedia projects. Not recommended for software.''', url='https://spdx.org/licenses/CC-BY-SA-4.0'),
  'CC-PDDC': License(title='''Creative Commons Public Domain Dedication and Certification''', description='''''', url='https://spdx.org/licenses/CC-PDDC'),
  'CC0-1.0': License(title='''Creative Commons Zero v1.0 Universal''', description='''The Creative Commons CC0 Public Domain Dedication waives copyright interest in a work you've created and dedicates it to the world-wide public domain. Use CC0 to opt out of copyright entirely and ensure your work has the widest reach. As with the Unlicense and typical software licenses, CC0 disclaims warranties. CC0 is very similar to the Unlicense.''', url='https://spdx.org/licenses/CC0-1.0'),
  'CDDL-1.0': License(title='''Common Development and Distribution License 1.0''', description='''''', url='https://spdx.org/licenses/CDDL-1.0'),
  'CDDL-1.1': License(title='''Common Development and Distribution License 1.1''', description='''''', url='https://spdx.org/licenses/CDDL-1.1'),
  'CDL-1.0': License(title='''Common Documentation License 1.0''', description='''''', url='https://spdx.org/licenses/CDL-1.0'),
  'CDLA-Permissive-1.0': License(title='''Community Data License Agreement Permissive 1.0''', description='''''', url='https://spdx.org/licenses/CDLA-Permissive-1.0'),
  'CDLA-Permissive-2.0': License(title='''Community Data License Agreement Permissive 2.0''', description='''''', url='https://spdx.org/licenses/CDLA-Permissive-2.0'),
  'CDLA-Sharing-1.0': License(title='''Community Data License Agreement Sharing 1.0''', description='''''', url='https://spdx.org/licenses/CDLA-Sharing-1.0'),
  'CECILL-1.0': License(title='''CeCILL Free Software License Agreement v1.0''', description='''''', url='https://spdx.org/licenses/CECILL-1.0'),
  'CECILL-1.1': License(title='''CeCILL Free Software License Agreement v1.1''', description='''''', url='https://spdx.org/licenses/CECILL-1.1'),
  'CECILL-2.0': License(title='''CeCILL Free Software License Agreement v2.0''', description='''''', url='https://spdx.org/licenses/CECILL-2.0'),
  'CECILL-2.1': License(title='''CeCILL Free Software License Agreement v2.1''', description='''Strong copyleft license made by three French public research organisations, CEA, CNRS, and Inria, compatible with AGPL-3.0, EUPL-1.1, and GPL-2.0, or later versions of those licenses.''', url='https://spdx.org/licenses/CECILL-2.1'),
  'CECILL-B': License(title='''CeCILL-B Free Software License Agreement''', description='''''', url='https://spdx.org/licenses/CECILL-B'),
  'CECILL-C': License(title='''CeCILL-C Free Software License Agreement''', description='''''', url='https://spdx.org/licenses/CECILL-C'),
  'CERN-OHL-1.1': License(title='''CERN Open Hardware Licence v1.1''', description='''''', url='https://spdx.org/licenses/CERN-OHL-1.1'),
  'CERN-OHL-1.2': License(title='''CERN Open Hardware Licence v1.2''', description='''''', url='https://spdx.org/licenses/CERN-OHL-1.2'),
  'CERN-OHL-P-2.0': License(title='''CERN Open Hardware Licence Version 2 - Permissive''', description='''A permissive license for hardware designs, with conditions only requiring preservation of notices. Contributors provide an express grant of patent rights. Licensed works, modifications, and larger works may be distributed under different terms and without sources.''', url='https://spdx.org/licenses/CERN-OHL-P-2.0'),
  'CERN-OHL-S-2.0': License(title='''CERN Open Hardware Licence Version 2 - Strongly Reciprocal''', description='''Permissions of this strongly reciprocal license for hardware designs are conditioned on making available complete sources of licensed works and modifications, which include larger works using a licensed work, under the same license. Notices must be preserved. Contributors provide an express grant of patent rights.''', url='https://spdx.org/licenses/CERN-OHL-S-2.0'),
  'CERN-OHL-W-2.0': License(title='''CERN Open Hardware Licence Version 2 - Weakly Reciprocal''', description='''Permissions of this weakly-reciprocal license are conditioned on making available complete sources of licensed works and modifications under the same license. Notices must be preserved. Contributors provide an express grant of patent rights. However, a larger work using the licensed work through interfaces provided by the licensed work may be distributed under different terms and without sources for the larger work.''', url='https://spdx.org/licenses/CERN-OHL-W-2.0'),
  'CFITSIO': License(title='''CFITSIO License''', description='''''', url='https://spdx.org/licenses/CFITSIO'),
  'checkmk': License(title='''Checkmk License''', description='''''', url='https://spdx.org/licenses/checkmk'),
  'ClArtistic': License(title='''Clarified Artistic License''', description='''''', url='https://spdx.org/licenses/ClArtistic'),
  'Clips': License(title='''Clips License''', description='''''', url='https://spdx.org/licenses/Clips'),
  'CMU-Mach': License(title='''CMU Mach License''', description='''''', url='https://spdx.org/licenses/CMU-Mach'),
  'CNRI-Jython': License(title='''CNRI Jython License''', description='''''', url='https://spdx.org/licenses/CNRI-Jython'),
  'CNRI-Python': License(title='''CNRI Python License''', description='''''', url='https://spdx.org/licenses/CNRI-Python'),
  'CNRI-Python-GPL-Compatible': License(title='''CNRI Python Open Source GPL Compatible License Agreement''', description='''''', url='https://spdx.org/licenses/CNRI-Python-GPL-Compatible'),
  'COIL-1.0': License(title='''Copyfree Open Innovation License''', description='''''', url='https://spdx.org/licenses/COIL-1.0'),
  'Community-Spec-1.0': License(title='''Community Specification License 1.0''', description='''''', url='https://spdx.org/licenses/Community-Spec-1.0'),
  'Condor-1.1': License(title='''Condor Public License v1.1''', description='''''', url='https://spdx.org/licenses/Condor-1.1'),
  'copyleft-next-0.3.0': License(title='''copyleft-next 0.3.0''', description='''''', url='https://spdx.org/licenses/copyleft-next-0.3.0'),
  'copyleft-next-0.3.1': License(title='''copyleft-next 0.3.1''', description='''''', url='https://spdx.org/licenses/copyleft-next-0.3.1'),
  'Cornell-Lossless-JPEG': License(title='''Cornell Lossless JPEG License''', description='''''', url='https://spdx.org/licenses/Cornell-Lossless-JPEG'),
  'CPAL-1.0': License(title='''Common Public Attribution License 1.0''', description='''''', url='https://spdx.org/licenses/CPAL-1.0'),
  'CPL-1.0': License(title='''Common Public License 1.0''', description='''''', url='https://spdx.org/licenses/CPL-1.0'),
  'CPOL-1.02': License(title='''Code Project Open License 1.02''', description='''''', url='https://spdx.org/licenses/CPOL-1.02'),
  'Crossword': License(title='''Crossword License''', description='''''', url='https://spdx.org/licenses/Crossword'),
  'CrystalStacker': License(title='''CrystalStacker License''', description='''''', url='https://spdx.org/licenses/CrystalStacker'),
  'CUA-OPL-1.0': License(title='''CUA Office Public License v1.0''', description='''''', url='https://spdx.org/licenses/CUA-OPL-1.0'),
  'Cube': License(title='''Cube License''', description='''''', url='https://spdx.org/licenses/Cube'),
  'curl': License(title='''curl License''', description='''''', url='https://spdx.org/licenses/curl'),
  'D-FSL-1.0': License(title='''Deutsche Freie Software Lizenz''', description='''''', url='https://spdx.org/licenses/D-FSL-1.0'),
  'diffmark': License(title='''diffmark license''', description='''''', url='https://spdx.org/licenses/diffmark'),
  'DL-DE-BY-2.0': License(title='''Data licence Germany – attribution – version 2.0''', description='''''', url='https://spdx.org/licenses/DL-DE-BY-2.0'),
  'DOC': License(title='''DOC License''', description='''''', url='https://spdx.org/licenses/DOC'),
  'Dotseqn': License(title='''Dotseqn License''', description='''''', url='https://spdx.org/licenses/Dotseqn'),
  'DRL-1.0': License(title='''Detection Rule License 1.0''', description='''''', url='https://spdx.org/licenses/DRL-1.0'),
  'DSDP': License(title='''DSDP License''', description='''''', url='https://spdx.org/licenses/DSDP'),
  'dvipdfm': License(title='''dvipdfm License''', description='''''', url='https://spdx.org/licenses/dvipdfm'),
  'ECL-1.0': License(title='''Educational Community License v1.0''', description='''''', url='https://spdx.org/licenses/ECL-1.0'),
  'ECL-2.0': License(title='''Educational Community License v2.0''', description='''The Educational Community License version 2.0 ("ECL") consists of the Apache 2.0 license, modified to change the scope of the patent grant in section 3 to be specific to the needs of the education communities using this license.''', url='https://spdx.org/licenses/ECL-2.0'),
  'eCos-2.0': License(title='''eCos license version 2.0''', description='''''', url='https://spdx.org/licenses/eCos-2.0'),
  'EFL-1.0': License(title='''Eiffel Forum License v1.0''', description='''''', url='https://spdx.org/licenses/EFL-1.0'),
  'EFL-2.0': License(title='''Eiffel Forum License v2.0''', description='''''', url='https://spdx.org/licenses/EFL-2.0'),
  'eGenix': License(title='''eGenix.com Public License 1.1.0''', description='''''', url='https://spdx.org/licenses/eGenix'),
  'Elastic-2.0': License(title='''Elastic License 2.0''', description='''''', url='https://spdx.org/licenses/Elastic-2.0'),
  'Entessa': License(title='''Entessa Public License v1.0''', description='''''', url='https://spdx.org/licenses/Entessa'),
  'EPICS': License(title='''EPICS Open License''', description='''''', url='https://spdx.org/licenses/EPICS'),
  'EPL-1.0': License(title='''Eclipse Public License 1.0''', description='''This commercially-friendly copyleft license provides the ability to commercially license binaries; a modern royalty-free patent license grant; and the ability for linked works to use other licenses, including commercial ones.''', url='https://spdx.org/licenses/EPL-1.0'),
  'EPL-2.0': License(title='''Eclipse Public License 2.0''', description='''This commercially-friendly copyleft license provides the ability to commercially license binaries; a modern royalty-free patent license grant; and the ability for linked works to use other licenses, including commercial ones.''', url='https://spdx.org/licenses/EPL-2.0'),
  'ErlPL-1.1': License(title='''Erlang Public License v1.1''', description='''''', url='https://spdx.org/licenses/ErlPL-1.1'),
  'etalab-2.0': License(title='''Etalab Open License 2.0''', description='''''', url='https://spdx.org/licenses/etalab-2.0'),
  'EUDatagrid': License(title='''EU DataGrid Software License''', description='''''', url='https://spdx.org/licenses/EUDatagrid'),
  'EUPL-1.0': License(title='''European Union Public License 1.0''', description='''''', url='https://spdx.org/licenses/EUPL-1.0'),
  'EUPL-1.1': License(title='''European Union Public License 1.1''', description='''The “European Union Public Licence” (EUPL) is a copyleft free/open source software license created on the initiative of and approved by the European Commission in 22 official languages of the European Union.''', url='https://spdx.org/licenses/EUPL-1.1'),
  'EUPL-1.2': License(title='''European Union Public License 1.2''', description='''The European Union Public Licence (EUPL) is a copyleft free/open source software license created on the initiative of and approved by the European Commission in 23 official languages of the European Union.''', url='https://spdx.org/licenses/EUPL-1.2'),
  'Eurosym': License(title='''Eurosym License''', description='''''', url='https://spdx.org/licenses/Eurosym'),
  'Fair': License(title='''Fair License''', description='''''', url='https://spdx.org/licenses/Fair'),
  'FDK-AAC': License(title='''Fraunhofer FDK AAC Codec Library''', description='''''', url='https://spdx.org/licenses/FDK-AAC'),
  'Frameworx-1.0': License(title='''Frameworx Open License 1.0''', description='''''', url='https://spdx.org/licenses/Frameworx-1.0'),
  'FreeBSD-DOC': License(title='''FreeBSD Documentation License''', description='''''', url='https://spdx.org/licenses/FreeBSD-DOC'),
  'FreeImage': License(title='''FreeImage Public License v1.0''', description='''''', url='https://spdx.org/licenses/FreeImage'),
  'FSFAP': License(title='''FSF All Permissive License''', description='''''', url='https://spdx.org/licenses/FSFAP'),
  'FSFUL': License(title='''FSF Unlimited License''', description='''''', url='https://spdx.org/licenses/FSFUL'),
  'FSFULLR': License(title='''FSF Unlimited License (with License Retention)''', description='''''', url='https://spdx.org/licenses/FSFULLR'),
  'FSFULLRWD': License(title='''FSF Unlimited License (With License Retention    and Warranty Disclaimer)''', description='''''', url='https://spdx.org/licenses/FSFULLRWD'),
  'FTL': License(title='''Freetype Project License''', description='''''', url='https://spdx.org/licenses/FTL'),
  'GD': License(title='''GD License''', description='''''', url='https://spdx.org/licenses/GD'),
  'GFDL-1.1': License(title='''GNU Free Documentation License v1.1''', description='''''', url='https://spdx.org/licenses/GFDL-1.1'),
  'GFDL-1.1-invariants-only': License(title='''GNU Free Documentation License v1.1 only - invariants''', description='''''', url='https://spdx.org/licenses/GFDL-1.1-invariants-only'),
  'GFDL-1.1-invariants-or-later': License(title='''GNU Free Documentation License v1.1 or later - invariants''', description='''''', url='https://spdx.org/licenses/GFDL-1.1-invariants-or-later'),
  'GFDL-1.1-no-invariants-only': License(title='''GNU Free Documentation License v1.1 only - no invariants''', description='''''', url='https://spdx.org/licenses/GFDL-1.1-no-invariants-only'),
  'GFDL-1.1-no-invariants-or-later': License(title='''GNU Free Documentation License v1.1 or later - no invariants''', description='''''', url='https://spdx.org/licenses/GFDL-1.1-no-invariants-or-later'),
  'GFDL-1.1-only': License(title='''GNU Free Documentation License v1.1 only''', description='''''', url='https://spdx.org/licenses/GFDL-1.1-only'),
  'GFDL-1.1-or-later': License(title='''GNU Free Documentation License v1.1 or later''', description='''''', url='https://spdx.org/licenses/GFDL-1.1-or-later'),
  'GFDL-1.2': License(title='''GNU Free Documentation License v1.2''', description='''''', url='https://spdx.org/licenses/GFDL-1.2'),
  'GFDL-1.2-invariants-only': License(title='''GNU Free Documentation License v1.2 only - invariants''', description='''''', url='https://spdx.org/licenses/GFDL-1.2-invariants-only'),
  'GFDL-1.2-invariants-or-later': License(title='''GNU Free Documentation License v1.2 or later - invariants''', description='''''', url='https://spdx.org/licenses/GFDL-1.2-invariants-or-later'),
  'GFDL-1.2-no-invariants-only': License(title='''GNU Free Documentation License v1.2 only - no invariants''', description='''''', url='https://spdx.org/licenses/GFDL-1.2-no-invariants-only'),
  'GFDL-1.2-no-invariants-or-later': License(title='''GNU Free Documentation License v1.2 or later - no invariants''', description='''''', url='https://spdx.org/licenses/GFDL-1.2-no-invariants-or-later'),
  'GFDL-1.2-only': License(title='''GNU Free Documentation License v1.2 only''', description='''''', url='https://spdx.org/licenses/GFDL-1.2-only'),
  'GFDL-1.2-or-later': License(title='''GNU Free Documentation License v1.2 or later''', description='''''', url='https://spdx.org/licenses/GFDL-1.2-or-later'),
  'GFDL-1.3': License(title='''GNU Free Documentation License v1.3''', description='''A copyleft license for software documentation.''', url='https://spdx.org/licenses/GFDL-1.3'),
  'GFDL-1.3-invariants-only': License(title='''GNU Free Documentation License v1.3 only - invariants''', description='''''', url='https://spdx.org/licenses/GFDL-1.3-invariants-only'),
  'GFDL-1.3-invariants-or-later': License(title='''GNU Free Documentation License v1.3 or later - invariants''', description='''''', url='https://spdx.org/licenses/GFDL-1.3-invariants-or-later'),
  'GFDL-1.3-no-invariants-only': License(title='''GNU Free Documentation License v1.3 only - no invariants''', description='''''', url='https://spdx.org/licenses/GFDL-1.3-no-invariants-only'),
  'GFDL-1.3-no-invariants-or-later': License(title='''GNU Free Documentation License v1.3 or later - no invariants''', description='''''', url='https://spdx.org/licenses/GFDL-1.3-no-invariants-or-later'),
  'GFDL-1.3-only': License(title='''GNU Free Documentation License v1.3 only''', description='''''', url='https://spdx.org/licenses/GFDL-1.3-only'),
  'GFDL-1.3-or-later': License(title='''GNU Free Documentation License v1.3 or later''', description='''''', url='https://spdx.org/licenses/GFDL-1.3-or-later'),
  'Giftware': License(title='''Giftware License''', description='''''', url='https://spdx.org/licenses/Giftware'),
  'GL2PS': License(title='''GL2PS License''', description='''''', url='https://spdx.org/licenses/GL2PS'),
  'Glide': License(title='''3dfx Glide License''', description='''''', url='https://spdx.org/licenses/Glide'),
  'Glulxe': License(title='''Glulxe License''', description='''''', url='https://spdx.org/licenses/Glulxe'),
  'GLWTPL': License(title='''Good Luck With That Public License''', description='''''', url='https://spdx.org/licenses/GLWTPL'),
  'gnuplot': License(title='''gnuplot License''', description='''''', url='https://spdx.org/licenses/gnuplot'),
  'GPL-1.0': License(title='''GNU General Public License v1.0 only''', description='''''', url='https://spdx.org/licenses/GPL-1.0'),
  'GPL-1.0+': License(title='''GNU General Public License v1.0 or later''', description='''''', url='https://spdx.org/licenses/GPL-1.0+'),
  'GPL-1.0-only': License(title='''GNU General Public License v1.0 only''', description='''''', url='https://spdx.org/licenses/GPL-1.0-only'),
  'GPL-1.0-or-later': License(title='''GNU General Public License v1.0 or later''', description='''''', url='https://spdx.org/licenses/GPL-1.0-or-later'),
  'GPL-2.0': License(title='''GNU General Public License v2.0 only''', description='''The GNU GPL is the most widely used free software license and has a strong copyleft requirement. When distributing derived works, the source code of the work must be made available under the same license. There are multiple variants of the GNU GPL, each with different requirements.''', url='https://spdx.org/licenses/GPL-2.0'),
  'GPL-2.0+': License(title='''GNU General Public License v2.0 or later''', description='''''', url='https://spdx.org/licenses/GPL-2.0+'),
  'GPL-2.0-only': License(title='''GNU General Public License v2.0 only''', description='''''', url='https://spdx.org/licenses/GPL-2.0-only'),
  'GPL-2.0-or-later': License(title='''GNU General Public License v2.0 or later''', description='''''', url='https://spdx.org/licenses/GPL-2.0-or-later'),
  'GPL-2.0-with-autoconf-exception': License(title='''GNU General Public License v2.0 w/Autoconf exception''', description='''''', url='https://spdx.org/licenses/GPL-2.0-with-autoconf-exception'),
  'GPL-2.0-with-bison-exception': License(title='''GNU General Public License v2.0 w/Bison exception''', description='''''', url='https://spdx.org/licenses/GPL-2.0-with-bison-exception'),
  'GPL-2.0-with-classpath-exception': License(title='''GNU General Public License v2.0 w/Classpath exception''', description='''''', url='https://spdx.org/licenses/GPL-2.0-with-classpath-exception'),
  'GPL-2.0-with-font-exception': License(title='''GNU General Public License v2.0 w/Font exception''', description='''''', url='https://spdx.org/licenses/GPL-2.0-with-font-exception'),
  'GPL-2.0-with-GCC-exception': License(title='''GNU General Public License v2.0 w/GCC Runtime Library exception''', description='''''', url='https://spdx.org/licenses/GPL-2.0-with-GCC-exception'),
  'GPL-3.0': License(title='''GNU General Public License v3.0 only''', description='''Permissions of this strong copyleft license are conditioned on making available complete source code of licensed works and modifications, which include larger works using a licensed work, under the same license. Copyright and license notices must be preserved. Contributors provide an express grant of patent rights.''', url='https://spdx.org/licenses/GPL-3.0'),
  'GPL-3.0+': License(title='''GNU General Public License v3.0 or later''', description='''''', url='https://spdx.org/licenses/GPL-3.0+'),
  'GPL-3.0-only': License(title='''GNU General Public License v3.0 only''', description='''''', url='https://spdx.org/licenses/GPL-3.0-only'),
  'GPL-3.0-or-later': License(title='''GNU General Public License v3.0 or later''', description='''''', url='https://spdx.org/licenses/GPL-3.0-or-later'),
  'GPL-3.0-with-autoconf-exception': License(title='''GNU General Public License v3.0 w/Autoconf exception''', description='''''', url='https://spdx.org/licenses/GPL-3.0-with-autoconf-exception'),
  'GPL-3.0-with-GCC-exception': License(title='''GNU General Public License v3.0 w/GCC Runtime Library exception''', description='''''', url='https://spdx.org/licenses/GPL-3.0-with-GCC-exception'),
  'Graphics-Gems': License(title='''Graphics Gems License''', description='''''', url='https://spdx.org/licenses/Graphics-Gems'),
  'gSOAP-1.3b': License(title='''gSOAP Public License v1.3b''', description='''''', url='https://spdx.org/licenses/gSOAP-1.3b'),
  'HaskellReport': License(title='''Haskell Language Report License''', description='''''', url='https://spdx.org/licenses/HaskellReport'),
  'Hippocratic-2.1': License(title='''Hippocratic License 2.1''', description='''''', url='https://spdx.org/licenses/Hippocratic-2.1'),
  'HP-1986': License(title='''Hewlett-Packard 1986 License''', description='''''', url='https://spdx.org/licenses/HP-1986'),
  'HPND': License(title='''Historical Permission Notice and Disclaimer''', description='''''', url='https://spdx.org/licenses/HPND'),
  'HPND-export-US': License(title='''HPND with US Government export control warning''', description='''''', url='https://spdx.org/licenses/HPND-export-US'),
  'HPND-Markus-Kuhn': License(title='''Historical Permission Notice and Disclaimer - Markus Kuhn variant''', description='''''', url='https://spdx.org/licenses/HPND-Markus-Kuhn'),
  'HPND-sell-variant': License(title='''Historical Permission Notice and Disclaimer - sell variant''', description='''''', url='https://spdx.org/licenses/HPND-sell-variant'),
  'HPND-sell-variant-MIT-disclaimer': License(title='''HPND sell variant with MIT disclaimer''', description='''''', url='https://spdx.org/licenses/HPND-sell-variant-MIT-disclaimer'),
  'HTMLTIDY': License(title='''HTML Tidy License''', description='''''', url='https://spdx.org/licenses/HTMLTIDY'),
  'IBM-pibs': License(title='''IBM PowerPC Initialization and Boot Software''', description='''''', url='https://spdx.org/licenses/IBM-pibs'),
  'ICU': License(title='''ICU License''', description='''''', url='https://spdx.org/licenses/ICU'),
  'IEC-Code-Components-EULA': License(title='''IEC    Code Components End-user licence agreement''', description='''''', url='https://spdx.org/licenses/IEC-Code-Components-EULA'),
  'IJG': License(title='''Independent JPEG Group License''', description='''''', url='https://spdx.org/licenses/IJG'),
  'IJG-short': License(title='''Independent JPEG Group License - short''', description='''''', url='https://spdx.org/licenses/IJG-short'),
  'ImageMagick': License(title='''ImageMagick License''', description='''''', url='https://spdx.org/licenses/ImageMagick'),
  'iMatix': License(title='''iMatix Standard Function Library Agreement''', description='''''', url='https://spdx.org/licenses/iMatix'),
  'Imlib2': License(title='''Imlib2 License''', description='''''', url='https://spdx.org/licenses/Imlib2'),
  'Info-ZIP': License(title='''Info-ZIP License''', description='''''', url='https://spdx.org/licenses/Info-ZIP'),
  'Intel': License(title='''Intel Open Source License''', description='''''', url='https://spdx.org/licenses/Intel'),
  'Intel-ACPI': License(title='''Intel ACPI Software License Agreement''', description='''''', url='https://spdx.org/licenses/Intel-ACPI'),
  'Interbase-1.0': License(title='''Interbase Public License v1.0''', description='''''', url='https://spdx.org/licenses/Interbase-1.0'),
  'IPA': License(title='''IPA Font License''', description='''''', url='https://spdx.org/licenses/IPA'),
  'IPL-1.0': License(title='''IBM Public License v1.0''', description='''''', url='https://spdx.org/licenses/IPL-1.0'),
  'ISC': License(title='''ISC License''', description='''A permissive license lets people do anything with your code with proper attribution and without warranty. The ISC license is functionally equivalent to the BSD 2-Clause and MIT licenses, removing some language that is no longer necessary.''', url='https://spdx.org/licenses/ISC'),
  'Jam': License(title='''Jam License''', description='''''', url='https://spdx.org/licenses/Jam'),
  'JasPer-2.0': License(title='''JasPer License''', description='''''', url='https://spdx.org/licenses/JasPer-2.0'),
  'JPL-image': License(title='''JPL Image Use Policy''', description='''''', url='https://spdx.org/licenses/JPL-image'),
  'JPNIC': License(title='''Japan Network Information Center License''', description='''''', url='https://spdx.org/licenses/JPNIC'),
  'JSON': License(title='''JSON License''', description='''''', url='https://spdx.org/licenses/JSON'),
  'Kazlib': License(title='''Kazlib License''', description='''''', url='https://spdx.org/licenses/Kazlib'),
  'Knuth-CTAN': License(title='''Knuth CTAN License''', description='''''', url='https://spdx.org/licenses/Knuth-CTAN'),
  'LAL-1.2': License(title='''Licence Art Libre 1.2''', description='''''', url='https://spdx.org/licenses/LAL-1.2'),
  'LAL-1.3': License(title='''Licence Art Libre 1.3''', description='''''', url='https://spdx.org/licenses/LAL-1.3'),
  'Latex2e': License(title='''Latex2e License''', description='''''', url='https://spdx.org/licenses/Latex2e'),
  'Leptonica': License(title='''Leptonica License''', description='''''', url='https://spdx.org/licenses/Leptonica'),
  'LGPL-2.0': License(title='''GNU Library General Public License v2 only''', description='''''', url='https://spdx.org/licenses/LGPL-2.0'),
  'LGPL-2.0+': License(title='''GNU Library General Public License v2 or later''', description='''''', url='https://spdx.org/licenses/LGPL-2.0+'),
  'LGPL-2.0-only': License(title='''GNU Library General Public License v2 only''', description='''''', url='https://spdx.org/licenses/LGPL-2.0-only'),
  'LGPL-2.0-or-later': License(title='''GNU Library General Public License v2 or later''', description='''''', url='https://spdx.org/licenses/LGPL-2.0-or-later'),
  'LGPL-2.1': License(title='''GNU Lesser General Public License v2.1 only''', description='''Primarily used for software libraries, the GNU LGPL requires that derived works be licensed under the same license, but works that only link to it do not fall under this restriction. There are two commonly used versions of the GNU LGPL.''', url='https://spdx.org/licenses/LGPL-2.1'),
  'LGPL-2.1+': License(title='''GNU Lesser General Public License v2.1 or later''', description='''''', url='https://spdx.org/licenses/LGPL-2.1+'),
  'LGPL-2.1-only': License(title='''GNU Lesser General Public License v2.1 only''', description='''''', url='https://spdx.org/licenses/LGPL-2.1-only'),
  'LGPL-2.1-or-later': License(title='''GNU Lesser General Public License v2.1 or later''', description='''''', url='https://spdx.org/licenses/LGPL-2.1-or-later'),
  'LGPL-3.0': License(title='''GNU Lesser General Public License v3.0 only''', description='''Permissions of this copyleft license are conditioned on making available complete source code of licensed works and modifications under the same license or the GNU GPLv3. Copyright and license notices must be preserved. Contributors provide an express grant of patent rights. However, a larger work using the licensed work through interfaces provided by the licensed work may be distributed under different terms and without source code for the larger work.''', url='https://spdx.org/licenses/LGPL-3.0'),
  'LGPL-3.0+': License(title='''GNU Lesser General Public License v3.0 or later''', description='''''', url='https://spdx.org/licenses/LGPL-3.0+'),
  'LGPL-3.0-only': License(title='''GNU Lesser General Public License v3.0 only''', description='''''', url='https://spdx.org/licenses/LGPL-3.0-only'),
  'LGPL-3.0-or-later': License(title='''GNU Lesser General Public License v3.0 or later''', description='''''', url='https://spdx.org/licenses/LGPL-3.0-or-later'),
  'LGPLLR': License(title='''Lesser General Public License For Linguistic Resources''', description='''''', url='https://spdx.org/licenses/LGPLLR'),
  'Libpng': License(title='''libpng License''', description='''''', url='https://spdx.org/licenses/Libpng'),
  'libpng-2.0': License(title='''PNG Reference Library version 2''', description='''''', url='https://spdx.org/licenses/libpng-2.0'),
  'libselinux-1.0': License(title='''libselinux public domain notice''', description='''''', url='https://spdx.org/licenses/libselinux-1.0'),
  'libtiff': License(title='''libtiff License''', description='''''', url='https://spdx.org/licenses/libtiff'),
  'libutil-David-Nugent': License(title='''libutil David Nugent License''', description='''''', url='https://spdx.org/licenses/libutil-David-Nugent'),
  'LiLiQ-P-1.1': License(title='''Licence Libre du Québec – Permissive version 1.1''', description='''''', url='https://spdx.org/licenses/LiLiQ-P-1.1'),
  'LiLiQ-R-1.1': License(title='''Licence Libre du Québec – Réciprocité version 1.1''', description='''''', url='https://spdx.org/licenses/LiLiQ-R-1.1'),
  'LiLiQ-Rplus-1.1': License(title='''Licence Libre du Québec – Réciprocité forte version 1.1''', description='''''', url='https://spdx.org/licenses/LiLiQ-Rplus-1.1'),
  'Linux-man-pages-copyleft': License(title='''Linux man-pages Copyleft''', description='''''', url='https://spdx.org/licenses/Linux-man-pages-copyleft'),
  'Linux-OpenIB': License(title='''Linux Kernel Variant of OpenIB.org license''', description='''''', url='https://spdx.org/licenses/Linux-OpenIB'),
  'LOOP': License(title='''Common Lisp LOOP License''', description='''''', url='https://spdx.org/licenses/LOOP'),
  'LPL-1.0': License(title='''Lucent Public License Version 1.0''', description='''''', url='https://spdx.org/licenses/LPL-1.0'),
  'LPL-1.02': License(title='''Lucent Public License v1.02''', description='''''', url='https://spdx.org/licenses/LPL-1.02'),
  'LPPL-1.0': License(title='''LaTeX Project Public License v1.0''', description='''''', url='https://spdx.org/licenses/LPPL-1.0'),
  'LPPL-1.1': License(title='''LaTeX Project Public License v1.1''', description='''''', url='https://spdx.org/licenses/LPPL-1.1'),
  'LPPL-1.2': License(title='''LaTeX Project Public License v1.2''', description='''''', url='https://spdx.org/licenses/LPPL-1.2'),
  'LPPL-1.3a': License(title='''LaTeX Project Public License v1.3a''', description='''''', url='https://spdx.org/licenses/LPPL-1.3a'),
  'LPPL-1.3c': License(title='''LaTeX Project Public License v1.3c''', description='''The LaTeX Project Public License (LPPL) is the primary license under which the LaTeX kernel and the base LaTeX packages are distributed.''', url='https://spdx.org/licenses/LPPL-1.3c'),
  'LZMA-SDK-9.11-to-9.20': License(title='''LZMA SDK License (versions 9.11 to 9.20)''', description='''''', url='https://spdx.org/licenses/LZMA-SDK-9.11-to-9.20'),
  'LZMA-SDK-9.22': License(title='''LZMA SDK License (versions 9.22 and beyond)''', description='''''', url='https://spdx.org/licenses/LZMA-SDK-9.22'),
  'MakeIndex': License(title='''MakeIndex License''', description='''''', url='https://spdx.org/licenses/MakeIndex'),
  'Martin-Birgmeier': License(title='''Martin Birgmeier License''', description='''''', url='https://spdx.org/licenses/Martin-Birgmeier'),
  'Minpack': License(title='''Minpack License''', description='''''', url='https://spdx.org/licenses/Minpack'),
  'MirOS': License(title='''The MirOS Licence''', description='''''', url='https://spdx.org/licenses/MirOS'),
  'MIT': License(title='''MIT License''', description='''A short and simple permissive license with conditions only requiring preservation of copyright and license notices. Licensed works, modifications, and larger works may be distributed under different terms and without source code.''', url='https://spdx.org/licenses/MIT'),
  'MIT-0': License(title='''MIT No Attribution''', description='''A short and simple permissive license with no conditions, not even requiring preservation of copyright and license notices. Licensed works, modifications, and larger works may be distributed under different terms and without source code.''', url='https://spdx.org/licenses/MIT-0'),
  'MIT-advertising': License(title='''Enlightenment License (e16)''', description='''''', url='https://spdx.org/licenses/MIT-advertising'),
  'MIT-CMU': License(title='''CMU License''', description='''''', url='https://spdx.org/licenses/MIT-CMU'),
  'MIT-enna': License(title='''enna License''', description='''''', url='https://spdx.org/licenses/MIT-enna'),
  'MIT-feh': License(title='''feh License''', description='''''', url='https://spdx.org/licenses/MIT-feh'),
  'MIT-Modern-Variant': License(title='''MIT License Modern Variant''', description='''''', url='https://spdx.org/licenses/MIT-Modern-Variant'),
  'MIT-open-group': License(title='''MIT Open Group variant''', description='''''', url='https://spdx.org/licenses/MIT-open-group'),
  'MIT-Wu': License(title='''MIT Tom Wu Variant''', description='''''', url='https://spdx.org/licenses/MIT-Wu'),
  'MITNFA': License(title='''MIT +no-false-attribs license''', description='''''', url='https://spdx.org/licenses/MITNFA'),
  'Motosoto': License(title='''Motosoto License''', description='''''', url='https://spdx.org/licenses/Motosoto'),
  'mpi-permissive': License(title='''mpi Permissive License''', description='''''', url='https://spdx.org/licenses/mpi-permissive'),
  'mpich2': License(title='''mpich2 License''', description='''''', url='https://spdx.org/licenses/mpich2'),
  'MPL-1.0': License(title='''Mozilla Public License 1.0''', description='''''', url='https://spdx.org/licenses/MPL-1.0'),
  'MPL-1.1': License(title='''Mozilla Public License 1.1''', description='''''', url='https://spdx.org/licenses/MPL-1.1'),
  'MPL-2.0': License(title='''Mozilla Public License 2.0''', description='''Permissions of this weak copyleft license are conditioned on making available source code of licensed files and modifications of those files under the same license (or in certain cases, one of the GNU licenses). Copyright and license notices must be preserved. Contributors provide an express grant of patent rights. However, a larger work using the licensed work may be distributed under different terms and without source code for files added in the larger work.''', url='https://spdx.org/licenses/MPL-2.0'),
  'MPL-2.0-no-copyleft-exception': License(title='''Mozilla Public License 2.0 (no copyleft exception)''', description='''''', url='https://spdx.org/licenses/MPL-2.0-no-copyleft-exception'),
  'mplus': License(title='''mplus Font License''', description='''''', url='https://spdx.org/licenses/mplus'),
  'MS-LPL': License(title='''Microsoft Limited Public License''', description='''''', url='https://spdx.org/licenses/MS-LPL'),
  'MS-PL': License(title='''Microsoft Public License''', description='''An open source license with a patent grant.''', url='https://spdx.org/licenses/MS-PL'),
  'MS-RL': License(title='''Microsoft Reciprocal License''', description='''An open source license with a patent grant similar to the Microsoft Public License, with the additional condition that any source code for any derived file be provided under this license.''', url='https://spdx.org/licenses/MS-RL'),
  'MTLL': License(title='''Matrix Template Library License''', description='''''', url='https://spdx.org/licenses/MTLL'),
  'MulanPSL-1.0': License(title='''Mulan Permissive Software License, Version 1''', description='''''', url='https://spdx.org/licenses/MulanPSL-1.0'),
  'MulanPSL-2.0': License(title='''Mulan Permissive Software License, Version 2''', description='''A permissive license similar to the Apache License, but explicitly states that it does not grant trademark rights. Mulan Permissive Software License is the first open source license in both Chinese and English approved by OSI.''', url='https://spdx.org/licenses/MulanPSL-2.0'),
  'Multics': License(title='''Multics License''', description='''''', url='https://spdx.org/licenses/Multics'),
  'Mup': License(title='''Mup License''', description='''''', url='https://spdx.org/licenses/Mup'),
  'NAIST-2003': License(title='''Nara Institute of Science and Technology License (2003)''', description='''''', url='https://spdx.org/licenses/NAIST-2003'),
  'NASA-1.3': License(title='''NASA Open Source Agreement 1.3''', description='''''', url='https://spdx.org/licenses/NASA-1.3'),
  'Naumen': License(title='''Naumen Public License''', description='''''', url='https://spdx.org/licenses/Naumen'),
  'NBPL-1.0': License(title='''Net Boolean Public License v1''', description='''''', url='https://spdx.org/licenses/NBPL-1.0'),
  'NCGL-UK-2.0': License(title='''Non-Commercial Government Licence''', description='''''', url='https://spdx.org/licenses/NCGL-UK-2.0'),
  'NCSA': License(title='''University of Illinois/NCSA Open Source License''', description='''The University of Illinois/NCSA Open Source License, or UIUC license, is a permissive free software license, based on the MIT/X11 license  and the BSD 3-clause License. Its conditions include requiring the preservation of copyright and license notices both in source and in binary distributions and the prohibition of using the names of the authors or the project organization to promote or endorse derived products.''', url='https://spdx.org/licenses/NCSA'),
  'Net-SNMP': License(title='''Net-SNMP License''', description='''''', url='https://spdx.org/licenses/Net-SNMP'),
  'NetCDF': License(title='''NetCDF license''', description='''''', url='https://spdx.org/licenses/NetCDF'),
  'Newsletr': License(title='''Newsletr License''', description='''''', url='https://spdx.org/licenses/Newsletr'),
  'NGPL': License(title='''Nethack General Public License''', description='''''', url='https://spdx.org/licenses/NGPL'),
  'NICTA-1.0': License(title='''NICTA Public Software License, Version 1.0''', description='''''', url='https://spdx.org/licenses/NICTA-1.0'),
  'NIST-PD': License(title='''NIST Public Domain Notice''', description='''''', url='https://spdx.org/licenses/NIST-PD'),
  'NIST-PD-fallback': License(title='''NIST Public Domain Notice with license fallback''', description='''''', url='https://spdx.org/licenses/NIST-PD-fallback'),
  'NLOD-1.0': License(title='''Norwegian Licence for Open Government Data (NLOD) 1.0''', description='''''', url='https://spdx.org/licenses/NLOD-1.0'),
  'NLOD-2.0': License(title='''Norwegian Licence for Open Government Data (NLOD) 2.0''', description='''''', url='https://spdx.org/licenses/NLOD-2.0'),
  'NLPL': License(title='''No Limit Public License''', description='''''', url='https://spdx.org/licenses/NLPL'),
  'Nokia': License(title='''Nokia Open Source License''', description='''''', url='https://spdx.org/licenses/Nokia'),
  'NOSL': License(title='''Netizen Open Source License''', description='''''', url='https://spdx.org/licenses/NOSL'),
  'Noweb': License(title='''Noweb License''', description='''''', url='https://spdx.org/licenses/Noweb'),
  'NPL-1.0': License(title='''Netscape Public License v1.0''', description='''''', url='https://spdx.org/licenses/NPL-1.0'),
  'NPL-1.1': License(title='''Netscape Public License v1.1''', description='''''', url='https://spdx.org/licenses/NPL-1.1'),
  'NPOSL-3.0': License(title='''Non-Profit Open Software License 3.0''', description='''''', url='https://spdx.org/licenses/NPOSL-3.0'),
  'NRL': License(title='''NRL License''', description='''''', url='https://spdx.org/licenses/NRL'),
  'NTP': License(title='''NTP License''', description='''''', url='https://spdx.org/licenses/NTP'),
  'NTP-0': License(title='''NTP No Attribution''', description='''''', url='https://spdx.org/licenses/NTP-0'),
  'Nunit': License(title='''Nunit License''', description='''''', url='https://spdx.org/licenses/Nunit'),
  'O-UDA-1.0': License(title='''Open Use of Data Agreement v1.0''', description='''''', url='https://spdx.org/licenses/O-UDA-1.0'),
  'OCCT-PL': License(title='''Open CASCADE Technology Public License''', description='''''', url='https://spdx.org/licenses/OCCT-PL'),
  'OCLC-2.0': License(title='''OCLC Research Public License 2.0''', description='''''', url='https://spdx.org/licenses/OCLC-2.0'),
  'ODbL-1.0': License(title='''Open Data Commons Open Database License v1.0''', description='''The Open Database License (ODbL) is a license agreement intended to allow users to freely share, modify, and use a database while maintaining this same freedom for others.''', url='https://spdx.org/licenses/ODbL-1.0'),
  'ODC-By-1.0': License(title='''Open Data Commons Attribution License v1.0''', description='''''', url='https://spdx.org/licenses/ODC-By-1.0'),
  'OFFIS': License(title='''OFFIS License''', description='''''', url='https://spdx.org/licenses/OFFIS'),
  'OFL-1.0': License(title='''SIL Open Font License 1.0''', description='''''', url='https://spdx.org/licenses/OFL-1.0'),
  'OFL-1.0-no-RFN': License(title='''SIL Open Font License 1.0 with no Reserved Font Name''', description='''''', url='https://spdx.org/licenses/OFL-1.0-no-RFN'),
  'OFL-1.0-RFN': License(title='''SIL Open Font License 1.0 with Reserved Font Name''', description='''''', url='https://spdx.org/licenses/OFL-1.0-RFN'),
  'OFL-1.1': License(title='''SIL Open Font License 1.1''', description='''The Open Font License (OFL) is maintained by SIL International. It attempts to be a compromise between the values of the free software and typeface design communities. It is used for almost all open source font projects, including those by Adobe, Google and Mozilla.''', url='https://spdx.org/licenses/OFL-1.1'),
  'OFL-1.1-no-RFN': License(title='''SIL Open Font License 1.1 with no Reserved Font Name''', description='''''', url='https://spdx.org/licenses/OFL-1.1-no-RFN'),
  'OFL-1.1-RFN': License(title='''SIL Open Font License 1.1 with Reserved Font Name''', description='''''', url='https://spdx.org/licenses/OFL-1.1-RFN'),
  'OGC-1.0': License(title='''OGC Software License, Version 1.0''', description='''''', url='https://spdx.org/licenses/OGC-1.0'),
  'OGDL-Taiwan-1.0': License(title='''Taiwan Open Government Data License, version 1.0''', description='''''', url='https://spdx.org/licenses/OGDL-Taiwan-1.0'),
  'OGL-Canada-2.0': License(title='''Open Government Licence - Canada''', description='''''', url='https://spdx.org/licenses/OGL-Canada-2.0'),
  'OGL-UK-1.0': License(title='''Open Government Licence v1.0''', description='''''', url='https://spdx.org/licenses/OGL-UK-1.0'),
  'OGL-UK-2.0': License(title='''Open Government Licence v2.0''', description='''''', url='https://spdx.org/licenses/OGL-UK-2.0'),
  'OGL-UK-3.0': License(title='''Open Government Licence v3.0''', description='''''', url='https://spdx.org/licenses/OGL-UK-3.0'),
  'OGTSL': License(title='''Open Group Test Suite License''', description='''''', url='https://spdx.org/licenses/OGTSL'),
  'OLDAP-1.1': License(title='''Open LDAP Public License v1.1''', description='''''', url='https://spdx.org/licenses/OLDAP-1.1'),
  'OLDAP-1.2': License(title='''Open LDAP Public License v1.2''', description='''''', url='https://spdx.org/licenses/OLDAP-1.2'),
  'OLDAP-1.3': License(title='''Open LDAP Public License v1.3''', description='''''', url='https://spdx.org/licenses/OLDAP-1.3'),
  'OLDAP-1.4': License(title='''Open LDAP Public License v1.4''', description='''''', url='https://spdx.org/licenses/OLDAP-1.4'),
  'OLDAP-2.0': License(title='''Open LDAP Public License v2.0 (or possibly 2.0A and 2.0B)''', description='''''', url='https://spdx.org/licenses/OLDAP-2.0'),
  'OLDAP-2.0.1': License(title='''Open LDAP Public License v2.0.1''', description='''''', url='https://spdx.org/licenses/OLDAP-2.0.1'),
  'OLDAP-2.1': License(title='''Open LDAP Public License v2.1''', description='''''', url='https://spdx.org/licenses/OLDAP-2.1'),
  'OLDAP-2.2': License(title='''Open LDAP Public License v2.2''', description='''''', url='https://spdx.org/licenses/OLDAP-2.2'),
  'OLDAP-2.2.1': License(title='''Open LDAP Public License v2.2.1''', description='''''', url='https://spdx.org/licenses/OLDAP-2.2.1'),
  'OLDAP-2.2.2': License(title='''Open LDAP Public License 2.2.2''', description='''''', url='https://spdx.org/licenses/OLDAP-2.2.2'),
  'OLDAP-2.3': License(title='''Open LDAP Public License v2.3''', description='''''', url='https://spdx.org/licenses/OLDAP-2.3'),
  'OLDAP-2.4': License(title='''Open LDAP Public License v2.4''', description='''''', url='https://spdx.org/licenses/OLDAP-2.4'),
  'OLDAP-2.5': License(title='''Open LDAP Public License v2.5''', description='''''', url='https://spdx.org/licenses/OLDAP-2.5'),
  'OLDAP-2.6': License(title='''Open LDAP Public License v2.6''', description='''''', url='https://spdx.org/licenses/OLDAP-2.6'),
  'OLDAP-2.7': License(title='''Open LDAP Public License v2.7''', description='''''', url='https://spdx.org/licenses/OLDAP-2.7'),
  'OLDAP-2.8': License(title='''Open LDAP Public License v2.8''', description='''''', url='https://spdx.org/licenses/OLDAP-2.8'),
  'OML': License(title='''Open Market License''', description='''''', url='https://spdx.org/licenses/OML'),
  'OpenPBS-2.3': License(title='''OpenPBS v2.3 Software License''', description='''''', url='https://spdx.org/licenses/OpenPBS-2.3'),
  'OpenSSL': License(title='''OpenSSL License''', description='''''', url='https://spdx.org/licenses/OpenSSL'),
  'OPL-1.0': License(title='''Open Public License v1.0''', description='''''', url='https://spdx.org/licenses/OPL-1.0'),
  'OPUBL-1.0': License(title='''Open Publication License v1.0''', description='''''', url='https://spdx.org/licenses/OPUBL-1.0'),
  'OSET-PL-2.1': License(title='''OSET Public License version 2.1''', description='''''', url='https://spdx.org/licenses/OSET-PL-2.1'),
  'OSL-1.0': License(title='''Open Software License 1.0''', description='''''', url='https://spdx.org/licenses/OSL-1.0'),
  'OSL-1.1': License(title='''Open Software License 1.1''', description='''''', url='https://spdx.org/licenses/OSL-1.1'),
  'OSL-2.0': License(title='''Open Software License 2.0''', description='''''', url='https://spdx.org/licenses/OSL-2.0'),
  'OSL-2.1': License(title='''Open Software License 2.1''', description='''''', url='https://spdx.org/licenses/OSL-2.1'),
  'OSL-3.0': License(title='''Open Software License 3.0''', description='''OSL 3.0 is a copyleft license that does not require reciprocal licensing on linked works. It also provides an express grant of patent rights from contributors to users, with a termination clause triggered if a user files a patent infringement lawsuit.''', url='https://spdx.org/licenses/OSL-3.0'),
  'Parity-6.0.0': License(title='''The Parity Public License 6.0.0''', description='''''', url='https://spdx.org/licenses/Parity-6.0.0'),
  'Parity-7.0.0': License(title='''The Parity Public License 7.0.0''', description='''''', url='https://spdx.org/licenses/Parity-7.0.0'),
  'PDDL-1.0': License(title='''Open Data Commons Public Domain Dedication & License 1.0''', description='''''', url='https://spdx.org/licenses/PDDL-1.0'),
  'PHP-3.0': License(title='''PHP License v3.0''', description='''''', url='https://spdx.org/licenses/PHP-3.0'),
  'PHP-3.01': License(title='''PHP License v3.01''', description='''''', url='https://spdx.org/licenses/PHP-3.01'),
  'Plexus': License(title='''Plexus Classworlds License''', description='''''', url='https://spdx.org/licenses/Plexus'),
  'PolyForm-Noncommercial-1.0.0': License(title='''PolyForm Noncommercial License 1.0.0''', description='''''', url='https://spdx.org/licenses/PolyForm-Noncommercial-1.0.0'),
  'PolyForm-Small-Business-1.0.0': License(title='''PolyForm Small Business License 1.0.0''', description='''''', url='https://spdx.org/licenses/PolyForm-Small-Business-1.0.0'),
  'PostgreSQL': License(title='''PostgreSQL License''', description='''A very short, BSD-style license, used specifically for PostgreSQL.''', url='https://spdx.org/licenses/PostgreSQL'),
  'PSF-2.0': License(title='''Python Software Foundation License 2.0''', description='''''', url='https://spdx.org/licenses/PSF-2.0'),
  'psfrag': License(title='''psfrag License''', description='''''', url='https://spdx.org/licenses/psfrag'),
  'psutils': License(title='''psutils License''', description='''''', url='https://spdx.org/licenses/psutils'),
  'Python-2.0': License(title='''Python License 2.0''', description='''''', url='https://spdx.org/licenses/Python-2.0'),
  'Python-2.0.1': License(title='''Python License 2.0.1''', description='''''', url='https://spdx.org/licenses/Python-2.0.1'),
  'Qhull': License(title='''Qhull License''', description='''''', url='https://spdx.org/licenses/Qhull'),
  'QPL-1.0': License(title='''Q Public License 1.0''', description='''''', url='https://spdx.org/licenses/QPL-1.0'),
  'QPL-1.0-INRIA-2004': License(title='''Q Public License 1.0 - INRIA 2004 variant''', description='''''', url='https://spdx.org/licenses/QPL-1.0-INRIA-2004'),
  'Rdisc': License(title='''Rdisc License''', description='''''', url='https://spdx.org/licenses/Rdisc'),
  'RHeCos-1.1': License(title='''Red Hat eCos Public License v1.1''', description='''''', url='https://spdx.org/licenses/RHeCos-1.1'),
  'RPL-1.1': License(title='''Reciprocal Public License 1.1''', description='''''', url='https://spdx.org/licenses/RPL-1.1'),
  'RPL-1.5': License(title='''Reciprocal Public License 1.5''', description='''''', url='https://spdx.org/licenses/RPL-1.5'),
  'RPSL-1.0': License(title='''RealNetworks Public Source License v1.0''', description='''''', url='https://spdx.org/licenses/RPSL-1.0'),
  'RSA-MD': License(title='''RSA Message-Digest License''', description='''''', url='https://spdx.org/licenses/RSA-MD'),
  'RSCPL': License(title='''Ricoh Source Code Public License''', description='''''', url='https://spdx.org/licenses/RSCPL'),
  'Ruby': License(title='''Ruby License''', description='''''', url='https://spdx.org/licenses/Ruby'),
  'SAX-PD': License(title='''Sax Public Domain Notice''', description='''''', url='https://spdx.org/licenses/SAX-PD'),
  'Saxpath': License(title='''Saxpath License''', description='''''', url='https://spdx.org/licenses/Saxpath'),
  'SCEA': License(title='''SCEA Shared Source License''', description='''''', url='https://spdx.org/licenses/SCEA'),
  'SchemeReport': License(title='''Scheme Language Report License''', description='''''', url='https://spdx.org/licenses/SchemeReport'),
  'Sendmail': License(title='''Sendmail License''', description='''''', url='https://spdx.org/licenses/Sendmail'),
  'Sendmail-8.23': License(title='''Sendmail License 8.23''', description='''''', url='https://spdx.org/licenses/Sendmail-8.23'),
  'SGI-B-1.0': License(title='''SGI Free Software License B v1.0''', description='''''', url='https://spdx.org/licenses/SGI-B-1.0'),
  'SGI-B-1.1': License(title='''SGI Free Software License B v1.1''', description='''''', url='https://spdx.org/licenses/SGI-B-1.1'),
  'SGI-B-2.0': License(title='''SGI Free Software License B v2.0''', description='''''', url='https://spdx.org/licenses/SGI-B-2.0'),
  'SHL-0.5': License(title='''Solderpad Hardware License v0.5''', description='''''', url='https://spdx.org/licenses/SHL-0.5'),
  'SHL-0.51': License(title='''Solderpad Hardware License, Version 0.51''', description='''''', url='https://spdx.org/licenses/SHL-0.51'),
  'SimPL-2.0': License(title='''Simple Public License 2.0''', description='''''', url='https://spdx.org/licenses/SimPL-2.0'),
  'SISSL': License(title='''Sun Industry Standards Source License v1.1''', description='''''', url='https://spdx.org/licenses/SISSL'),
  'SISSL-1.2': License(title='''Sun Industry Standards Source License v1.2''', description='''''', url='https://spdx.org/licenses/SISSL-1.2'),
  'Sleepycat': License(title='''Sleepycat License''', description='''''', url='https://spdx.org/licenses/Sleepycat'),
  'SMLNJ': License(title='''Standard ML of New Jersey License''', description='''''', url='https://spdx.org/licenses/SMLNJ'),
  'SMPPL': License(title='''Secure Messaging Protocol Public License''', description='''''', url='https://spdx.org/licenses/SMPPL'),
  'SNIA': License(title='''SNIA Public License 1.1''', description='''''', url='https://spdx.org/licenses/SNIA'),
  'snprintf': License(title='''snprintf License''', description='''''', url='https://spdx.org/licenses/snprintf'),
  'Spencer-86': License(title='''Spencer License 86''', description='''''', url='https://spdx.org/licenses/Spencer-86'),
  'Spencer-94': License(title='''Spencer License 94''', description='''''', url='https://spdx.org/licenses/Spencer-94'),
  'Spencer-99': License(title='''Spencer License 99''', description='''''', url='https://spdx.org/licenses/Spencer-99'),
  'SPL-1.0': License(title='''Sun Public License v1.0''', description='''''', url='https://spdx.org/licenses/SPL-1.0'),
  'SSH-OpenSSH': License(title='''SSH OpenSSH license''', description='''''', url='https://spdx.org/licenses/SSH-OpenSSH'),
  'SSH-short': License(title='''SSH short notice''', description='''''', url='https://spdx.org/licenses/SSH-short'),
  'SSPL-1.0': License(title='''Server Side Public License, v 1''', description='''''', url='https://spdx.org/licenses/SSPL-1.0'),
  'StandardML-NJ': License(title='''Standard ML of New Jersey License''', description='''''', url='https://spdx.org/licenses/StandardML-NJ'),
  'SugarCRM-1.1.3': License(title='''SugarCRM Public License v1.1.3''', description='''''', url='https://spdx.org/licenses/SugarCRM-1.1.3'),
  'SunPro': License(title='''SunPro License''', description='''''', url='https://spdx.org/licenses/SunPro'),
  'SWL': License(title='''Scheme Widget Library (SWL) Software License Agreement''', description='''''', url='https://spdx.org/licenses/SWL'),
  'Symlinks': License(title='''Symlinks License''', description='''''', url='https://spdx.org/licenses/Symlinks'),
  'TAPR-OHL-1.0': License(title='''TAPR Open Hardware License v1.0''', description='''''', url='https://spdx.org/licenses/TAPR-OHL-1.0'),
  'TCL': License(title='''TCL/TK License''', description='''''', url='https://spdx.org/licenses/TCL'),
  'TCP-wrappers': License(title='''TCP Wrappers License''', description='''''', url='https://spdx.org/licenses/TCP-wrappers'),
  'TMate': License(title='''TMate Open Source License''', description='''''', url='https://spdx.org/licenses/TMate'),
  'TORQUE-1.1': License(title='''TORQUE v2.5+ Software License v1.1''', description='''''', url='https://spdx.org/licenses/TORQUE-1.1'),
  'TOSL': License(title='''Trusster Open Source License''', description='''''', url='https://spdx.org/licenses/TOSL'),
  'TPDL': License(title='''Time::ParseDate License''', description='''''', url='https://spdx.org/licenses/TPDL'),
  'TPL-1.0': License(title='''THOR Public License 1.0''', description='''''', url='https://spdx.org/licenses/TPL-1.0'),
  'TTWL': License(title='''Text-Tabs+Wrap License''', description='''''', url='https://spdx.org/licenses/TTWL'),
  'TU-Berlin-1.0': License(title='''Technische Universitaet Berlin License 1.0''', description='''''', url='https://spdx.org/licenses/TU-Berlin-1.0'),
  'TU-Berlin-2.0': License(title='''Technische Universitaet Berlin License 2.0''', description='''''', url='https://spdx.org/licenses/TU-Berlin-2.0'),
  'UCAR': License(title='''UCAR License''', description='''''', url='https://spdx.org/licenses/UCAR'),
  'UCL-1.0': License(title='''Upstream Compatibility License v1.0''', description='''''', url='https://spdx.org/licenses/UCL-1.0'),
  'Unicode-DFS-2015': License(title='''Unicode License Agreement - Data Files and Software (2015)''', description='''''', url='https://spdx.org/licenses/Unicode-DFS-2015'),
  'Unicode-DFS-2016': License(title='''Unicode License Agreement - Data Files and Software (2016)''', description='''''', url='https://spdx.org/licenses/Unicode-DFS-2016'),
  'Unicode-TOU': License(title='''Unicode Terms of Use''', description='''''', url='https://spdx.org/licenses/Unicode-TOU'),
  'Unlicense': License(title='''The Unlicense''', description='''A license with no conditions whatsoever which dedicates works to the public domain. Unlicensed works, modifications, and larger works may be distributed under different terms and without source code.''', url='https://spdx.org/licenses/Unlicense'),
  'UPL-1.0': License(title='''Universal Permissive License v1.0''', description='''A permissive, OSI and FSF approved, GPL compatible license, expressly allowing attribution with just a copyright notice and a short form link rather than the full text of the license.  Includes an express grant of patent rights.  Licensed works and modifications may be distributed under different terms and without source code, and the patent grant may also optionally be expanded to larger works to permit use as a contributor license agreement.''', url='https://spdx.org/licenses/UPL-1.0'),
  'Vim': License(title='''Vim License''', description='''There are no restrictions on using or distributing an unmodified copy of the software. Parts of the software may also be distributed, but the license text must always be included. For modified versions a few restrictions apply. The license is GPL compatible, you may compile the software with GPL libraries and distribute it.''', url='https://spdx.org/licenses/Vim'),
  'VOSTROM': License(title='''VOSTROM Public License for Open Source''', description='''''', url='https://spdx.org/licenses/VOSTROM'),
  'VSL-1.0': License(title='''Vovida Software License v1.0''', description='''''', url='https://spdx.org/licenses/VSL-1.0'),
  'W3C': License(title='''W3C Software Notice and License (2002-12-31)''', description='''''', url='https://spdx.org/licenses/W3C'),
  'W3C-19980720': License(title='''W3C Software Notice and License (1998-07-20)''', description='''''', url='https://spdx.org/licenses/W3C-19980720'),
  'W3C-20150513': License(title='''W3C Software Notice and Document License (2015-05-13)''', description='''''', url='https://spdx.org/licenses/W3C-20150513'),
  'w3m': License(title='''w3m License''', description='''''', url='https://spdx.org/licenses/w3m'),
  'Watcom-1.0': License(title='''Sybase Open Watcom Public License 1.0''', description='''''', url='https://spdx.org/licenses/Watcom-1.0'),
  'Wsuipa': License(title='''Wsuipa License''', description='''''', url='https://spdx.org/licenses/Wsuipa'),
  'WTFPL': License(title='''Do What The F*ck You Want To Public License''', description='''The easiest license out there. It gives the user permissions to do whatever they want with your code.''', url='https://spdx.org/licenses/WTFPL'),
  'wxWindows': License(title='''wxWindows Library License''', description='''''', url='https://spdx.org/licenses/wxWindows'),
  'X11': License(title='''X11 License''', description='''''', url='https://spdx.org/licenses/X11'),
  'X11-distribute-modifications-variant': License(title='''X11 License Distribution Modification Variant''', description='''''', url='https://spdx.org/licenses/X11-distribute-modifications-variant'),
  'Xerox': License(title='''Xerox License''', description='''''', url='https://spdx.org/licenses/Xerox'),
  'XFree86-1.1': License(title='''XFree86 License 1.1''', description='''''', url='https://spdx.org/licenses/XFree86-1.1'),
  'xinetd': License(title='''xinetd License''', description='''''', url='https://spdx.org/licenses/xinetd'),
  'xlock': License(title='''xlock License''', description='''''', url='https://spdx.org/licenses/xlock'),
  'Xnet': License(title='''X.Net License''', description='''''', url='https://spdx.org/licenses/Xnet'),
  'xpp': License(title='''XPP License''', description='''''', url='https://spdx.org/licenses/xpp'),
  'XSkat': License(title='''XSkat License''', description='''''', url='https://spdx.org/licenses/XSkat'),
  'YPL-1.0': License(title='''Yahoo! Public License v1.0''', description='''''', url='https://spdx.org/licenses/YPL-1.0'),
  'YPL-1.1': License(title='''Yahoo! Public License v1.1''', description='''''', url='https://spdx.org/licenses/YPL-1.1'),
  'Zed': License(title='''Zed License''', description='''''', url='https://spdx.org/licenses/Zed'),
  'Zend-2.0': License(title='''Zend License v2.0''', description='''''', url='https://spdx.org/licenses/Zend-2.0'),
  'Zimbra-1.3': License(title='''Zimbra Public License v1.3''', description='''''', url='https://spdx.org/licenses/Zimbra-1.3'),
  'Zimbra-1.4': License(title='''Zimbra Public License v1.4''', description='''''', url='https://spdx.org/licenses/Zimbra-1.4'),
  'Zlib': License(title='''zlib License''', description='''A short permissive license, compatible with GPL. Requires altered source versions to be documented as such.''', url='https://spdx.org/licenses/Zlib'),
  'zlib-acknowledgement': License(title='''zlib/libpng License with Acknowledgement''', description='''''', url='https://spdx.org/licenses/zlib-acknowledgement'),
  'ZPL-1.1': License(title='''Zope Public License 1.1''', description='''''', url='https://spdx.org/licenses/ZPL-1.1'),
  'ZPL-2.0': License(title='''Zope Public License 2.0''', description='''''', url='https://spdx.org/licenses/ZPL-2.0'),
  'ZPL-2.1': License(title='''Zope Public License 2.1''', description='''''', url='https://spdx.org/licenses/ZPL-2.1'),
})



# The following is a hash table of all known common SPDX URLs (including
# variants) mapping to the SPDX license id's. We use this to look up SPDX
# license identifiers given license URLs.

LICENSE_URLS = CaseFoldDict({
  'http://spdx.org/licenses/0BSD': '0BSD',
  'https://spdx.org/licenses/0BSD': '0BSD',
  'http://landley.net/toybox/license.html': '0BSD',
  'https://landley.net/toybox/license.html': '0BSD',
  'http://opensource.org/licenses/0BSD': '0BSD',
  'https://opensource.org/licenses/0BSD': '0BSD',
  'http://spdx.org/licenses/AAL': 'AAL',
  'https://spdx.org/licenses/AAL': 'AAL',
  'http://opensource.org/licenses/attribution': 'AAL',
  'https://opensource.org/licenses/attribution': 'AAL',
  'http://spdx.org/licenses/Abstyles': 'Abstyles',
  'https://spdx.org/licenses/Abstyles': 'Abstyles',
  'http://fedoraproject.org/wiki/Licensing/Abstyles': 'Abstyles',
  'https://fedoraproject.org/wiki/Licensing/Abstyles': 'Abstyles',
  'http://spdx.org/licenses/AdaCore-doc': 'AdaCore-doc',
  'https://spdx.org/licenses/AdaCore-doc': 'AdaCore-doc',
  'http://github.com/AdaCore/xmlada/blob/master/docs/index.rst': 'AdaCore-doc',
  'https://github.com/AdaCore/xmlada/blob/master/docs/index.rst': 'AdaCore-doc',
  'http://github.com/AdaCore/gnatcoll-core/blob/master/docs/index.rst': 'AdaCore-doc',
  'https://github.com/AdaCore/gnatcoll-core/blob/master/docs/index.rst': 'AdaCore-doc',
  'http://github.com/AdaCore/gnatcoll-db/blob/master/docs/index.rst': 'AdaCore-doc',
  'https://github.com/AdaCore/gnatcoll-db/blob/master/docs/index.rst': 'AdaCore-doc',
  'http://spdx.org/licenses/Adobe-2006': 'Adobe-2006',
  'https://spdx.org/licenses/Adobe-2006': 'Adobe-2006',
  'http://fedoraproject.org/wiki/Licensing/AdobeLicense': 'Adobe-2006',
  'https://fedoraproject.org/wiki/Licensing/AdobeLicense': 'Adobe-2006',
  'http://spdx.org/licenses/Adobe-Glyp': 'Adobe-Glyph',
  'https://spdx.org/licenses/Adobe-Glyp': 'Adobe-Glyph',
  'http://fedoraproject.org/wiki/Licensing/MIT#AdobeGlyph': 'Adobe-Glyph',
  'https://fedoraproject.org/wiki/Licensing/MIT#AdobeGlyph': 'Adobe-Glyph',
  'http://spdx.org/licenses/ADSL': 'ADSL',
  'https://spdx.org/licenses/ADSL': 'ADSL',
  'http://fedoraproject.org/wiki/Licensing/AmazonDigitalServicesLicense': 'ADSL',
  'https://fedoraproject.org/wiki/Licensing/AmazonDigitalServicesLicense': 'ADSL',
  'http://spdx.org/licenses/AFL-1.1': 'AFL-1.1',
  'https://spdx.org/licenses/AFL-1.1': 'AFL-1.1',
  'http://opensource.linux-mirror.org/licenses/afl-1.1.txt': 'AFL-1.1',
  'https://opensource.linux-mirror.org/licenses/afl-1.1.txt': 'AFL-1.1',
  'http://wayback.archive.org/web/20021004124254/http://www.opensource.org/licenses/academic.php': 'AFL-1.1',
  'https://wayback.archive.org/web/20021004124254/https://www.opensource.org/licenses/academic.php': 'AFL-1.1',
  'http://spdx.org/licenses/AFL-1.2': 'AFL-1.2',
  'https://spdx.org/licenses/AFL-1.2': 'AFL-1.2',
  'http://opensource.linux-mirror.org/licenses/afl-1.2.txt': 'AFL-1.2',
  'https://opensource.linux-mirror.org/licenses/afl-1.2.txt': 'AFL-1.2',
  'http://wayback.archive.org/web/20021204204652/http://www.opensource.org/licenses/academic.php': 'AFL-1.2',
  'https://wayback.archive.org/web/20021204204652/https://www.opensource.org/licenses/academic.php': 'AFL-1.2',
  'http://spdx.org/licenses/AFL-2.0': 'AFL-2.0',
  'https://spdx.org/licenses/AFL-2.0': 'AFL-2.0',
  'http://wayback.archive.org/web/20060924134533/http://www.opensource.org/licenses/afl-2.0.txt': 'AFL-2.0',
  'https://wayback.archive.org/web/20060924134533/https://www.opensource.org/licenses/afl-2.0.txt': 'AFL-2.0',
  'http://spdx.org/licenses/AFL-2.1': 'AFL-2.1',
  'https://spdx.org/licenses/AFL-2.1': 'AFL-2.1',
  'http://opensource.linux-mirror.org/licenses/afl-2.1.txt': 'AFL-2.1',
  'https://opensource.linux-mirror.org/licenses/afl-2.1.txt': 'AFL-2.1',
  'http://spdx.org/licenses/AFL-3.0': 'AFL-3.0',
  'https://spdx.org/licenses/AFL-3.0': 'AFL-3.0',
  'http://www.rosenlaw.com/AFL3.0.htm': 'AFL-3.0',
  'https://www.rosenlaw.com/AFL3.0.htm': 'AFL-3.0',
  'http://opensource.org/licenses/afl-3.0': 'AFL-3.0',
  'https://opensource.org/licenses/afl-3.0': 'AFL-3.0',
  'http://spdx.org/licenses/Afmparse': 'Afmparse',
  'https://spdx.org/licenses/Afmparse': 'Afmparse',
  'http://fedoraproject.org/wiki/Licensing/Afmparse': 'Afmparse',
  'https://fedoraproject.org/wiki/Licensing/Afmparse': 'Afmparse',
  'http://spdx.org/licenses/AGPL-1.0': 'AGPL-1.0',
  'https://spdx.org/licenses/AGPL-1.0': 'AGPL-1.0',
  'http://www.affero.org/oagpl.html': 'AGPL-1.0-or-later',
  'https://www.affero.org/oagpl.html': 'AGPL-1.0-or-later',
  'http://spdx.org/licenses/AGPL-1.0-only': 'AGPL-1.0-only',
  'https://spdx.org/licenses/AGPL-1.0-only': 'AGPL-1.0-only',
  'http://spdx.org/licenses/AGPL-1.0-or-later': 'AGPL-1.0-or-later',
  'https://spdx.org/licenses/AGPL-1.0-or-later': 'AGPL-1.0-or-later',
  'http://spdx.org/licenses/AGPL-3.0': 'AGPL-3.0',
  'https://spdx.org/licenses/AGPL-3.0': 'AGPL-3.0',
  'http://www.gnu.org/licenses/agpl.txt': 'AGPL-3.0-or-later',
  'https://www.gnu.org/licenses/agpl.txt': 'AGPL-3.0-or-later',
  'http://opensource.org/licenses/AGPL-3.0': 'AGPL-3.0-or-later',
  'https://opensource.org/licenses/AGPL-3.0': 'AGPL-3.0-or-later',
  'http://spdx.org/licenses/AGPL-3.0-only': 'AGPL-3.0-only',
  'https://spdx.org/licenses/AGPL-3.0-only': 'AGPL-3.0-only',
  'http://spdx.org/licenses/AGPL-3.0-or-later': 'AGPL-3.0-or-later',
  'https://spdx.org/licenses/AGPL-3.0-or-later': 'AGPL-3.0-or-later',
  'http://spdx.org/licenses/Aladdin': 'Aladdin',
  'https://spdx.org/licenses/Aladdin': 'Aladdin',
  'http://pages.cs.wisc.edu/~ghost/doc/AFPL/6.01/Public.htm': 'Aladdin',
  'https://pages.cs.wisc.edu/~ghost/doc/AFPL/6.01/Public.htm': 'Aladdin',
  'http://spdx.org/licenses/AMDPLPA': 'AMDPLPA',
  'https://spdx.org/licenses/AMDPLPA': 'AMDPLPA',
  'http://fedoraproject.org/wiki/Licensing/AMD_plpa_map_License': 'AMDPLPA',
  'https://fedoraproject.org/wiki/Licensing/AMD_plpa_map_License': 'AMDPLPA',
  'http://spdx.org/licenses/AML': 'AML',
  'https://spdx.org/licenses/AML': 'AML',
  'http://fedoraproject.org/wiki/Licensing/Apple_MIT_License': 'AML',
  'https://fedoraproject.org/wiki/Licensing/Apple_MIT_License': 'AML',
  'http://spdx.org/licenses/AMPAS': 'AMPAS',
  'https://spdx.org/licenses/AMPAS': 'AMPAS',
  'http://fedoraproject.org/wiki/Licensing/BSD#AMPASBSD': 'AMPAS',
  'https://fedoraproject.org/wiki/Licensing/BSD#AMPASBSD': 'AMPAS',
  'http://spdx.org/licenses/ANTLR-PD': 'ANTLR-PD',
  'https://spdx.org/licenses/ANTLR-PD': 'ANTLR-PD',
  'http://www.antlr2.org/license.html': 'ANTLR-PD-fallback',
  'https://www.antlr2.org/license.html': 'ANTLR-PD-fallback',
  'http://spdx.org/licenses/ANTLR-PD-fallback': 'ANTLR-PD-fallback',
  'https://spdx.org/licenses/ANTLR-PD-fallback': 'ANTLR-PD-fallback',
  'http://spdx.org/licenses/Apache-1.0': 'Apache-1.0',
  'https://spdx.org/licenses/Apache-1.0': 'Apache-1.0',
  'http://www.apache.org/licenses/LICENSE-1.0': 'Apache-1.0',
  'https://www.apache.org/licenses/LICENSE-1.0': 'Apache-1.0',
  'http://spdx.org/licenses/Apache-1.1': 'Apache-1.1',
  'https://spdx.org/licenses/Apache-1.1': 'Apache-1.1',
  'http://apache.org/licenses/LICENSE-1.1': 'Apache-1.1',
  'https://apache.org/licenses/LICENSE-1.1': 'Apache-1.1',
  'http://opensource.org/licenses/Apache-1.1': 'Apache-1.1',
  'https://opensource.org/licenses/Apache-1.1': 'Apache-1.1',
  'http://spdx.org/licenses/Apache-2.0': 'Apache-2.0',
  'https://spdx.org/licenses/Apache-2.0': 'Apache-2.0',
  'http://www.apache.org/licenses/LICENSE-2.0': 'Apache-2.0',
  'https://www.apache.org/licenses/LICENSE-2.0': 'Apache-2.0',
  'http://opensource.org/licenses/Apache-2.0': 'Apache-2.0',
  'https://opensource.org/licenses/Apache-2.0': 'Apache-2.0',
  'http://spdx.org/licenses/APAFML': 'APAFML',
  'https://spdx.org/licenses/APAFML': 'APAFML',
  'http://fedoraproject.org/wiki/Licensing/AdobePostscriptAFM': 'APAFML',
  'https://fedoraproject.org/wiki/Licensing/AdobePostscriptAFM': 'APAFML',
  'http://spdx.org/licenses/APL-1.0': 'APL-1.0',
  'https://spdx.org/licenses/APL-1.0': 'APL-1.0',
  'http://opensource.org/licenses/APL-1.0': 'APL-1.0',
  'https://opensource.org/licenses/APL-1.0': 'APL-1.0',
  'http://spdx.org/licenses/App-s2p': 'App-s2p',
  'https://spdx.org/licenses/App-s2p': 'App-s2p',
  'http://fedoraproject.org/wiki/Licensing/App-s2p': 'App-s2p',
  'https://fedoraproject.org/wiki/Licensing/App-s2p': 'App-s2p',
  'http://spdx.org/licenses/APSL-1.0': 'APSL-1.0',
  'https://spdx.org/licenses/APSL-1.0': 'APSL-1.0',
  'http://fedoraproject.org/wiki/Licensing/Apple_Public_Source_License_1.0': 'APSL-1.0',
  'https://fedoraproject.org/wiki/Licensing/Apple_Public_Source_License_1.0': 'APSL-1.0',
  'http://spdx.org/licenses/APSL-1.1': 'APSL-1.1',
  'https://spdx.org/licenses/APSL-1.1': 'APSL-1.1',
  'http://www.opensource.apple.com/source/IOSerialFamily/IOSerialFamily-7/APPLE_LICENSE': 'APSL-1.1',
  'https://www.opensource.apple.com/source/IOSerialFamily/IOSerialFamily-7/APPLE_LICENSE': 'APSL-1.1',
  'http://spdx.org/licenses/APSL-1.2': 'APSL-1.2',
  'https://spdx.org/licenses/APSL-1.2': 'APSL-1.2',
  'http://www.samurajdata.se/opensource/mirror/licenses/apsl.php': 'APSL-1.2',
  'https://www.samurajdata.se/opensource/mirror/licenses/apsl.php': 'APSL-1.2',
  'http://spdx.org/licenses/APSL-2.0': 'APSL-2.0',
  'https://spdx.org/licenses/APSL-2.0': 'APSL-2.0',
  'http://www.opensource.apple.com/license/apsl/': 'APSL-2.0',
  'https://www.opensource.apple.com/license/apsl/': 'APSL-2.0',
  'http://spdx.org/licenses/Arphic-1999': 'Arphic-1999',
  'https://spdx.org/licenses/Arphic-1999': 'Arphic-1999',
  'http://ftp.gnu.org/gnu/non-gnu/chinese-fonts-truetype/LICENSE': 'Arphic-1999',
  'https://ftp.gnu.org/gnu/non-gnu/chinese-fonts-truetype/LICENSE': 'Arphic-1999',
  'http://spdx.org/licenses/Artistic-1.0': 'Artistic-1.0',
  'https://spdx.org/licenses/Artistic-1.0': 'Artistic-1.0',
  'http://opensource.org/licenses/Artistic-1.0': 'Artistic-1.0-cl8',
  'https://opensource.org/licenses/Artistic-1.0': 'Artistic-1.0-cl8',
  'http://spdx.org/licenses/Artistic-1.0-cl8': 'Artistic-1.0-cl8',
  'https://spdx.org/licenses/Artistic-1.0-cl8': 'Artistic-1.0-cl8',
  'http://spdx.org/licenses/Artistic-1.0-Per': 'Artistic-1.0-Perl',
  'https://spdx.org/licenses/Artistic-1.0-Per': 'Artistic-1.0-Perl',
  'http://dev.perl.org/licenses/artistic.html': 'Artistic-1.0-Perl',
  'https://dev.perl.org/licenses/artistic.html': 'Artistic-1.0-Perl',
  'http://spdx.org/licenses/Artistic-2.0': 'Artistic-2.0',
  'https://spdx.org/licenses/Artistic-2.0': 'Artistic-2.0',
  'http://www.perlfoundation.org/artistic_license_2_0': 'Artistic-2.0',
  'https://www.perlfoundation.org/artistic_license_2_0': 'Artistic-2.0',
  'http://www.perlfoundation.org/artistic-license-20.html': 'Artistic-2.0',
  'https://www.perlfoundation.org/artistic-license-20.html': 'Artistic-2.0',
  'http://opensource.org/licenses/artistic-license-2.0': 'Artistic-2.0',
  'https://opensource.org/licenses/artistic-license-2.0': 'Artistic-2.0',
  'http://spdx.org/licenses/Baekmuk': 'Baekmuk',
  'https://spdx.org/licenses/Baekmuk': 'Baekmuk',
  'http://fedoraproject.org/wiki/Licensing:Baekmuk?rd=Licensing/Baekmuk': 'Baekmuk',
  'https://fedoraproject.org/wiki/Licensing:Baekmuk?rd=Licensing/Baekmuk': 'Baekmuk',
  'http://spdx.org/licenses/Bahyp': 'Bahyph',
  'https://spdx.org/licenses/Bahyp': 'Bahyph',
  'http://fedoraproject.org/wiki/Licensing/Bahyph': 'Bahyph',
  'https://fedoraproject.org/wiki/Licensing/Bahyph': 'Bahyph',
  'http://spdx.org/licenses/Barr': 'Barr',
  'https://spdx.org/licenses/Barr': 'Barr',
  'http://fedoraproject.org/wiki/Licensing/Barr': 'Barr',
  'https://fedoraproject.org/wiki/Licensing/Barr': 'Barr',
  'http://spdx.org/licenses/Beerware': 'Beerware',
  'https://spdx.org/licenses/Beerware': 'Beerware',
  'http://fedoraproject.org/wiki/Licensing/Beerware': 'Beerware',
  'https://fedoraproject.org/wiki/Licensing/Beerware': 'Beerware',
  'http://people.freebsd.org/~phk/': 'Beerware',
  'https://people.freebsd.org/~phk/': 'Beerware',
  'http://spdx.org/licenses/Bitstream-Charter': 'Bitstream-Charter',
  'https://spdx.org/licenses/Bitstream-Charter': 'Bitstream-Charter',
  'http://fedoraproject.org/wiki/Licensing/Charter#License_Text': 'Bitstream-Charter',
  'https://fedoraproject.org/wiki/Licensing/Charter#License_Text': 'Bitstream-Charter',
  'http://raw.githubusercontent.com/blackhole89/notekit/master/data/fonts/Charter%20license.txt': 'Bitstream-Charter',
  'https://raw.githubusercontent.com/blackhole89/notekit/master/data/fonts/Charter%20license.txt': 'Bitstream-Charter',
  'http://spdx.org/licenses/Bitstream-Vera': 'Bitstream-Vera',
  'https://spdx.org/licenses/Bitstream-Vera': 'Bitstream-Vera',
  'http://web.archive.org/web/20080207013128/http://www.gnome.org/fonts/': 'Bitstream-Vera',
  'https://web.archive.org/web/20080207013128/https://www.gnome.org/fonts/': 'Bitstream-Vera',
  'http://docubrain.com/sites/default/files/licenses/bitstream-vera.html': 'Bitstream-Vera',
  'https://docubrain.com/sites/default/files/licenses/bitstream-vera.html': 'Bitstream-Vera',
  'http://spdx.org/licenses/BitTorrent-1.0': 'BitTorrent-1.0',
  'https://spdx.org/licenses/BitTorrent-1.0': 'BitTorrent-1.0',
  'http://sources.gentoo.org/cgi-bin/viewvc.cgi/gentoo-x86/licenses/BitTorrent?r1=1.1&r2=1.1.1.1&diff_format=s': 'BitTorrent-1.0',
  'https://sources.gentoo.org/cgi-bin/viewvc.cgi/gentoo-x86/licenses/BitTorrent?r1=1.1&r2=1.1.1.1&diff_format=s': 'BitTorrent-1.0',
  'http://spdx.org/licenses/BitTorrent-1.1': 'BitTorrent-1.1',
  'https://spdx.org/licenses/BitTorrent-1.1': 'BitTorrent-1.1',
  'http://directory.fsf.org/wiki/License:BitTorrentOSL1.1': 'BitTorrent-1.1',
  'https://directory.fsf.org/wiki/License:BitTorrentOSL1.1': 'BitTorrent-1.1',
  'http://spdx.org/licenses/blessing': 'blessing',
  'https://spdx.org/licenses/blessing': 'blessing',
  'http://www.sqlite.org/src/artifact/e33a4df7e32d742a?ln=4-9': 'blessing',
  'https://www.sqlite.org/src/artifact/e33a4df7e32d742a?ln=4-9': 'blessing',
  'http://sqlite.org/src/artifact/df5091916dbb40e6': 'blessing',
  'https://sqlite.org/src/artifact/df5091916dbb40e6': 'blessing',
  'http://spdx.org/licenses/BlueOak-1.0.0': 'BlueOak-1.0.0',
  'https://spdx.org/licenses/BlueOak-1.0.0': 'BlueOak-1.0.0',
  'http://blueoakcouncil.org/license/1.0.0': 'BlueOak-1.0.0',
  'https://blueoakcouncil.org/license/1.0.0': 'BlueOak-1.0.0',
  'http://spdx.org/licenses/Borceux': 'Borceux',
  'https://spdx.org/licenses/Borceux': 'Borceux',
  'http://fedoraproject.org/wiki/Licensing/Borceux': 'Borceux',
  'https://fedoraproject.org/wiki/Licensing/Borceux': 'Borceux',
  'http://spdx.org/licenses/Brian-Gladman-3-Clause': 'Brian-Gladman-3-Clause',
  'https://spdx.org/licenses/Brian-Gladman-3-Clause': 'Brian-Gladman-3-Clause',
  'http://github.com/SWI-Prolog/packages-clib/blob/master/sha1/brg_endian.h': 'Brian-Gladman-3-Clause',
  'https://github.com/SWI-Prolog/packages-clib/blob/master/sha1/brg_endian.h': 'Brian-Gladman-3-Clause',
  'http://spdx.org/licenses/BSD-1-Clause': 'BSD-1-Clause',
  'https://spdx.org/licenses/BSD-1-Clause': 'BSD-1-Clause',
  'http://svnweb.freebsd.org/base/head/include/ifaddrs.h?revision=326823': 'BSD-1-Clause',
  'https://svnweb.freebsd.org/base/head/include/ifaddrs.h?revision=326823': 'BSD-1-Clause',
  'http://spdx.org/licenses/BSD-2-Clause': 'BSD-2-Clause',
  'https://spdx.org/licenses/BSD-2-Clause': 'BSD-2-Clause',
  'http://opensource.org/licenses/BSD-2-Clause': 'BSD-2-Clause',
  'https://opensource.org/licenses/BSD-2-Clause': 'BSD-2-Clause',
  'http://spdx.org/licenses/BSD-2-Clause-FreeBSD': 'BSD-2-Clause-FreeBSD',
  'https://spdx.org/licenses/BSD-2-Clause-FreeBSD': 'BSD-2-Clause-FreeBSD',
  'http://www.freebsd.org/copyright/freebsd-license.html': 'BSD-2-Clause-Views',
  'https://www.freebsd.org/copyright/freebsd-license.html': 'BSD-2-Clause-Views',
  'http://spdx.org/licenses/BSD-2-Clause-NetBSD': 'BSD-2-Clause-NetBSD',
  'https://spdx.org/licenses/BSD-2-Clause-NetBSD': 'BSD-2-Clause-NetBSD',
  'http://www.netbsd.org/about/redistribution.html#default': 'BSD-2-Clause-NetBSD',
  'https://www.netbsd.org/about/redistribution.html#default': 'BSD-2-Clause-NetBSD',
  'http://spdx.org/licenses/BSD-2-Clause-Paten': 'BSD-2-Clause-Patent',
  'https://spdx.org/licenses/BSD-2-Clause-Paten': 'BSD-2-Clause-Patent',
  'http://opensource.org/licenses/BSDplusPatent': 'BSD-2-Clause-Patent',
  'https://opensource.org/licenses/BSDplusPatent': 'BSD-2-Clause-Patent',
  'http://spdx.org/licenses/BSD-2-Clause-Views': 'BSD-2-Clause-Views',
  'https://spdx.org/licenses/BSD-2-Clause-Views': 'BSD-2-Clause-Views',
  'http://people.freebsd.org/~ivoras/wine/patch-wine-nvidia.sh': 'BSD-2-Clause-Views',
  'https://people.freebsd.org/~ivoras/wine/patch-wine-nvidia.sh': 'BSD-2-Clause-Views',
  'http://github.com/protegeproject/protege/blob/master/license.txt': 'BSD-2-Clause-Views',
  'https://github.com/protegeproject/protege/blob/master/license.txt': 'BSD-2-Clause-Views',
  'http://spdx.org/licenses/BSD-3-Clause': 'BSD-3-Clause',
  'https://spdx.org/licenses/BSD-3-Clause': 'BSD-3-Clause',
  'http://opensource.org/licenses/BSD-3-Clause': 'BSD-3-Clause',
  'https://opensource.org/licenses/BSD-3-Clause': 'BSD-3-Clause',
  'http://www.eclipse.org/org/documents/edl-v10.php': 'BSD-3-Clause',
  'https://www.eclipse.org/org/documents/edl-v10.php': 'BSD-3-Clause',
  'http://spdx.org/licenses/BSD-3-Clause-Attribution': 'BSD-3-Clause-Attribution',
  'https://spdx.org/licenses/BSD-3-Clause-Attribution': 'BSD-3-Clause-Attribution',
  'http://fedoraproject.org/wiki/Licensing/BSD_with_Attribution': 'BSD-3-Clause-Attribution',
  'https://fedoraproject.org/wiki/Licensing/BSD_with_Attribution': 'BSD-3-Clause-Attribution',
  'http://spdx.org/licenses/BSD-3-Clause-Clear': 'BSD-3-Clause-Clear',
  'https://spdx.org/licenses/BSD-3-Clause-Clear': 'BSD-3-Clause-Clear',
  'http://labs.metacarta.com/license-explanation.html#license': 'BSD-3-Clause-Clear',
  'https://labs.metacarta.com/license-explanation.html#license': 'BSD-3-Clause-Clear',
  'http://spdx.org/licenses/BSD-3-Clause-LBNL': 'BSD-3-Clause-LBNL',
  'https://spdx.org/licenses/BSD-3-Clause-LBNL': 'BSD-3-Clause-LBNL',
  'http://fedoraproject.org/wiki/Licensing/LBNLBSD': 'BSD-3-Clause-LBNL',
  'https://fedoraproject.org/wiki/Licensing/LBNLBSD': 'BSD-3-Clause-LBNL',
  'http://spdx.org/licenses/BSD-3-Clause-Modification': 'BSD-3-Clause-Modification',
  'https://spdx.org/licenses/BSD-3-Clause-Modification': 'BSD-3-Clause-Modification',
  'http://fedoraproject.org/wiki/Licensing:BSD#Modification_Variant': 'BSD-3-Clause-Modification',
  'https://fedoraproject.org/wiki/Licensing:BSD#Modification_Variant': 'BSD-3-Clause-Modification',
  'http://spdx.org/licenses/BSD-3-Clause-No-Military-License': 'BSD-3-Clause-No-Military-License',
  'https://spdx.org/licenses/BSD-3-Clause-No-Military-License': 'BSD-3-Clause-No-Military-License',
  'http://gitlab.syncad.com/hive/dhive/-/blob/master/LICENSE': 'BSD-3-Clause-No-Military-License',
  'https://gitlab.syncad.com/hive/dhive/-/blob/master/LICENSE': 'BSD-3-Clause-No-Military-License',
  'http://github.com/greymass/swift-eosio/blob/master/LICENSE': 'BSD-3-Clause-No-Military-License',
  'https://github.com/greymass/swift-eosio/blob/master/LICENSE': 'BSD-3-Clause-No-Military-License',
  'http://spdx.org/licenses/BSD-3-Clause-No-Nuclear-License': 'BSD-3-Clause-No-Nuclear-License',
  'https://spdx.org/licenses/BSD-3-Clause-No-Nuclear-License': 'BSD-3-Clause-No-Nuclear-License',
  'http://download.oracle.com/otn-pub/java/licenses/bsd.txt?AuthParam=1467140197_43d516ce1776bd08a58235a7785be1cc': 'BSD-3-Clause-No-Nuclear-License',
  'https://download.oracle.com/otn-pub/java/licenses/bsd.txt?AuthParam=1467140197_43d516ce1776bd08a58235a7785be1cc': 'BSD-3-Clause-No-Nuclear-License',
  'http://spdx.org/licenses/BSD-3-Clause-No-Nuclear-License-2014': 'BSD-3-Clause-No-Nuclear-License-2014',
  'https://spdx.org/licenses/BSD-3-Clause-No-Nuclear-License-2014': 'BSD-3-Clause-No-Nuclear-License-2014',
  'http://java.net/projects/javaeetutorial/pages/BerkeleyLicense': 'BSD-3-Clause-No-Nuclear-License-2014',
  'https://java.net/projects/javaeetutorial/pages/BerkeleyLicense': 'BSD-3-Clause-No-Nuclear-License-2014',
  'http://spdx.org/licenses/BSD-3-Clause-No-Nuclear-Warranty': 'BSD-3-Clause-No-Nuclear-Warranty',
  'https://spdx.org/licenses/BSD-3-Clause-No-Nuclear-Warranty': 'BSD-3-Clause-No-Nuclear-Warranty',
  'http://jogamp.org/git/?p=gluegen.git;a=blob_plain;f=LICENSE.txt': 'BSD-3-Clause-No-Nuclear-Warranty',
  'https://jogamp.org/git/?p=gluegen.git;a=blob_plain;f=LICENSE.txt': 'BSD-3-Clause-No-Nuclear-Warranty',
  'http://spdx.org/licenses/BSD-3-Clause-Open-MPI': 'BSD-3-Clause-Open-MPI',
  'https://spdx.org/licenses/BSD-3-Clause-Open-MPI': 'BSD-3-Clause-Open-MPI',
  'http://www.open-mpi.org/community/license.php': 'BSD-3-Clause-Open-MPI',
  'https://www.open-mpi.org/community/license.php': 'BSD-3-Clause-Open-MPI',
  'http://www.netlib.org/lapack/LICENSE.txt': 'BSD-3-Clause-Open-MPI',
  'https://www.netlib.org/lapack/LICENSE.txt': 'BSD-3-Clause-Open-MPI',
  'http://spdx.org/licenses/BSD-4-Clause': 'BSD-4-Clause',
  'https://spdx.org/licenses/BSD-4-Clause': 'BSD-4-Clause',
  'http://directory.fsf.org/wiki/License:BSD_4Clause': 'BSD-4-Clause',
  'https://directory.fsf.org/wiki/License:BSD_4Clause': 'BSD-4-Clause',
  'http://spdx.org/licenses/BSD-4-Clause-Shortened': 'BSD-4-Clause-Shortened',
  'https://spdx.org/licenses/BSD-4-Clause-Shortened': 'BSD-4-Clause-Shortened',
  'http://metadata.ftp-master.debian.org/changelogs//main/a/arpwatch/arpwatch_2.1a15-7_copyright': 'BSD-4-Clause-Shortened',
  'https://metadata.ftp-master.debian.org/changelogs//main/a/arpwatch/arpwatch_2.1a15-7_copyright': 'BSD-4-Clause-Shortened',
  'http://spdx.org/licenses/BSD-4-Clause-UC': 'BSD-4-Clause-UC',
  'https://spdx.org/licenses/BSD-4-Clause-UC': 'BSD-4-Clause-UC',
  'http://www.freebsd.org/copyright/license.html': 'BSD-4-Clause-UC',
  'https://www.freebsd.org/copyright/license.html': 'BSD-4-Clause-UC',
  'http://spdx.org/licenses/BSD-4.3RENO': 'BSD-4.3RENO',
  'https://spdx.org/licenses/BSD-4.3RENO': 'BSD-4.3RENO',
  'http://sourceware.org/git/?p=binutils-gdb.git;a=blob;f=libiberty/strcasecmp.c;h=131d81c2ce7881fa48c363dc5bf5fb302c61ce0b;hb=HEAD': 'BSD-4.3RENO',
  'https://sourceware.org/git/?p=binutils-gdb.git;a=blob;f=libiberty/strcasecmp.c;h=131d81c2ce7881fa48c363dc5bf5fb302c61ce0b;hb=HEAD': 'BSD-4.3RENO',
  'http://spdx.org/licenses/BSD-4.3TAHOE': 'BSD-4.3TAHOE',
  'https://spdx.org/licenses/BSD-4.3TAHOE': 'BSD-4.3TAHOE',
  'http://github.com/389ds/389-ds-base/blob/main/ldap/include/sysexits-compat.h#L15': 'BSD-4.3TAHOE',
  'https://github.com/389ds/389-ds-base/blob/main/ldap/include/sysexits-compat.h#L15': 'BSD-4.3TAHOE',
  'http://spdx.org/licenses/BSD-Advertising-Acknowledgemen': 'BSD-Advertising-Acknowledgement',
  'https://spdx.org/licenses/BSD-Advertising-Acknowledgemen': 'BSD-Advertising-Acknowledgement',
  'http://github.com/python-excel/xlrd/blob/master/LICENSE#L33': 'BSD-Advertising-Acknowledgement',
  'https://github.com/python-excel/xlrd/blob/master/LICENSE#L33': 'BSD-Advertising-Acknowledgement',
  'http://spdx.org/licenses/BSD-Attribution-HPND-disclaimer': 'BSD-Attribution-HPND-disclaimer',
  'https://spdx.org/licenses/BSD-Attribution-HPND-disclaimer': 'BSD-Attribution-HPND-disclaimer',
  'http://github.com/cyrusimap/cyrus-sasl/blob/master/COPYING': 'BSD-Attribution-HPND-disclaimer',
  'https://github.com/cyrusimap/cyrus-sasl/blob/master/COPYING': 'BSD-Attribution-HPND-disclaimer',
  'http://spdx.org/licenses/BSD-Protection': 'BSD-Protection',
  'https://spdx.org/licenses/BSD-Protection': 'BSD-Protection',
  'http://fedoraproject.org/wiki/Licensing/BSD_Protection_License': 'BSD-Protection',
  'https://fedoraproject.org/wiki/Licensing/BSD_Protection_License': 'BSD-Protection',
  'http://spdx.org/licenses/BSD-Source-Code': 'BSD-Source-Code',
  'https://spdx.org/licenses/BSD-Source-Code': 'BSD-Source-Code',
  'http://github.com/robbiehanson/CocoaHTTPServer/blob/master/LICENSE.txt': 'BSD-Source-Code',
  'https://github.com/robbiehanson/CocoaHTTPServer/blob/master/LICENSE.txt': 'BSD-Source-Code',
  'http://spdx.org/licenses/BSL-1.0': 'BSL-1.0',
  'https://spdx.org/licenses/BSL-1.0': 'BSL-1.0',
  'http://www.boost.org/LICENSE_1_0.txt': 'BSL-1.0',
  'https://www.boost.org/LICENSE_1_0.txt': 'BSL-1.0',
  'http://opensource.org/licenses/BSL-1.0': 'BSL-1.0',
  'https://opensource.org/licenses/BSL-1.0': 'BSL-1.0',
  'http://spdx.org/licenses/BUSL-1.1': 'BUSL-1.1',
  'https://spdx.org/licenses/BUSL-1.1': 'BUSL-1.1',
  'http://mariadb.com/bsl11/': 'BUSL-1.1',
  'https://mariadb.com/bsl11/': 'BUSL-1.1',
  'http://spdx.org/licenses/bzip2-1.0.5': 'bzip2-1.0.5',
  'https://spdx.org/licenses/bzip2-1.0.5': 'bzip2-1.0.5',
  'http://sourceware.org/bzip2/1.0.5/bzip2-manual-1.0.5.html': 'bzip2-1.0.5',
  'https://sourceware.org/bzip2/1.0.5/bzip2-manual-1.0.5.html': 'bzip2-1.0.5',
  'http://bzip.org/1.0.5/bzip2-manual-1.0.5.html': 'bzip2-1.0.6',
  'https://bzip.org/1.0.5/bzip2-manual-1.0.5.html': 'bzip2-1.0.6',
  'http://spdx.org/licenses/bzip2-1.0.6': 'bzip2-1.0.6',
  'https://spdx.org/licenses/bzip2-1.0.6': 'bzip2-1.0.6',
  'http://sourceware.org/git/?p=bzip2.git;a=blob;f=LICENSE;hb=bzip2-1.0.6': 'bzip2-1.0.6',
  'https://sourceware.org/git/?p=bzip2.git;a=blob;f=LICENSE;hb=bzip2-1.0.6': 'bzip2-1.0.6',
  'http://spdx.org/licenses/C-UDA-1.0': 'C-UDA-1.0',
  'https://spdx.org/licenses/C-UDA-1.0': 'C-UDA-1.0',
  'http://github.com/microsoft/Computational-Use-of-Data-Agreement/blob/master/C-UDA-1.0.md': 'C-UDA-1.0',
  'https://github.com/microsoft/Computational-Use-of-Data-Agreement/blob/master/C-UDA-1.0.md': 'C-UDA-1.0',
  'http://cdla.dev/computational-use-of-data-agreement-v1-0/': 'C-UDA-1.0',
  'https://cdla.dev/computational-use-of-data-agreement-v1-0/': 'C-UDA-1.0',
  'http://spdx.org/licenses/CAL-1.0': 'CAL-1.0',
  'https://spdx.org/licenses/CAL-1.0': 'CAL-1.0',
  'http://cryptographicautonomylicense.com/license-text.html': 'CAL-1.0-Combined-Work-Exception',
  'https://cryptographicautonomylicense.com/license-text.html': 'CAL-1.0-Combined-Work-Exception',
  'http://opensource.org/licenses/CAL-1.0': 'CAL-1.0-Combined-Work-Exception',
  'https://opensource.org/licenses/CAL-1.0': 'CAL-1.0-Combined-Work-Exception',
  'http://spdx.org/licenses/CAL-1.0-Combined-Work-Exception': 'CAL-1.0-Combined-Work-Exception',
  'https://spdx.org/licenses/CAL-1.0-Combined-Work-Exception': 'CAL-1.0-Combined-Work-Exception',
  'http://spdx.org/licenses/Caldera': 'Caldera',
  'https://spdx.org/licenses/Caldera': 'Caldera',
  'http://www.lemis.com/grog/UNIX/ancient-source-all.pdf': 'Caldera',
  'https://www.lemis.com/grog/UNIX/ancient-source-all.pdf': 'Caldera',
  'http://spdx.org/licenses/CATOSL-1.1': 'CATOSL-1.1',
  'https://spdx.org/licenses/CATOSL-1.1': 'CATOSL-1.1',
  'http://opensource.org/licenses/CATOSL-1.1': 'CATOSL-1.1',
  'https://opensource.org/licenses/CATOSL-1.1': 'CATOSL-1.1',
  'http://spdx.org/licenses/CC-BY-1.0': 'CC-BY-1.0',
  'https://spdx.org/licenses/CC-BY-1.0': 'CC-BY-1.0',
  'http://creativecommons.org/licenses/by/1.0/legalcode': 'CC-BY-1.0',
  'https://creativecommons.org/licenses/by/1.0/legalcode': 'CC-BY-1.0',
  'http://spdx.org/licenses/CC-BY-2.0': 'CC-BY-2.0',
  'https://spdx.org/licenses/CC-BY-2.0': 'CC-BY-2.0',
  'http://creativecommons.org/licenses/by/2.0/legalcode': 'CC-BY-2.0',
  'https://creativecommons.org/licenses/by/2.0/legalcode': 'CC-BY-2.0',
  'http://spdx.org/licenses/CC-BY-2.5': 'CC-BY-2.5',
  'https://spdx.org/licenses/CC-BY-2.5': 'CC-BY-2.5',
  'http://creativecommons.org/licenses/by/2.5/legalcode': 'CC-BY-2.5',
  'https://creativecommons.org/licenses/by/2.5/legalcode': 'CC-BY-2.5',
  'http://spdx.org/licenses/CC-BY-2.5-AU': 'CC-BY-2.5-AU',
  'https://spdx.org/licenses/CC-BY-2.5-AU': 'CC-BY-2.5-AU',
  'http://creativecommons.org/licenses/by/2.5/au/legalcode': 'CC-BY-2.5-AU',
  'https://creativecommons.org/licenses/by/2.5/au/legalcode': 'CC-BY-2.5-AU',
  'http://spdx.org/licenses/CC-BY-3.0': 'CC-BY-3.0',
  'https://spdx.org/licenses/CC-BY-3.0': 'CC-BY-3.0',
  'http://creativecommons.org/licenses/by/3.0/legalcode': 'CC-BY-3.0',
  'https://creativecommons.org/licenses/by/3.0/legalcode': 'CC-BY-3.0',
  'http://spdx.org/licenses/CC-BY-3.0-AT': 'CC-BY-3.0-AT',
  'https://spdx.org/licenses/CC-BY-3.0-AT': 'CC-BY-3.0-AT',
  'http://creativecommons.org/licenses/by/3.0/at/legalcode': 'CC-BY-3.0-AT',
  'https://creativecommons.org/licenses/by/3.0/at/legalcode': 'CC-BY-3.0-AT',
  'http://spdx.org/licenses/CC-BY-3.0-DE': 'CC-BY-3.0-DE',
  'https://spdx.org/licenses/CC-BY-3.0-DE': 'CC-BY-3.0-DE',
  'http://creativecommons.org/licenses/by/3.0/de/legalcode': 'CC-BY-3.0-DE',
  'https://creativecommons.org/licenses/by/3.0/de/legalcode': 'CC-BY-3.0-DE',
  'http://spdx.org/licenses/CC-BY-3.0-IGO': 'CC-BY-3.0-IGO',
  'https://spdx.org/licenses/CC-BY-3.0-IGO': 'CC-BY-3.0-IGO',
  'http://creativecommons.org/licenses/by/3.0/igo/legalcode': 'CC-BY-3.0-IGO',
  'https://creativecommons.org/licenses/by/3.0/igo/legalcode': 'CC-BY-3.0-IGO',
  'http://spdx.org/licenses/CC-BY-3.0-NL': 'CC-BY-3.0-NL',
  'https://spdx.org/licenses/CC-BY-3.0-NL': 'CC-BY-3.0-NL',
  'http://creativecommons.org/licenses/by/3.0/nl/legalcode': 'CC-BY-3.0-NL',
  'https://creativecommons.org/licenses/by/3.0/nl/legalcode': 'CC-BY-3.0-NL',
  'http://spdx.org/licenses/CC-BY-3.0-US': 'CC-BY-3.0-US',
  'https://spdx.org/licenses/CC-BY-3.0-US': 'CC-BY-3.0-US',
  'http://creativecommons.org/licenses/by/3.0/us/legalcode': 'CC-BY-3.0-US',
  'https://creativecommons.org/licenses/by/3.0/us/legalcode': 'CC-BY-3.0-US',
  'http://spdx.org/licenses/CC-BY-4.0': 'CC-BY-4.0',
  'https://spdx.org/licenses/CC-BY-4.0': 'CC-BY-4.0',
  'http://creativecommons.org/licenses/by/4.0/legalcode': 'CC-BY-4.0',
  'https://creativecommons.org/licenses/by/4.0/legalcode': 'CC-BY-4.0',
  'http://spdx.org/licenses/CC-BY-NC-1.0': 'CC-BY-NC-1.0',
  'https://spdx.org/licenses/CC-BY-NC-1.0': 'CC-BY-NC-1.0',
  'http://creativecommons.org/licenses/by-nc/1.0/legalcode': 'CC-BY-NC-1.0',
  'https://creativecommons.org/licenses/by-nc/1.0/legalcode': 'CC-BY-NC-1.0',
  'http://spdx.org/licenses/CC-BY-NC-2.0': 'CC-BY-NC-2.0',
  'https://spdx.org/licenses/CC-BY-NC-2.0': 'CC-BY-NC-2.0',
  'http://creativecommons.org/licenses/by-nc/2.0/legalcode': 'CC-BY-NC-2.0',
  'https://creativecommons.org/licenses/by-nc/2.0/legalcode': 'CC-BY-NC-2.0',
  'http://spdx.org/licenses/CC-BY-NC-2.5': 'CC-BY-NC-2.5',
  'https://spdx.org/licenses/CC-BY-NC-2.5': 'CC-BY-NC-2.5',
  'http://creativecommons.org/licenses/by-nc/2.5/legalcode': 'CC-BY-NC-2.5',
  'https://creativecommons.org/licenses/by-nc/2.5/legalcode': 'CC-BY-NC-2.5',
  'http://spdx.org/licenses/CC-BY-NC-3.0': 'CC-BY-NC-3.0',
  'https://spdx.org/licenses/CC-BY-NC-3.0': 'CC-BY-NC-3.0',
  'http://creativecommons.org/licenses/by-nc/3.0/legalcode': 'CC-BY-NC-3.0',
  'https://creativecommons.org/licenses/by-nc/3.0/legalcode': 'CC-BY-NC-3.0',
  'http://spdx.org/licenses/CC-BY-NC-3.0-DE': 'CC-BY-NC-3.0-DE',
  'https://spdx.org/licenses/CC-BY-NC-3.0-DE': 'CC-BY-NC-3.0-DE',
  'http://creativecommons.org/licenses/by-nc/3.0/de/legalcode': 'CC-BY-NC-3.0-DE',
  'https://creativecommons.org/licenses/by-nc/3.0/de/legalcode': 'CC-BY-NC-3.0-DE',
  'http://spdx.org/licenses/CC-BY-NC-4.0': 'CC-BY-NC-4.0',
  'https://spdx.org/licenses/CC-BY-NC-4.0': 'CC-BY-NC-4.0',
  'http://creativecommons.org/licenses/by-nc/4.0/legalcode': 'CC-BY-NC-4.0',
  'https://creativecommons.org/licenses/by-nc/4.0/legalcode': 'CC-BY-NC-4.0',
  'http://spdx.org/licenses/CC-BY-NC-ND-1.0': 'CC-BY-NC-ND-1.0',
  'https://spdx.org/licenses/CC-BY-NC-ND-1.0': 'CC-BY-NC-ND-1.0',
  'http://creativecommons.org/licenses/by-nd-nc/1.0/legalcode': 'CC-BY-NC-ND-1.0',
  'https://creativecommons.org/licenses/by-nd-nc/1.0/legalcode': 'CC-BY-NC-ND-1.0',
  'http://spdx.org/licenses/CC-BY-NC-ND-2.0': 'CC-BY-NC-ND-2.0',
  'https://spdx.org/licenses/CC-BY-NC-ND-2.0': 'CC-BY-NC-ND-2.0',
  'http://creativecommons.org/licenses/by-nc-nd/2.0/legalcode': 'CC-BY-NC-ND-2.0',
  'https://creativecommons.org/licenses/by-nc-nd/2.0/legalcode': 'CC-BY-NC-ND-2.0',
  'http://spdx.org/licenses/CC-BY-NC-ND-2.5': 'CC-BY-NC-ND-2.5',
  'https://spdx.org/licenses/CC-BY-NC-ND-2.5': 'CC-BY-NC-ND-2.5',
  'http://creativecommons.org/licenses/by-nc-nd/2.5/legalcode': 'CC-BY-NC-ND-2.5',
  'https://creativecommons.org/licenses/by-nc-nd/2.5/legalcode': 'CC-BY-NC-ND-2.5',
  'http://spdx.org/licenses/CC-BY-NC-ND-3.0': 'CC-BY-NC-ND-3.0',
  'https://spdx.org/licenses/CC-BY-NC-ND-3.0': 'CC-BY-NC-ND-3.0',
  'http://creativecommons.org/licenses/by-nc-nd/3.0/legalcode': 'CC-BY-NC-ND-3.0',
  'https://creativecommons.org/licenses/by-nc-nd/3.0/legalcode': 'CC-BY-NC-ND-3.0',
  'http://spdx.org/licenses/CC-BY-NC-ND-3.0-DE': 'CC-BY-NC-ND-3.0-DE',
  'https://spdx.org/licenses/CC-BY-NC-ND-3.0-DE': 'CC-BY-NC-ND-3.0-DE',
  'http://creativecommons.org/licenses/by-nc-nd/3.0/de/legalcode': 'CC-BY-NC-ND-3.0-DE',
  'https://creativecommons.org/licenses/by-nc-nd/3.0/de/legalcode': 'CC-BY-NC-ND-3.0-DE',
  'http://spdx.org/licenses/CC-BY-NC-ND-3.0-IGO': 'CC-BY-NC-ND-3.0-IGO',
  'https://spdx.org/licenses/CC-BY-NC-ND-3.0-IGO': 'CC-BY-NC-ND-3.0-IGO',
  'http://creativecommons.org/licenses/by-nc-nd/3.0/igo/legalcode': 'CC-BY-NC-ND-3.0-IGO',
  'https://creativecommons.org/licenses/by-nc-nd/3.0/igo/legalcode': 'CC-BY-NC-ND-3.0-IGO',
  'http://spdx.org/licenses/CC-BY-NC-ND-4.0': 'CC-BY-NC-ND-4.0',
  'https://spdx.org/licenses/CC-BY-NC-ND-4.0': 'CC-BY-NC-ND-4.0',
  'http://creativecommons.org/licenses/by-nc-nd/4.0/legalcode': 'CC-BY-NC-ND-4.0',
  'https://creativecommons.org/licenses/by-nc-nd/4.0/legalcode': 'CC-BY-NC-ND-4.0',
  'http://spdx.org/licenses/CC-BY-NC-SA-1.0': 'CC-BY-NC-SA-1.0',
  'https://spdx.org/licenses/CC-BY-NC-SA-1.0': 'CC-BY-NC-SA-1.0',
  'http://creativecommons.org/licenses/by-nc-sa/1.0/legalcode': 'CC-BY-NC-SA-1.0',
  'https://creativecommons.org/licenses/by-nc-sa/1.0/legalcode': 'CC-BY-NC-SA-1.0',
  'http://spdx.org/licenses/CC-BY-NC-SA-2.0': 'CC-BY-NC-SA-2.0',
  'https://spdx.org/licenses/CC-BY-NC-SA-2.0': 'CC-BY-NC-SA-2.0',
  'http://creativecommons.org/licenses/by-nc-sa/2.0/legalcode': 'CC-BY-NC-SA-2.0',
  'https://creativecommons.org/licenses/by-nc-sa/2.0/legalcode': 'CC-BY-NC-SA-2.0',
  'http://spdx.org/licenses/CC-BY-NC-SA-2.0-DE': 'CC-BY-NC-SA-2.0-DE',
  'https://spdx.org/licenses/CC-BY-NC-SA-2.0-DE': 'CC-BY-NC-SA-2.0-DE',
  'http://creativecommons.org/licenses/by-nc-sa/2.0/de/legalcode': 'CC-BY-NC-SA-2.0-DE',
  'https://creativecommons.org/licenses/by-nc-sa/2.0/de/legalcode': 'CC-BY-NC-SA-2.0-DE',
  'http://spdx.org/licenses/CC-BY-NC-SA-2.0-FR': 'CC-BY-NC-SA-2.0-FR',
  'https://spdx.org/licenses/CC-BY-NC-SA-2.0-FR': 'CC-BY-NC-SA-2.0-FR',
  'http://creativecommons.org/licenses/by-nc-sa/2.0/fr/legalcode': 'CC-BY-NC-SA-2.0-FR',
  'https://creativecommons.org/licenses/by-nc-sa/2.0/fr/legalcode': 'CC-BY-NC-SA-2.0-FR',
  'http://spdx.org/licenses/CC-BY-NC-SA-2.0-UK': 'CC-BY-NC-SA-2.0-UK',
  'https://spdx.org/licenses/CC-BY-NC-SA-2.0-UK': 'CC-BY-NC-SA-2.0-UK',
  'http://creativecommons.org/licenses/by-nc-sa/2.0/uk/legalcode': 'CC-BY-NC-SA-2.0-UK',
  'https://creativecommons.org/licenses/by-nc-sa/2.0/uk/legalcode': 'CC-BY-NC-SA-2.0-UK',
  'http://spdx.org/licenses/CC-BY-NC-SA-2.5': 'CC-BY-NC-SA-2.5',
  'https://spdx.org/licenses/CC-BY-NC-SA-2.5': 'CC-BY-NC-SA-2.5',
  'http://creativecommons.org/licenses/by-nc-sa/2.5/legalcode': 'CC-BY-NC-SA-2.5',
  'https://creativecommons.org/licenses/by-nc-sa/2.5/legalcode': 'CC-BY-NC-SA-2.5',
  'http://spdx.org/licenses/CC-BY-NC-SA-3.0': 'CC-BY-NC-SA-3.0',
  'https://spdx.org/licenses/CC-BY-NC-SA-3.0': 'CC-BY-NC-SA-3.0',
  'http://creativecommons.org/licenses/by-nc-sa/3.0/legalcode': 'CC-BY-NC-SA-3.0',
  'https://creativecommons.org/licenses/by-nc-sa/3.0/legalcode': 'CC-BY-NC-SA-3.0',
  'http://spdx.org/licenses/CC-BY-NC-SA-3.0-DE': 'CC-BY-NC-SA-3.0-DE',
  'https://spdx.org/licenses/CC-BY-NC-SA-3.0-DE': 'CC-BY-NC-SA-3.0-DE',
  'http://creativecommons.org/licenses/by-nc-sa/3.0/de/legalcode': 'CC-BY-NC-SA-3.0-DE',
  'https://creativecommons.org/licenses/by-nc-sa/3.0/de/legalcode': 'CC-BY-NC-SA-3.0-DE',
  'http://spdx.org/licenses/CC-BY-NC-SA-3.0-IGO': 'CC-BY-NC-SA-3.0-IGO',
  'https://spdx.org/licenses/CC-BY-NC-SA-3.0-IGO': 'CC-BY-NC-SA-3.0-IGO',
  'http://creativecommons.org/licenses/by-nc-sa/3.0/igo/legalcode': 'CC-BY-NC-SA-3.0-IGO',
  'https://creativecommons.org/licenses/by-nc-sa/3.0/igo/legalcode': 'CC-BY-NC-SA-3.0-IGO',
  'http://spdx.org/licenses/CC-BY-NC-SA-4.0': 'CC-BY-NC-SA-4.0',
  'https://spdx.org/licenses/CC-BY-NC-SA-4.0': 'CC-BY-NC-SA-4.0',
  'http://creativecommons.org/licenses/by-nc-sa/4.0/legalcode': 'CC-BY-NC-SA-4.0',
  'https://creativecommons.org/licenses/by-nc-sa/4.0/legalcode': 'CC-BY-NC-SA-4.0',
  'http://spdx.org/licenses/CC-BY-ND-1.0': 'CC-BY-ND-1.0',
  'https://spdx.org/licenses/CC-BY-ND-1.0': 'CC-BY-ND-1.0',
  'http://creativecommons.org/licenses/by-nd/1.0/legalcode': 'CC-BY-ND-1.0',
  'https://creativecommons.org/licenses/by-nd/1.0/legalcode': 'CC-BY-ND-1.0',
  'http://spdx.org/licenses/CC-BY-ND-2.0': 'CC-BY-ND-2.0',
  'https://spdx.org/licenses/CC-BY-ND-2.0': 'CC-BY-ND-2.0',
  'http://creativecommons.org/licenses/by-nd/2.0/legalcode': 'CC-BY-ND-2.0',
  'https://creativecommons.org/licenses/by-nd/2.0/legalcode': 'CC-BY-ND-2.0',
  'http://spdx.org/licenses/CC-BY-ND-2.5': 'CC-BY-ND-2.5',
  'https://spdx.org/licenses/CC-BY-ND-2.5': 'CC-BY-ND-2.5',
  'http://creativecommons.org/licenses/by-nd/2.5/legalcode': 'CC-BY-ND-2.5',
  'https://creativecommons.org/licenses/by-nd/2.5/legalcode': 'CC-BY-ND-2.5',
  'http://spdx.org/licenses/CC-BY-ND-3.0': 'CC-BY-ND-3.0',
  'https://spdx.org/licenses/CC-BY-ND-3.0': 'CC-BY-ND-3.0',
  'http://creativecommons.org/licenses/by-nd/3.0/legalcode': 'CC-BY-ND-3.0',
  'https://creativecommons.org/licenses/by-nd/3.0/legalcode': 'CC-BY-ND-3.0',
  'http://spdx.org/licenses/CC-BY-ND-3.0-DE': 'CC-BY-ND-3.0-DE',
  'https://spdx.org/licenses/CC-BY-ND-3.0-DE': 'CC-BY-ND-3.0-DE',
  'http://creativecommons.org/licenses/by-nd/3.0/de/legalcode': 'CC-BY-ND-3.0-DE',
  'https://creativecommons.org/licenses/by-nd/3.0/de/legalcode': 'CC-BY-ND-3.0-DE',
  'http://spdx.org/licenses/CC-BY-ND-4.0': 'CC-BY-ND-4.0',
  'https://spdx.org/licenses/CC-BY-ND-4.0': 'CC-BY-ND-4.0',
  'http://creativecommons.org/licenses/by-nd/4.0/legalcode': 'CC-BY-ND-4.0',
  'https://creativecommons.org/licenses/by-nd/4.0/legalcode': 'CC-BY-ND-4.0',
  'http://spdx.org/licenses/CC-BY-SA-1.0': 'CC-BY-SA-1.0',
  'https://spdx.org/licenses/CC-BY-SA-1.0': 'CC-BY-SA-1.0',
  'http://creativecommons.org/licenses/by-sa/1.0/legalcode': 'CC-BY-SA-1.0',
  'https://creativecommons.org/licenses/by-sa/1.0/legalcode': 'CC-BY-SA-1.0',
  'http://spdx.org/licenses/CC-BY-SA-2.0': 'CC-BY-SA-2.0',
  'https://spdx.org/licenses/CC-BY-SA-2.0': 'CC-BY-SA-2.0',
  'http://creativecommons.org/licenses/by-sa/2.0/legalcode': 'CC-BY-SA-2.0',
  'https://creativecommons.org/licenses/by-sa/2.0/legalcode': 'CC-BY-SA-2.0',
  'http://spdx.org/licenses/CC-BY-SA-2.0-UK': 'CC-BY-SA-2.0-UK',
  'https://spdx.org/licenses/CC-BY-SA-2.0-UK': 'CC-BY-SA-2.0-UK',
  'http://creativecommons.org/licenses/by-sa/2.0/uk/legalcode': 'CC-BY-SA-2.0-UK',
  'https://creativecommons.org/licenses/by-sa/2.0/uk/legalcode': 'CC-BY-SA-2.0-UK',
  'http://spdx.org/licenses/CC-BY-SA-2.1-JP': 'CC-BY-SA-2.1-JP',
  'https://spdx.org/licenses/CC-BY-SA-2.1-JP': 'CC-BY-SA-2.1-JP',
  'http://creativecommons.org/licenses/by-sa/2.1/jp/legalcode': 'CC-BY-SA-2.1-JP',
  'https://creativecommons.org/licenses/by-sa/2.1/jp/legalcode': 'CC-BY-SA-2.1-JP',
  'http://spdx.org/licenses/CC-BY-SA-2.5': 'CC-BY-SA-2.5',
  'https://spdx.org/licenses/CC-BY-SA-2.5': 'CC-BY-SA-2.5',
  'http://creativecommons.org/licenses/by-sa/2.5/legalcode': 'CC-BY-SA-2.5',
  'https://creativecommons.org/licenses/by-sa/2.5/legalcode': 'CC-BY-SA-2.5',
  'http://spdx.org/licenses/CC-BY-SA-3.0': 'CC-BY-SA-3.0',
  'https://spdx.org/licenses/CC-BY-SA-3.0': 'CC-BY-SA-3.0',
  'http://creativecommons.org/licenses/by-sa/3.0/legalcode': 'CC-BY-SA-3.0',
  'https://creativecommons.org/licenses/by-sa/3.0/legalcode': 'CC-BY-SA-3.0',
  'http://spdx.org/licenses/CC-BY-SA-3.0-AT': 'CC-BY-SA-3.0-AT',
  'https://spdx.org/licenses/CC-BY-SA-3.0-AT': 'CC-BY-SA-3.0-AT',
  'http://creativecommons.org/licenses/by-sa/3.0/at/legalcode': 'CC-BY-SA-3.0-AT',
  'https://creativecommons.org/licenses/by-sa/3.0/at/legalcode': 'CC-BY-SA-3.0-AT',
  'http://spdx.org/licenses/CC-BY-SA-3.0-DE': 'CC-BY-SA-3.0-DE',
  'https://spdx.org/licenses/CC-BY-SA-3.0-DE': 'CC-BY-SA-3.0-DE',
  'http://creativecommons.org/licenses/by-sa/3.0/de/legalcode': 'CC-BY-SA-3.0-DE',
  'https://creativecommons.org/licenses/by-sa/3.0/de/legalcode': 'CC-BY-SA-3.0-DE',
  'http://spdx.org/licenses/CC-BY-SA-4.0': 'CC-BY-SA-4.0',
  'https://spdx.org/licenses/CC-BY-SA-4.0': 'CC-BY-SA-4.0',
  'http://creativecommons.org/licenses/by-sa/4.0/legalcode': 'CC-BY-SA-4.0',
  'https://creativecommons.org/licenses/by-sa/4.0/legalcode': 'CC-BY-SA-4.0',
  'http://spdx.org/licenses/CC-PDDC': 'CC-PDDC',
  'https://spdx.org/licenses/CC-PDDC': 'CC-PDDC',
  'http://creativecommons.org/licenses/publicdomain/': 'CC-PDDC',
  'https://creativecommons.org/licenses/publicdomain/': 'CC-PDDC',
  'http://spdx.org/licenses/CC0-1.0': 'CC0-1.0',
  'https://spdx.org/licenses/CC0-1.0': 'CC0-1.0',
  'http://creativecommons.org/publicdomain/zero/1.0/legalcode': 'CC0-1.0',
  'https://creativecommons.org/publicdomain/zero/1.0/legalcode': 'CC0-1.0',
  'http://spdx.org/licenses/CDDL-1.0': 'CDDL-1.0',
  'https://spdx.org/licenses/CDDL-1.0': 'CDDL-1.0',
  'http://opensource.org/licenses/cddl1': 'CDDL-1.0',
  'https://opensource.org/licenses/cddl1': 'CDDL-1.0',
  'http://spdx.org/licenses/CDDL-1.1': 'CDDL-1.1',
  'https://spdx.org/licenses/CDDL-1.1': 'CDDL-1.1',
  'http://glassfish.java.net/public/CDDL+GPL_1_1.html': 'CDDL-1.1',
  'https://glassfish.java.net/public/CDDL+GPL_1_1.html': 'CDDL-1.1',
  'http://javaee.github.io/glassfish/LICENSE': 'CDDL-1.1',
  'https://javaee.github.io/glassfish/LICENSE': 'CDDL-1.1',
  'http://spdx.org/licenses/CDL-1.0': 'CDL-1.0',
  'https://spdx.org/licenses/CDL-1.0': 'CDL-1.0',
  'http://www.opensource.apple.com/cdl/': 'CDL-1.0',
  'https://www.opensource.apple.com/cdl/': 'CDL-1.0',
  'http://fedoraproject.org/wiki/Licensing/Common_Documentation_License': 'CDL-1.0',
  'https://fedoraproject.org/wiki/Licensing/Common_Documentation_License': 'CDL-1.0',
  'http://www.gnu.org/licenses/license-list.html#ACDL': 'CDL-1.0',
  'https://www.gnu.org/licenses/license-list.html#ACDL': 'CDL-1.0',
  'http://spdx.org/licenses/CDLA-Permissive-1.0': 'CDLA-Permissive-1.0',
  'https://spdx.org/licenses/CDLA-Permissive-1.0': 'CDLA-Permissive-1.0',
  'http://cdla.io/permissive-1-0': 'CDLA-Permissive-1.0',
  'https://cdla.io/permissive-1-0': 'CDLA-Permissive-1.0',
  'http://spdx.org/licenses/CDLA-Permissive-2.0': 'CDLA-Permissive-2.0',
  'https://spdx.org/licenses/CDLA-Permissive-2.0': 'CDLA-Permissive-2.0',
  'http://cdla.dev/permissive-2-0': 'CDLA-Permissive-2.0',
  'https://cdla.dev/permissive-2-0': 'CDLA-Permissive-2.0',
  'http://spdx.org/licenses/CDLA-Sharing-1.0': 'CDLA-Sharing-1.0',
  'https://spdx.org/licenses/CDLA-Sharing-1.0': 'CDLA-Sharing-1.0',
  'http://cdla.io/sharing-1-0': 'CDLA-Sharing-1.0',
  'https://cdla.io/sharing-1-0': 'CDLA-Sharing-1.0',
  'http://spdx.org/licenses/CECILL-1.0': 'CECILL-1.0',
  'https://spdx.org/licenses/CECILL-1.0': 'CECILL-1.0',
  'http://www.cecill.info/licences/Licence_CeCILL_V1-fr.html': 'CECILL-1.0',
  'https://www.cecill.info/licences/Licence_CeCILL_V1-fr.html': 'CECILL-1.0',
  'http://spdx.org/licenses/CECILL-1.1': 'CECILL-1.1',
  'https://spdx.org/licenses/CECILL-1.1': 'CECILL-1.1',
  'http://www.cecill.info/licences/Licence_CeCILL_V1.1-US.html': 'CECILL-1.1',
  'https://www.cecill.info/licences/Licence_CeCILL_V1.1-US.html': 'CECILL-1.1',
  'http://spdx.org/licenses/CECILL-2.0': 'CECILL-2.0',
  'https://spdx.org/licenses/CECILL-2.0': 'CECILL-2.0',
  'http://www.cecill.info/licences/Licence_CeCILL_V2-en.html': 'CECILL-2.0',
  'https://www.cecill.info/licences/Licence_CeCILL_V2-en.html': 'CECILL-2.0',
  'http://spdx.org/licenses/CECILL-2.1': 'CECILL-2.1',
  'https://spdx.org/licenses/CECILL-2.1': 'CECILL-2.1',
  'http://www.cecill.info/licences/Licence_CeCILL_V2.1-en.html': 'CECILL-2.1',
  'https://www.cecill.info/licences/Licence_CeCILL_V2.1-en.html': 'CECILL-2.1',
  'http://spdx.org/licenses/CECILL-B': 'CECILL-B',
  'https://spdx.org/licenses/CECILL-B': 'CECILL-B',
  'http://www.cecill.info/licences/Licence_CeCILL-B_V1-en.html': 'CECILL-B',
  'https://www.cecill.info/licences/Licence_CeCILL-B_V1-en.html': 'CECILL-B',
  'http://spdx.org/licenses/CECILL-C': 'CECILL-C',
  'https://spdx.org/licenses/CECILL-C': 'CECILL-C',
  'http://www.cecill.info/licences/Licence_CeCILL-C_V1-en.html': 'CECILL-C',
  'https://www.cecill.info/licences/Licence_CeCILL-C_V1-en.html': 'CECILL-C',
  'http://spdx.org/licenses/CERN-OHL-1.1': 'CERN-OHL-1.1',
  'https://spdx.org/licenses/CERN-OHL-1.1': 'CERN-OHL-1.1',
  'http://www.ohwr.org/project/licenses/wikis/cern-ohl-v1.1': 'CERN-OHL-1.1',
  'https://www.ohwr.org/project/licenses/wikis/cern-ohl-v1.1': 'CERN-OHL-1.1',
  'http://spdx.org/licenses/CERN-OHL-1.2': 'CERN-OHL-1.2',
  'https://spdx.org/licenses/CERN-OHL-1.2': 'CERN-OHL-1.2',
  'http://www.ohwr.org/project/licenses/wikis/cern-ohl-v1.2': 'CERN-OHL-1.2',
  'https://www.ohwr.org/project/licenses/wikis/cern-ohl-v1.2': 'CERN-OHL-1.2',
  'http://spdx.org/licenses/CERN-OHL-P-2.0': 'CERN-OHL-P-2.0',
  'https://spdx.org/licenses/CERN-OHL-P-2.0': 'CERN-OHL-P-2.0',
  'http://www.ohwr.org/project/cernohl/wikis/Documents/CERN-OHL-version-2': 'CERN-OHL-W-2.0',
  'https://www.ohwr.org/project/cernohl/wikis/Documents/CERN-OHL-version-2': 'CERN-OHL-W-2.0',
  'http://spdx.org/licenses/CERN-OHL-S-2.0': 'CERN-OHL-S-2.0',
  'https://spdx.org/licenses/CERN-OHL-S-2.0': 'CERN-OHL-S-2.0',
  'http://spdx.org/licenses/CERN-OHL-W-2.0': 'CERN-OHL-W-2.0',
  'https://spdx.org/licenses/CERN-OHL-W-2.0': 'CERN-OHL-W-2.0',
  'http://spdx.org/licenses/CFITSIO': 'CFITSIO',
  'https://spdx.org/licenses/CFITSIO': 'CFITSIO',
  'http://heasarc.gsfc.nasa.gov/docs/software/fitsio/c/f_user/node9.html': 'CFITSIO',
  'https://heasarc.gsfc.nasa.gov/docs/software/fitsio/c/f_user/node9.html': 'CFITSIO',
  'http://spdx.org/licenses/checkmk': 'checkmk',
  'https://spdx.org/licenses/checkmk': 'checkmk',
  'http://github.com/libcheck/check/blob/master/checkmk/checkmk.in': 'checkmk',
  'https://github.com/libcheck/check/blob/master/checkmk/checkmk.in': 'checkmk',
  'http://spdx.org/licenses/ClArtistic': 'ClArtistic',
  'https://spdx.org/licenses/ClArtistic': 'ClArtistic',
  'http://gianluca.dellavedova.org/2011/01/03/clarified-artistic-license/': 'ClArtistic',
  'https://gianluca.dellavedova.org/2011/01/03/clarified-artistic-license/': 'ClArtistic',
  'http://www.ncftp.com/ncftp/doc/LICENSE.txt': 'ClArtistic',
  'https://www.ncftp.com/ncftp/doc/LICENSE.txt': 'ClArtistic',
  'http://spdx.org/licenses/Clips': 'Clips',
  'https://spdx.org/licenses/Clips': 'Clips',
  'http://github.com/DrItanium/maya/blob/master/LICENSE.CLIPS': 'Clips',
  'https://github.com/DrItanium/maya/blob/master/LICENSE.CLIPS': 'Clips',
  'http://spdx.org/licenses/CMU-Mac': 'CMU-Mach',
  'https://spdx.org/licenses/CMU-Mac': 'CMU-Mach',
  'http://www.cs.cmu.edu/~410/licenses.html': 'CMU-Mach',
  'https://www.cs.cmu.edu/~410/licenses.html': 'CMU-Mach',
  'http://spdx.org/licenses/CNRI-Jython': 'CNRI-Jython',
  'https://spdx.org/licenses/CNRI-Jython': 'CNRI-Jython',
  'http://www.jython.org/license.html': 'CNRI-Jython',
  'https://www.jython.org/license.html': 'CNRI-Jython',
  'http://spdx.org/licenses/CNRI-Python': 'CNRI-Python',
  'https://spdx.org/licenses/CNRI-Python': 'CNRI-Python',
  'http://opensource.org/licenses/CNRI-Python': 'CNRI-Python',
  'https://opensource.org/licenses/CNRI-Python': 'CNRI-Python',
  'http://spdx.org/licenses/CNRI-Python-GPL-Compatible': 'CNRI-Python-GPL-Compatible',
  'https://spdx.org/licenses/CNRI-Python-GPL-Compatible': 'CNRI-Python-GPL-Compatible',
  'http://www.python.org/download/releases/1.6.1/download_win/': 'CNRI-Python-GPL-Compatible',
  'https://www.python.org/download/releases/1.6.1/download_win/': 'CNRI-Python-GPL-Compatible',
  'http://spdx.org/licenses/COIL-1.0': 'COIL-1.0',
  'https://spdx.org/licenses/COIL-1.0': 'COIL-1.0',
  'http://coil.apotheon.org/plaintext/01.0.txt': 'COIL-1.0',
  'https://coil.apotheon.org/plaintext/01.0.txt': 'COIL-1.0',
  'http://spdx.org/licenses/Community-Spec-1.0': 'Community-Spec-1.0',
  'https://spdx.org/licenses/Community-Spec-1.0': 'Community-Spec-1.0',
  'http://github.com/CommunitySpecification/1.0/blob/master/1._Community_Specification_License-v1.md': 'Community-Spec-1.0',
  'https://github.com/CommunitySpecification/1.0/blob/master/1._Community_Specification_License-v1.md': 'Community-Spec-1.0',
  'http://spdx.org/licenses/Condor-1.1': 'Condor-1.1',
  'https://spdx.org/licenses/Condor-1.1': 'Condor-1.1',
  'http://research.cs.wisc.edu/condor/license.html#condor': 'Condor-1.1',
  'https://research.cs.wisc.edu/condor/license.html#condor': 'Condor-1.1',
  'http://web.archive.org/web/20111123062036/http://research.cs.wisc.edu/condor/license.html#condor': 'Condor-1.1',
  'https://web.archive.org/web/20111123062036/https://research.cs.wisc.edu/condor/license.html#condor': 'Condor-1.1',
  'http://spdx.org/licenses/copyleft-next-0.3.0': 'copyleft-next-0.3.0',
  'https://spdx.org/licenses/copyleft-next-0.3.0': 'copyleft-next-0.3.0',
  'http://github.com/copyleft-next/copyleft-next/blob/master/Releases/copyleft-next-0.3.0': 'copyleft-next-0.3.0',
  'https://github.com/copyleft-next/copyleft-next/blob/master/Releases/copyleft-next-0.3.0': 'copyleft-next-0.3.0',
  'http://spdx.org/licenses/copyleft-next-0.3.1': 'copyleft-next-0.3.1',
  'https://spdx.org/licenses/copyleft-next-0.3.1': 'copyleft-next-0.3.1',
  'http://github.com/copyleft-next/copyleft-next/blob/master/Releases/copyleft-next-0.3.1': 'copyleft-next-0.3.1',
  'https://github.com/copyleft-next/copyleft-next/blob/master/Releases/copyleft-next-0.3.1': 'copyleft-next-0.3.1',
  'http://spdx.org/licenses/Cornell-Lossless-JPEG': 'Cornell-Lossless-JPEG',
  'https://spdx.org/licenses/Cornell-Lossless-JPEG': 'Cornell-Lossless-JPEG',
  'http://android.googlesource.com/platform/external/dng_sdk/+/refs/heads/master/source/dng_lossless_jpeg.cpp#16': 'Cornell-Lossless-JPEG',
  'https://android.googlesource.com/platform/external/dng_sdk/+/refs/heads/master/source/dng_lossless_jpeg.cpp#16': 'Cornell-Lossless-JPEG',
  'http://www.mssl.ucl.ac.uk/~mcrw/src/20050920/proto.h': 'Cornell-Lossless-JPEG',
  'https://www.mssl.ucl.ac.uk/~mcrw/src/20050920/proto.h': 'Cornell-Lossless-JPEG',
  'http://gitlab.freedesktop.org/libopenraw/libopenraw/blob/master/lib/ljpegdecompressor.cpp#L32': 'Cornell-Lossless-JPEG',
  'https://gitlab.freedesktop.org/libopenraw/libopenraw/blob/master/lib/ljpegdecompressor.cpp#L32': 'Cornell-Lossless-JPEG',
  'http://spdx.org/licenses/CPAL-1.0': 'CPAL-1.0',
  'https://spdx.org/licenses/CPAL-1.0': 'CPAL-1.0',
  'http://opensource.org/licenses/CPAL-1.0': 'CPAL-1.0',
  'https://opensource.org/licenses/CPAL-1.0': 'CPAL-1.0',
  'http://spdx.org/licenses/CPL-1.0': 'CPL-1.0',
  'https://spdx.org/licenses/CPL-1.0': 'CPL-1.0',
  'http://opensource.org/licenses/CPL-1.0': 'CPL-1.0',
  'https://opensource.org/licenses/CPL-1.0': 'CPL-1.0',
  'http://spdx.org/licenses/CPOL-1.02': 'CPOL-1.02',
  'https://spdx.org/licenses/CPOL-1.02': 'CPOL-1.02',
  'http://www.codeproject.com/info/cpol10.aspx': 'CPOL-1.02',
  'https://www.codeproject.com/info/cpol10.aspx': 'CPOL-1.02',
  'http://spdx.org/licenses/Crossword': 'Crossword',
  'https://spdx.org/licenses/Crossword': 'Crossword',
  'http://fedoraproject.org/wiki/Licensing/Crossword': 'Crossword',
  'https://fedoraproject.org/wiki/Licensing/Crossword': 'Crossword',
  'http://spdx.org/licenses/CrystalStacker': 'CrystalStacker',
  'https://spdx.org/licenses/CrystalStacker': 'CrystalStacker',
  'http://fedoraproject.org/wiki/Licensing:CrystalStacker?rd=Licensing/CrystalStacker': 'CrystalStacker',
  'https://fedoraproject.org/wiki/Licensing:CrystalStacker?rd=Licensing/CrystalStacker': 'CrystalStacker',
  'http://spdx.org/licenses/CUA-OPL-1.0': 'CUA-OPL-1.0',
  'https://spdx.org/licenses/CUA-OPL-1.0': 'CUA-OPL-1.0',
  'http://opensource.org/licenses/CUA-OPL-1.0': 'CUA-OPL-1.0',
  'https://opensource.org/licenses/CUA-OPL-1.0': 'CUA-OPL-1.0',
  'http://spdx.org/licenses/Cube': 'Cube',
  'https://spdx.org/licenses/Cube': 'Cube',
  'http://fedoraproject.org/wiki/Licensing/Cube': 'Cube',
  'https://fedoraproject.org/wiki/Licensing/Cube': 'Cube',
  'http://spdx.org/licenses/cur': 'curl',
  'https://spdx.org/licenses/cur': 'curl',
  'http://github.com/bagder/curl/blob/master/COPYING': 'curl',
  'https://github.com/bagder/curl/blob/master/COPYING': 'curl',
  'http://spdx.org/licenses/D-FSL-1.0': 'D-FSL-1.0',
  'https://spdx.org/licenses/D-FSL-1.0': 'D-FSL-1.0',
  'http://www.dipp.nrw.de/d-fsl/lizenzen/': 'D-FSL-1.0',
  'https://www.dipp.nrw.de/d-fsl/lizenzen/': 'D-FSL-1.0',
  'http://www.dipp.nrw.de/d-fsl/index_html/lizenzen/de/D-FSL-1_0_de.txt': 'D-FSL-1.0',
  'https://www.dipp.nrw.de/d-fsl/index_html/lizenzen/de/D-FSL-1_0_de.txt': 'D-FSL-1.0',
  'http://www.dipp.nrw.de/d-fsl/index_html/lizenzen/en/D-FSL-1_0_en.txt': 'D-FSL-1.0',
  'https://www.dipp.nrw.de/d-fsl/index_html/lizenzen/en/D-FSL-1_0_en.txt': 'D-FSL-1.0',
  'http://www.hbz-nrw.de/produkte/open-access/lizenzen/dfsl': 'D-FSL-1.0',
  'https://www.hbz-nrw.de/produkte/open-access/lizenzen/dfsl': 'D-FSL-1.0',
  'http://www.hbz-nrw.de/produkte/open-access/lizenzen/dfsl/deutsche-freie-software-lizenz': 'D-FSL-1.0',
  'https://www.hbz-nrw.de/produkte/open-access/lizenzen/dfsl/deutsche-freie-software-lizenz': 'D-FSL-1.0',
  'http://www.hbz-nrw.de/produkte/open-access/lizenzen/dfsl/german-free-software-license': 'D-FSL-1.0',
  'https://www.hbz-nrw.de/produkte/open-access/lizenzen/dfsl/german-free-software-license': 'D-FSL-1.0',
  'http://www.hbz-nrw.de/produkte/open-access/lizenzen/dfsl/D-FSL-1_0_de.txt/at_download/file': 'D-FSL-1.0',
  'https://www.hbz-nrw.de/produkte/open-access/lizenzen/dfsl/D-FSL-1_0_de.txt/at_download/file': 'D-FSL-1.0',
  'http://www.hbz-nrw.de/produkte/open-access/lizenzen/dfsl/D-FSL-1_0_en.txt/at_download/file': 'D-FSL-1.0',
  'https://www.hbz-nrw.de/produkte/open-access/lizenzen/dfsl/D-FSL-1_0_en.txt/at_download/file': 'D-FSL-1.0',
  'http://spdx.org/licenses/diffmark': 'diffmark',
  'https://spdx.org/licenses/diffmark': 'diffmark',
  'http://fedoraproject.org/wiki/Licensing/diffmark': 'diffmark',
  'https://fedoraproject.org/wiki/Licensing/diffmark': 'diffmark',
  'http://spdx.org/licenses/DL-DE-BY-2.0': 'DL-DE-BY-2.0',
  'https://spdx.org/licenses/DL-DE-BY-2.0': 'DL-DE-BY-2.0',
  'http://www.govdata.de/dl-de/by-2-0': 'DL-DE-BY-2.0',
  'https://www.govdata.de/dl-de/by-2-0': 'DL-DE-BY-2.0',
  'http://spdx.org/licenses/DOC': 'DOC',
  'https://spdx.org/licenses/DOC': 'DOC',
  'http://www.cs.wustl.edu/~schmidt/ACE-copying.html': 'DOC',
  'https://www.cs.wustl.edu/~schmidt/ACE-copying.html': 'DOC',
  'http://www.dre.vanderbilt.edu/~schmidt/ACE-copying.html': 'DOC',
  'https://www.dre.vanderbilt.edu/~schmidt/ACE-copying.html': 'DOC',
  'http://spdx.org/licenses/Dotseqn': 'Dotseqn',
  'https://spdx.org/licenses/Dotseqn': 'Dotseqn',
  'http://fedoraproject.org/wiki/Licensing/Dotseqn': 'Dotseqn',
  'https://fedoraproject.org/wiki/Licensing/Dotseqn': 'Dotseqn',
  'http://spdx.org/licenses/DRL-1.0': 'DRL-1.0',
  'https://spdx.org/licenses/DRL-1.0': 'DRL-1.0',
  'http://github.com/Neo23x0/sigma/blob/master/LICENSE.Detection.Rules.md': 'DRL-1.0',
  'https://github.com/Neo23x0/sigma/blob/master/LICENSE.Detection.Rules.md': 'DRL-1.0',
  'http://spdx.org/licenses/DSDP': 'DSDP',
  'https://spdx.org/licenses/DSDP': 'DSDP',
  'http://fedoraproject.org/wiki/Licensing/DSDP': 'DSDP',
  'https://fedoraproject.org/wiki/Licensing/DSDP': 'DSDP',
  'http://spdx.org/licenses/dvipdf': 'dvipdfm',
  'https://spdx.org/licenses/dvipdf': 'dvipdfm',
  'http://fedoraproject.org/wiki/Licensing/dvipdfm': 'dvipdfm',
  'https://fedoraproject.org/wiki/Licensing/dvipdfm': 'dvipdfm',
  'http://spdx.org/licenses/ECL-1.0': 'ECL-1.0',
  'https://spdx.org/licenses/ECL-1.0': 'ECL-1.0',
  'http://opensource.org/licenses/ECL-1.0': 'ECL-1.0',
  'https://opensource.org/licenses/ECL-1.0': 'ECL-1.0',
  'http://spdx.org/licenses/ECL-2.0': 'ECL-2.0',
  'https://spdx.org/licenses/ECL-2.0': 'ECL-2.0',
  'http://opensource.org/licenses/ECL-2.0': 'ECL-2.0',
  'https://opensource.org/licenses/ECL-2.0': 'ECL-2.0',
  'http://spdx.org/licenses/eCos-2.0': 'eCos-2.0',
  'https://spdx.org/licenses/eCos-2.0': 'eCos-2.0',
  'http://www.gnu.org/licenses/ecos-license.html': 'eCos-2.0',
  'https://www.gnu.org/licenses/ecos-license.html': 'eCos-2.0',
  'http://spdx.org/licenses/EFL-1.0': 'EFL-1.0',
  'https://spdx.org/licenses/EFL-1.0': 'EFL-1.0',
  'http://www.eiffel-nice.org/license/forum.txt': 'EFL-1.0',
  'https://www.eiffel-nice.org/license/forum.txt': 'EFL-1.0',
  'http://opensource.org/licenses/EFL-1.0': 'EFL-1.0',
  'https://opensource.org/licenses/EFL-1.0': 'EFL-1.0',
  'http://spdx.org/licenses/EFL-2.0': 'EFL-2.0',
  'https://spdx.org/licenses/EFL-2.0': 'EFL-2.0',
  'http://www.eiffel-nice.org/license/eiffel-forum-license-2.html': 'EFL-2.0',
  'https://www.eiffel-nice.org/license/eiffel-forum-license-2.html': 'EFL-2.0',
  'http://opensource.org/licenses/EFL-2.0': 'EFL-2.0',
  'https://opensource.org/licenses/EFL-2.0': 'EFL-2.0',
  'http://spdx.org/licenses/eGenix': 'eGenix',
  'https://spdx.org/licenses/eGenix': 'eGenix',
  'http://www.egenix.com/products/eGenix.com-Public-License-1.1.0.pdf': 'eGenix',
  'https://www.egenix.com/products/eGenix.com-Public-License-1.1.0.pdf': 'eGenix',
  'http://fedoraproject.org/wiki/Licensing/eGenix.com_Public_License_1.1.0': 'eGenix',
  'https://fedoraproject.org/wiki/Licensing/eGenix.com_Public_License_1.1.0': 'eGenix',
  'http://spdx.org/licenses/Elastic-2.0': 'Elastic-2.0',
  'https://spdx.org/licenses/Elastic-2.0': 'Elastic-2.0',
  'http://www.elastic.co/licensing/elastic-license': 'Elastic-2.0',
  'https://www.elastic.co/licensing/elastic-license': 'Elastic-2.0',
  'http://github.com/elastic/elasticsearch/blob/master/licenses/ELASTIC-LICENSE-2.0.txt': 'Elastic-2.0',
  'https://github.com/elastic/elasticsearch/blob/master/licenses/ELASTIC-LICENSE-2.0.txt': 'Elastic-2.0',
  'http://spdx.org/licenses/Entessa': 'Entessa',
  'https://spdx.org/licenses/Entessa': 'Entessa',
  'http://opensource.org/licenses/Entessa': 'Entessa',
  'https://opensource.org/licenses/Entessa': 'Entessa',
  'http://spdx.org/licenses/EPICS': 'EPICS',
  'https://spdx.org/licenses/EPICS': 'EPICS',
  'http://epics.anl.gov/license/open.php': 'EPICS',
  'https://epics.anl.gov/license/open.php': 'EPICS',
  'http://spdx.org/licenses/EPL-1.0': 'EPL-1.0',
  'https://spdx.org/licenses/EPL-1.0': 'EPL-1.0',
  'http://www.eclipse.org/legal/epl-v10.html': 'EPL-1.0',
  'https://www.eclipse.org/legal/epl-v10.html': 'EPL-1.0',
  'http://opensource.org/licenses/EPL-1.0': 'EPL-1.0',
  'https://opensource.org/licenses/EPL-1.0': 'EPL-1.0',
  'http://spdx.org/licenses/EPL-2.0': 'EPL-2.0',
  'https://spdx.org/licenses/EPL-2.0': 'EPL-2.0',
  'http://www.eclipse.org/legal/epl-2.0': 'EPL-2.0',
  'https://www.eclipse.org/legal/epl-2.0': 'EPL-2.0',
  'http://www.opensource.org/licenses/EPL-2.0': 'EPL-2.0',
  'https://www.opensource.org/licenses/EPL-2.0': 'EPL-2.0',
  'http://spdx.org/licenses/ErlPL-1.1': 'ErlPL-1.1',
  'https://spdx.org/licenses/ErlPL-1.1': 'ErlPL-1.1',
  'http://www.erlang.org/EPLICENSE': 'ErlPL-1.1',
  'https://www.erlang.org/EPLICENSE': 'ErlPL-1.1',
  'http://spdx.org/licenses/etalab-2.0': 'etalab-2.0',
  'https://spdx.org/licenses/etalab-2.0': 'etalab-2.0',
  'http://github.com/DISIC/politique-de-contribution-open-source/blob/master/LICENSE.pdf': 'etalab-2.0',
  'https://github.com/DISIC/politique-de-contribution-open-source/blob/master/LICENSE.pdf': 'etalab-2.0',
  'http://raw.githubusercontent.com/DISIC/politique-de-contribution-open-source/master/LICENSE': 'etalab-2.0',
  'https://raw.githubusercontent.com/DISIC/politique-de-contribution-open-source/master/LICENSE': 'etalab-2.0',
  'http://spdx.org/licenses/EUDatagrid': 'EUDatagrid',
  'https://spdx.org/licenses/EUDatagrid': 'EUDatagrid',
  'http://eu-datagrid.web.cern.ch/eu-datagrid/license.html': 'EUDatagrid',
  'https://eu-datagrid.web.cern.ch/eu-datagrid/license.html': 'EUDatagrid',
  'http://opensource.org/licenses/EUDatagrid': 'EUDatagrid',
  'https://opensource.org/licenses/EUDatagrid': 'EUDatagrid',
  'http://spdx.org/licenses/EUPL-1.0': 'EUPL-1.0',
  'https://spdx.org/licenses/EUPL-1.0': 'EUPL-1.0',
  'http://ec.europa.eu/idabc/en/document/7330.html': 'EUPL-1.0',
  'https://ec.europa.eu/idabc/en/document/7330.html': 'EUPL-1.0',
  'http://ec.europa.eu/idabc/servlets/Doc027f.pdf?id=31096': 'EUPL-1.0',
  'https://ec.europa.eu/idabc/servlets/Doc027f.pdf?id=31096': 'EUPL-1.0',
  'http://spdx.org/licenses/EUPL-1.1': 'EUPL-1.1',
  'https://spdx.org/licenses/EUPL-1.1': 'EUPL-1.1',
  'http://joinup.ec.europa.eu/software/page/eupl/licence-eupl': 'EUPL-1.1',
  'https://joinup.ec.europa.eu/software/page/eupl/licence-eupl': 'EUPL-1.1',
  'http://joinup.ec.europa.eu/sites/default/files/custom-page/attachment/eupl1.1.-licence-en_0.pdf': 'EUPL-1.1',
  'https://joinup.ec.europa.eu/sites/default/files/custom-page/attachment/eupl1.1.-licence-en_0.pdf': 'EUPL-1.1',
  'http://opensource.org/licenses/EUPL-1.1': 'EUPL-1.1',
  'https://opensource.org/licenses/EUPL-1.1': 'EUPL-1.1',
  'http://spdx.org/licenses/EUPL-1.2': 'EUPL-1.2',
  'https://spdx.org/licenses/EUPL-1.2': 'EUPL-1.2',
  'http://joinup.ec.europa.eu/page/eupl-text-11-12': 'EUPL-1.2',
  'https://joinup.ec.europa.eu/page/eupl-text-11-12': 'EUPL-1.2',
  'http://joinup.ec.europa.eu/sites/default/files/custom-page/attachment/eupl_v1.2_en.pdf': 'EUPL-1.2',
  'https://joinup.ec.europa.eu/sites/default/files/custom-page/attachment/eupl_v1.2_en.pdf': 'EUPL-1.2',
  'http://joinup.ec.europa.eu/sites/default/files/custom-page/attachment/2020-03/EUPL-1.2%20EN.txt': 'EUPL-1.2',
  'https://joinup.ec.europa.eu/sites/default/files/custom-page/attachment/2020-03/EUPL-1.2%20EN.txt': 'EUPL-1.2',
  'http://joinup.ec.europa.eu/sites/default/files/inline-files/EUPL%20v1_2%20EN(1).txt': 'EUPL-1.2',
  'https://joinup.ec.europa.eu/sites/default/files/inline-files/EUPL%20v1_2%20EN(1).txt': 'EUPL-1.2',
  'http://eur-lex.europa.eu/legal-content/EN/TXT/HTML/?uri=CELEX:32017D0863': 'EUPL-1.2',
  'https://eur-lex.europa.eu/legal-content/EN/TXT/HTML/?uri=CELEX:32017D0863': 'EUPL-1.2',
  'http://opensource.org/licenses/EUPL-1.2': 'EUPL-1.2',
  'https://opensource.org/licenses/EUPL-1.2': 'EUPL-1.2',
  'http://spdx.org/licenses/Eurosy': 'Eurosym',
  'https://spdx.org/licenses/Eurosy': 'Eurosym',
  'http://fedoraproject.org/wiki/Licensing/Eurosym': 'Eurosym',
  'https://fedoraproject.org/wiki/Licensing/Eurosym': 'Eurosym',
  'http://spdx.org/licenses/Fair': 'Fair',
  'https://spdx.org/licenses/Fair': 'Fair',
  'http://fairlicense.org/': 'Fair',
  'https://fairlicense.org/': 'Fair',
  'http://opensource.org/licenses/Fair': 'Fair',
  'https://opensource.org/licenses/Fair': 'Fair',
  'http://spdx.org/licenses/FDK-AAC': 'FDK-AAC',
  'https://spdx.org/licenses/FDK-AAC': 'FDK-AAC',
  'http://fedoraproject.org/wiki/Licensing/FDK-AAC': 'FDK-AAC',
  'https://fedoraproject.org/wiki/Licensing/FDK-AAC': 'FDK-AAC',
  'http://directory.fsf.org/wiki/License:Fdk': 'FDK-AAC',
  'https://directory.fsf.org/wiki/License:Fdk': 'FDK-AAC',
  'http://spdx.org/licenses/Frameworx-1.0': 'Frameworx-1.0',
  'https://spdx.org/licenses/Frameworx-1.0': 'Frameworx-1.0',
  'http://opensource.org/licenses/Frameworx-1.0': 'Frameworx-1.0',
  'https://opensource.org/licenses/Frameworx-1.0': 'Frameworx-1.0',
  'http://spdx.org/licenses/FreeBSD-DOC': 'FreeBSD-DOC',
  'https://spdx.org/licenses/FreeBSD-DOC': 'FreeBSD-DOC',
  'http://www.freebsd.org/copyright/freebsd-doc-license/': 'FreeBSD-DOC',
  'https://www.freebsd.org/copyright/freebsd-doc-license/': 'FreeBSD-DOC',
  'http://spdx.org/licenses/FreeImage': 'FreeImage',
  'https://spdx.org/licenses/FreeImage': 'FreeImage',
  'http://freeimage.sourceforge.net/freeimage-license.txt': 'FreeImage',
  'https://freeimage.sourceforge.net/freeimage-license.txt': 'FreeImage',
  'http://spdx.org/licenses/FSFAP': 'FSFAP',
  'https://spdx.org/licenses/FSFAP': 'FSFAP',
  'http://www.gnu.org/prep/maintain/html_node/License-Notices-for-Other-Files.html': 'FSFAP',
  'https://www.gnu.org/prep/maintain/html_node/License-Notices-for-Other-Files.html': 'FSFAP',
  'http://spdx.org/licenses/FSFUL': 'FSFUL',
  'https://spdx.org/licenses/FSFUL': 'FSFUL',
  'http://fedoraproject.org/wiki/Licensing/FSF_Unlimited_License': 'FSFUL',
  'https://fedoraproject.org/wiki/Licensing/FSF_Unlimited_License': 'FSFUL',
  'http://spdx.org/licenses/FSFULLR': 'FSFULLR',
  'https://spdx.org/licenses/FSFULLR': 'FSFULLR',
  'http://fedoraproject.org/wiki/Licensing/FSF_Unlimited_License#License_Retention_Variant': 'FSFULLR',
  'https://fedoraproject.org/wiki/Licensing/FSF_Unlimited_License#License_Retention_Variant': 'FSFULLR',
  'http://spdx.org/licenses/FSFULLRWD': 'FSFULLRWD',
  'https://spdx.org/licenses/FSFULLRWD': 'FSFULLRWD',
  'http://lists.gnu.org/archive/html/autoconf/2012-04/msg00061.html': 'FSFULLRWD',
  'https://lists.gnu.org/archive/html/autoconf/2012-04/msg00061.html': 'FSFULLRWD',
  'http://spdx.org/licenses/FTL': 'FTL',
  'https://spdx.org/licenses/FTL': 'FTL',
  'http://freetype.fis.uniroma2.it/FTL.TXT': 'FTL',
  'https://freetype.fis.uniroma2.it/FTL.TXT': 'FTL',
  'http://git.savannah.gnu.org/cgit/freetype/freetype2.git/tree/docs/FTL.TXT': 'FTL',
  'https://git.savannah.gnu.org/cgit/freetype/freetype2.git/tree/docs/FTL.TXT': 'FTL',
  'http://gitlab.freedesktop.org/freetype/freetype/-/raw/master/docs/FTL.TXT': 'FTL',
  'https://gitlab.freedesktop.org/freetype/freetype/-/raw/master/docs/FTL.TXT': 'FTL',
  'http://spdx.org/licenses/GD': 'GD',
  'https://spdx.org/licenses/GD': 'GD',
  'http://libgd.github.io/manuals/2.3.0/files/license-txt.html': 'GD',
  'https://libgd.github.io/manuals/2.3.0/files/license-txt.html': 'GD',
  'http://spdx.org/licenses/GFDL-1.1': 'GFDL-1.1',
  'https://spdx.org/licenses/GFDL-1.1': 'GFDL-1.1',
  'http://www.gnu.org/licenses/old-licenses/fdl-1.1.txt': 'GFDL-1.1-or-later',
  'https://www.gnu.org/licenses/old-licenses/fdl-1.1.txt': 'GFDL-1.1-or-later',
  'http://spdx.org/licenses/GFDL-1.1-invariants-only': 'GFDL-1.1-invariants-only',
  'https://spdx.org/licenses/GFDL-1.1-invariants-only': 'GFDL-1.1-invariants-only',
  'http://spdx.org/licenses/GFDL-1.1-invariants-or-later': 'GFDL-1.1-invariants-or-later',
  'https://spdx.org/licenses/GFDL-1.1-invariants-or-later': 'GFDL-1.1-invariants-or-later',
  'http://spdx.org/licenses/GFDL-1.1-no-invariants-only': 'GFDL-1.1-no-invariants-only',
  'https://spdx.org/licenses/GFDL-1.1-no-invariants-only': 'GFDL-1.1-no-invariants-only',
  'http://spdx.org/licenses/GFDL-1.1-no-invariants-or-later': 'GFDL-1.1-no-invariants-or-later',
  'https://spdx.org/licenses/GFDL-1.1-no-invariants-or-later': 'GFDL-1.1-no-invariants-or-later',
  'http://spdx.org/licenses/GFDL-1.1-only': 'GFDL-1.1-only',
  'https://spdx.org/licenses/GFDL-1.1-only': 'GFDL-1.1-only',
  'http://spdx.org/licenses/GFDL-1.1-or-later': 'GFDL-1.1-or-later',
  'https://spdx.org/licenses/GFDL-1.1-or-later': 'GFDL-1.1-or-later',
  'http://spdx.org/licenses/GFDL-1.2': 'GFDL-1.2',
  'https://spdx.org/licenses/GFDL-1.2': 'GFDL-1.2',
  'http://www.gnu.org/licenses/old-licenses/fdl-1.2.txt': 'GFDL-1.2-or-later',
  'https://www.gnu.org/licenses/old-licenses/fdl-1.2.txt': 'GFDL-1.2-or-later',
  'http://spdx.org/licenses/GFDL-1.2-invariants-only': 'GFDL-1.2-invariants-only',
  'https://spdx.org/licenses/GFDL-1.2-invariants-only': 'GFDL-1.2-invariants-only',
  'http://spdx.org/licenses/GFDL-1.2-invariants-or-later': 'GFDL-1.2-invariants-or-later',
  'https://spdx.org/licenses/GFDL-1.2-invariants-or-later': 'GFDL-1.2-invariants-or-later',
  'http://spdx.org/licenses/GFDL-1.2-no-invariants-only': 'GFDL-1.2-no-invariants-only',
  'https://spdx.org/licenses/GFDL-1.2-no-invariants-only': 'GFDL-1.2-no-invariants-only',
  'http://spdx.org/licenses/GFDL-1.2-no-invariants-or-later': 'GFDL-1.2-no-invariants-or-later',
  'https://spdx.org/licenses/GFDL-1.2-no-invariants-or-later': 'GFDL-1.2-no-invariants-or-later',
  'http://spdx.org/licenses/GFDL-1.2-only': 'GFDL-1.2-only',
  'https://spdx.org/licenses/GFDL-1.2-only': 'GFDL-1.2-only',
  'http://spdx.org/licenses/GFDL-1.2-or-later': 'GFDL-1.2-or-later',
  'https://spdx.org/licenses/GFDL-1.2-or-later': 'GFDL-1.2-or-later',
  'http://spdx.org/licenses/GFDL-1.3': 'GFDL-1.3',
  'https://spdx.org/licenses/GFDL-1.3': 'GFDL-1.3',
  'http://www.gnu.org/licenses/fdl-1.3.txt': 'GFDL-1.3-or-later',
  'https://www.gnu.org/licenses/fdl-1.3.txt': 'GFDL-1.3-or-later',
  'http://spdx.org/licenses/GFDL-1.3-invariants-only': 'GFDL-1.3-invariants-only',
  'https://spdx.org/licenses/GFDL-1.3-invariants-only': 'GFDL-1.3-invariants-only',
  'http://spdx.org/licenses/GFDL-1.3-invariants-or-later': 'GFDL-1.3-invariants-or-later',
  'https://spdx.org/licenses/GFDL-1.3-invariants-or-later': 'GFDL-1.3-invariants-or-later',
  'http://spdx.org/licenses/GFDL-1.3-no-invariants-only': 'GFDL-1.3-no-invariants-only',
  'https://spdx.org/licenses/GFDL-1.3-no-invariants-only': 'GFDL-1.3-no-invariants-only',
  'http://spdx.org/licenses/GFDL-1.3-no-invariants-or-later': 'GFDL-1.3-no-invariants-or-later',
  'https://spdx.org/licenses/GFDL-1.3-no-invariants-or-later': 'GFDL-1.3-no-invariants-or-later',
  'http://spdx.org/licenses/GFDL-1.3-only': 'GFDL-1.3-only',
  'https://spdx.org/licenses/GFDL-1.3-only': 'GFDL-1.3-only',
  'http://spdx.org/licenses/GFDL-1.3-or-later': 'GFDL-1.3-or-later',
  'https://spdx.org/licenses/GFDL-1.3-or-later': 'GFDL-1.3-or-later',
  'http://spdx.org/licenses/Giftware': 'Giftware',
  'https://spdx.org/licenses/Giftware': 'Giftware',
  'http://liballeg.org/license.html#allegro-4-the-giftware-license': 'Giftware',
  'https://liballeg.org/license.html#allegro-4-the-giftware-license': 'Giftware',
  'http://spdx.org/licenses/GL2PS': 'GL2PS',
  'https://spdx.org/licenses/GL2PS': 'GL2PS',
  'http://www.geuz.org/gl2ps/COPYING.GL2PS': 'GL2PS',
  'https://www.geuz.org/gl2ps/COPYING.GL2PS': 'GL2PS',
  'http://spdx.org/licenses/Glide': 'Glide',
  'https://spdx.org/licenses/Glide': 'Glide',
  'http://www.users.on.net/~triforce/glidexp/COPYING.txt': 'Glide',
  'https://www.users.on.net/~triforce/glidexp/COPYING.txt': 'Glide',
  'http://spdx.org/licenses/Glulxe': 'Glulxe',
  'https://spdx.org/licenses/Glulxe': 'Glulxe',
  'http://fedoraproject.org/wiki/Licensing/Glulxe': 'Glulxe',
  'https://fedoraproject.org/wiki/Licensing/Glulxe': 'Glulxe',
  'http://spdx.org/licenses/GLWTPL': 'GLWTPL',
  'https://spdx.org/licenses/GLWTPL': 'GLWTPL',
  'http://github.com/me-shaon/GLWTPL/commit/da5f6bc734095efbacb442c0b31e33a65b9d6e85': 'GLWTPL',
  'https://github.com/me-shaon/GLWTPL/commit/da5f6bc734095efbacb442c0b31e33a65b9d6e85': 'GLWTPL',
  'http://spdx.org/licenses/gnuplo': 'gnuplot',
  'https://spdx.org/licenses/gnuplo': 'gnuplot',
  'http://fedoraproject.org/wiki/Licensing/Gnuplot': 'gnuplot',
  'https://fedoraproject.org/wiki/Licensing/Gnuplot': 'gnuplot',
  'http://spdx.org/licenses/GPL-1.0': 'GPL-1.0',
  'https://spdx.org/licenses/GPL-1.0': 'GPL-1.0',
  'http://www.gnu.org/licenses/old-licenses/gpl-1.0-standalone.html': 'GPL-1.0-or-later',
  'https://www.gnu.org/licenses/old-licenses/gpl-1.0-standalone.html': 'GPL-1.0-or-later',
  'http://spdx.org/licenses/GPL-1.0+': 'GPL-1.0+',
  'https://spdx.org/licenses/GPL-1.0+': 'GPL-1.0+',
  'http://spdx.org/licenses/GPL-1.0-only': 'GPL-1.0-only',
  'https://spdx.org/licenses/GPL-1.0-only': 'GPL-1.0-only',
  'http://spdx.org/licenses/GPL-1.0-or-later': 'GPL-1.0-or-later',
  'https://spdx.org/licenses/GPL-1.0-or-later': 'GPL-1.0-or-later',
  'http://spdx.org/licenses/GPL-2.0': 'GPL-2.0',
  'https://spdx.org/licenses/GPL-2.0': 'GPL-2.0',
  'http://www.gnu.org/licenses/old-licenses/gpl-2.0-standalone.html': 'GPL-2.0-or-later',
  'https://www.gnu.org/licenses/old-licenses/gpl-2.0-standalone.html': 'GPL-2.0-or-later',
  'http://opensource.org/licenses/GPL-2.0': 'GPL-2.0-or-later',
  'https://opensource.org/licenses/GPL-2.0': 'GPL-2.0-or-later',
  'http://spdx.org/licenses/GPL-2.0+': 'GPL-2.0+',
  'https://spdx.org/licenses/GPL-2.0+': 'GPL-2.0+',
  'http://spdx.org/licenses/GPL-2.0-only': 'GPL-2.0-only',
  'https://spdx.org/licenses/GPL-2.0-only': 'GPL-2.0-only',
  'http://spdx.org/licenses/GPL-2.0-or-later': 'GPL-2.0-or-later',
  'https://spdx.org/licenses/GPL-2.0-or-later': 'GPL-2.0-or-later',
  'http://spdx.org/licenses/GPL-2.0-with-autoconf-exception': 'GPL-2.0-with-autoconf-exception',
  'https://spdx.org/licenses/GPL-2.0-with-autoconf-exception': 'GPL-2.0-with-autoconf-exception',
  'http://ac-archive.sourceforge.net/doc/copyright.html': 'GPL-2.0-with-autoconf-exception',
  'https://ac-archive.sourceforge.net/doc/copyright.html': 'GPL-2.0-with-autoconf-exception',
  'http://spdx.org/licenses/GPL-2.0-with-bison-exception': 'GPL-2.0-with-bison-exception',
  'https://spdx.org/licenses/GPL-2.0-with-bison-exception': 'GPL-2.0-with-bison-exception',
  'http://git.savannah.gnu.org/cgit/bison.git/tree/data/yacc.c?id=193d7c7054ba7197b0789e14965b739162319b5e#n141': 'GPL-2.0-with-bison-exception',
  'https://git.savannah.gnu.org/cgit/bison.git/tree/data/yacc.c?id=193d7c7054ba7197b0789e14965b739162319b5e#n141': 'GPL-2.0-with-bison-exception',
  'http://spdx.org/licenses/GPL-2.0-with-classpath-exception': 'GPL-2.0-with-classpath-exception',
  'https://spdx.org/licenses/GPL-2.0-with-classpath-exception': 'GPL-2.0-with-classpath-exception',
  'http://www.gnu.org/software/classpath/license.html': 'GPL-2.0-with-classpath-exception',
  'https://www.gnu.org/software/classpath/license.html': 'GPL-2.0-with-classpath-exception',
  'http://spdx.org/licenses/GPL-2.0-with-font-exception': 'GPL-2.0-with-font-exception',
  'https://spdx.org/licenses/GPL-2.0-with-font-exception': 'GPL-2.0-with-font-exception',
  'http://www.gnu.org/licenses/gpl-faq.html#FontException': 'GPL-2.0-with-font-exception',
  'https://www.gnu.org/licenses/gpl-faq.html#FontException': 'GPL-2.0-with-font-exception',
  'http://spdx.org/licenses/GPL-2.0-with-GCC-exception': 'GPL-2.0-with-GCC-exception',
  'https://spdx.org/licenses/GPL-2.0-with-GCC-exception': 'GPL-2.0-with-GCC-exception',
  'http://gcc.gnu.org/git/?p=gcc.git;a=blob;f=gcc/libgcc1.c;h=762f5143fc6eed57b6797c82710f3538aa52b40b;hb=cb143a3ce4fb417c68f5fa2691a1b1b1053dfba9#l10': 'GPL-2.0-with-GCC-exception',
  'https://gcc.gnu.org/git/?p=gcc.git;a=blob;f=gcc/libgcc1.c;h=762f5143fc6eed57b6797c82710f3538aa52b40b;hb=cb143a3ce4fb417c68f5fa2691a1b1b1053dfba9#l10': 'GPL-2.0-with-GCC-exception',
  'http://spdx.org/licenses/GPL-3.0': 'GPL-3.0',
  'https://spdx.org/licenses/GPL-3.0': 'GPL-3.0',
  'http://www.gnu.org/licenses/gpl-3.0-standalone.html': 'GPL-3.0-or-later',
  'https://www.gnu.org/licenses/gpl-3.0-standalone.html': 'GPL-3.0-or-later',
  'http://opensource.org/licenses/GPL-3.0': 'GPL-3.0-or-later',
  'https://opensource.org/licenses/GPL-3.0': 'GPL-3.0-or-later',
  'http://spdx.org/licenses/GPL-3.0+': 'GPL-3.0+',
  'https://spdx.org/licenses/GPL-3.0+': 'GPL-3.0+',
  'http://spdx.org/licenses/GPL-3.0-only': 'GPL-3.0-only',
  'https://spdx.org/licenses/GPL-3.0-only': 'GPL-3.0-only',
  'http://spdx.org/licenses/GPL-3.0-or-later': 'GPL-3.0-or-later',
  'https://spdx.org/licenses/GPL-3.0-or-later': 'GPL-3.0-or-later',
  'http://spdx.org/licenses/GPL-3.0-with-autoconf-exception': 'GPL-3.0-with-autoconf-exception',
  'https://spdx.org/licenses/GPL-3.0-with-autoconf-exception': 'GPL-3.0-with-autoconf-exception',
  'http://www.gnu.org/licenses/autoconf-exception-3.0.html': 'GPL-3.0-with-autoconf-exception',
  'https://www.gnu.org/licenses/autoconf-exception-3.0.html': 'GPL-3.0-with-autoconf-exception',
  'http://spdx.org/licenses/GPL-3.0-with-GCC-exception': 'GPL-3.0-with-GCC-exception',
  'https://spdx.org/licenses/GPL-3.0-with-GCC-exception': 'GPL-3.0-with-GCC-exception',
  'http://www.gnu.org/licenses/gcc-exception-3.1.html': 'GPL-3.0-with-GCC-exception',
  'https://www.gnu.org/licenses/gcc-exception-3.1.html': 'GPL-3.0-with-GCC-exception',
  'http://spdx.org/licenses/Graphics-Gems': 'Graphics-Gems',
  'https://spdx.org/licenses/Graphics-Gems': 'Graphics-Gems',
  'http://github.com/erich666/GraphicsGems/blob/master/LICENSE.md': 'Graphics-Gems',
  'https://github.com/erich666/GraphicsGems/blob/master/LICENSE.md': 'Graphics-Gems',
  'http://spdx.org/licenses/gSOAP-1.3b': 'gSOAP-1.3b',
  'https://spdx.org/licenses/gSOAP-1.3b': 'gSOAP-1.3b',
  'http://www.cs.fsu.edu/~engelen/license.html': 'gSOAP-1.3b',
  'https://www.cs.fsu.edu/~engelen/license.html': 'gSOAP-1.3b',
  'http://spdx.org/licenses/HaskellRepor': 'HaskellReport',
  'https://spdx.org/licenses/HaskellRepor': 'HaskellReport',
  'http://fedoraproject.org/wiki/Licensing/Haskell_Language_Report_License': 'HaskellReport',
  'https://fedoraproject.org/wiki/Licensing/Haskell_Language_Report_License': 'HaskellReport',
  'http://spdx.org/licenses/Hippocratic-2.1': 'Hippocratic-2.1',
  'https://spdx.org/licenses/Hippocratic-2.1': 'Hippocratic-2.1',
  'http://firstdonoharm.dev/version/2/1/license.html': 'Hippocratic-2.1',
  'https://firstdonoharm.dev/version/2/1/license.html': 'Hippocratic-2.1',
  'http://github.com/EthicalSource/hippocratic-license/blob/58c0e646d64ff6fbee275bfe2b9492f914e3ab2a/LICENSE.txt': 'Hippocratic-2.1',
  'https://github.com/EthicalSource/hippocratic-license/blob/58c0e646d64ff6fbee275bfe2b9492f914e3ab2a/LICENSE.txt': 'Hippocratic-2.1',
  'http://spdx.org/licenses/HP-1986': 'HP-1986',
  'https://spdx.org/licenses/HP-1986': 'HP-1986',
  'http://sourceware.org/git/?p=newlib-cygwin.git;a=blob;f=newlib/libc/machine/hppa/memchr.S;h=1cca3e5e8867aa4bffef1f75a5c1bba25c0c441e;hb=HEAD#l2': 'HP-1986',
  'https://sourceware.org/git/?p=newlib-cygwin.git;a=blob;f=newlib/libc/machine/hppa/memchr.S;h=1cca3e5e8867aa4bffef1f75a5c1bba25c0c441e;hb=HEAD#l2': 'HP-1986',
  'http://spdx.org/licenses/HPND': 'HPND',
  'https://spdx.org/licenses/HPND': 'HPND',
  'http://opensource.org/licenses/HPND': 'HPND',
  'https://opensource.org/licenses/HPND': 'HPND',
  'http://spdx.org/licenses/HPND-export-US': 'HPND-export-US',
  'https://spdx.org/licenses/HPND-export-US': 'HPND-export-US',
  'http://www.kermitproject.org/ck90.html#source': 'HPND-export-US',
  'https://www.kermitproject.org/ck90.html#source': 'HPND-export-US',
  'http://spdx.org/licenses/HPND-Markus-Kuhn': 'HPND-Markus-Kuhn',
  'https://spdx.org/licenses/HPND-Markus-Kuhn': 'HPND-Markus-Kuhn',
  'http://www.cl.cam.ac.uk/~mgk25/ucs/wcwidth.c': 'HPND-Markus-Kuhn',
  'https://www.cl.cam.ac.uk/~mgk25/ucs/wcwidth.c': 'HPND-Markus-Kuhn',
  'http://sourceware.org/git/?p=binutils-gdb.git;a=blob;f=readline/readline/support/wcwidth.c;h=0f5ec995796f4813abbcf4972aec0378ab74722a;hb=HEAD#l55': 'HPND-Markus-Kuhn',
  'https://sourceware.org/git/?p=binutils-gdb.git;a=blob;f=readline/readline/support/wcwidth.c;h=0f5ec995796f4813abbcf4972aec0378ab74722a;hb=HEAD#l55': 'HPND-Markus-Kuhn',
  'http://spdx.org/licenses/HPND-sell-varian': 'HPND-sell-variant',
  'https://spdx.org/licenses/HPND-sell-varian': 'HPND-sell-variant',
  'http://git.kernel.org/pub/scm/linux/kernel/git/torvalds/linux.git/tree/net/sunrpc/auth_gss/gss_generic_token.c?h=v4.19': 'HPND-sell-variant',
  'https://git.kernel.org/pub/scm/linux/kernel/git/torvalds/linux.git/tree/net/sunrpc/auth_gss/gss_generic_token.c?h=v4.19': 'HPND-sell-variant',
  'http://spdx.org/licenses/HPND-sell-variant-MIT-disclaimer': 'HPND-sell-variant-MIT-disclaimer',
  'https://spdx.org/licenses/HPND-sell-variant-MIT-disclaimer': 'HPND-sell-variant-MIT-disclaimer',
  'http://github.com/sigmavirus24/x11-ssh-askpass/blob/master/README': 'HPND-sell-variant-MIT-disclaimer',
  'https://github.com/sigmavirus24/x11-ssh-askpass/blob/master/README': 'HPND-sell-variant-MIT-disclaimer',
  'http://spdx.org/licenses/HTMLTIDY': 'HTMLTIDY',
  'https://spdx.org/licenses/HTMLTIDY': 'HTMLTIDY',
  'http://github.com/htacg/tidy-html5/blob/next/README/LICENSE.md': 'HTMLTIDY',
  'https://github.com/htacg/tidy-html5/blob/next/README/LICENSE.md': 'HTMLTIDY',
  'http://spdx.org/licenses/IBM-pibs': 'IBM-pibs',
  'https://spdx.org/licenses/IBM-pibs': 'IBM-pibs',
  'http://git.denx.de/?p=u-boot.git;a=blob;f=arch/powerpc/cpu/ppc4xx/miiphy.c;h=297155fdafa064b955e53e9832de93bfb0cfb85b;hb=9fab4bf4cc077c21e43941866f3f2c196f28670d': 'IBM-pibs',
  'https://git.denx.de/?p=u-boot.git;a=blob;f=arch/powerpc/cpu/ppc4xx/miiphy.c;h=297155fdafa064b955e53e9832de93bfb0cfb85b;hb=9fab4bf4cc077c21e43941866f3f2c196f28670d': 'IBM-pibs',
  'http://spdx.org/licenses/ICU': 'ICU',
  'https://spdx.org/licenses/ICU': 'ICU',
  'http://source.icu-project.org/repos/icu/icu/trunk/license.html': 'ICU',
  'https://source.icu-project.org/repos/icu/icu/trunk/license.html': 'ICU',
  'http://spdx.org/licenses/IEC-Code-Components-EULA': 'IEC-Code-Components-EULA',
  'https://spdx.org/licenses/IEC-Code-Components-EULA': 'IEC-Code-Components-EULA',
  'http://www.iec.ch/webstore/custserv/pdf/CC-EULA.pdf': 'IEC-Code-Components-EULA',
  'https://www.iec.ch/webstore/custserv/pdf/CC-EULA.pdf': 'IEC-Code-Components-EULA',
  'http://www.iec.ch/CCv1': 'IEC-Code-Components-EULA',
  'https://www.iec.ch/CCv1': 'IEC-Code-Components-EULA',
  'http://www.iec.ch/copyright': 'IEC-Code-Components-EULA',
  'https://www.iec.ch/copyright': 'IEC-Code-Components-EULA',
  'http://spdx.org/licenses/IJG': 'IJG',
  'https://spdx.org/licenses/IJG': 'IJG',
  'http://dev.w3.org/cvsweb/Amaya/libjpeg/Attic/README?rev=1.2': 'IJG',
  'https://dev.w3.org/cvsweb/Amaya/libjpeg/Attic/README?rev=1.2': 'IJG',
  'http://spdx.org/licenses/IJG-shor': 'IJG-short',
  'https://spdx.org/licenses/IJG-shor': 'IJG-short',
  'http://sourceforge.net/p/xmedcon/code/ci/master/tree/libs/ljpg/': 'IJG-short',
  'https://sourceforge.net/p/xmedcon/code/ci/master/tree/libs/ljpg/': 'IJG-short',
  'http://spdx.org/licenses/ImageMagick': 'ImageMagick',
  'https://spdx.org/licenses/ImageMagick': 'ImageMagick',
  'http://www.imagemagick.org/script/license.php': 'ImageMagick',
  'https://www.imagemagick.org/script/license.php': 'ImageMagick',
  'http://spdx.org/licenses/iMatix': 'iMatix',
  'https://spdx.org/licenses/iMatix': 'iMatix',
  'http://legacy.imatix.com/html/sfl/sfl4.htm#license': 'iMatix',
  'https://legacy.imatix.com/html/sfl/sfl4.htm#license': 'iMatix',
  'http://spdx.org/licenses/Imlib2': 'Imlib2',
  'https://spdx.org/licenses/Imlib2': 'Imlib2',
  'http://trac.enlightenment.org/e/browser/trunk/imlib2/COPYING': 'Imlib2',
  'https://trac.enlightenment.org/e/browser/trunk/imlib2/COPYING': 'Imlib2',
  'http://git.enlightenment.org/legacy/imlib2.git/tree/COPYING': 'Imlib2',
  'https://git.enlightenment.org/legacy/imlib2.git/tree/COPYING': 'Imlib2',
  'http://spdx.org/licenses/Info-ZIP': 'Info-ZIP',
  'https://spdx.org/licenses/Info-ZIP': 'Info-ZIP',
  'http://www.info-zip.org/license.html': 'Info-ZIP',
  'https://www.info-zip.org/license.html': 'Info-ZIP',
  'http://spdx.org/licenses/Inte': 'Intel',
  'https://spdx.org/licenses/Inte': 'Intel',
  'http://opensource.org/licenses/Intel': 'Intel',
  'https://opensource.org/licenses/Intel': 'Intel',
  'http://spdx.org/licenses/Intel-ACPI': 'Intel-ACPI',
  'https://spdx.org/licenses/Intel-ACPI': 'Intel-ACPI',
  'http://fedoraproject.org/wiki/Licensing/Intel_ACPI_Software_License_Agreement': 'Intel-ACPI',
  'https://fedoraproject.org/wiki/Licensing/Intel_ACPI_Software_License_Agreement': 'Intel-ACPI',
  'http://spdx.org/licenses/Interbase-1.0': 'Interbase-1.0',
  'https://spdx.org/licenses/Interbase-1.0': 'Interbase-1.0',
  'http://web.archive.org/web/20060319014854/http://info.borland.com/devsupport/interbase/opensource/IPL.html': 'Interbase-1.0',
  'https://web.archive.org/web/20060319014854/https://info.borland.com/devsupport/interbase/opensource/IPL.html': 'Interbase-1.0',
  'http://spdx.org/licenses/IPA': 'IPA',
  'https://spdx.org/licenses/IPA': 'IPA',
  'http://opensource.org/licenses/IPA': 'IPA',
  'https://opensource.org/licenses/IPA': 'IPA',
  'http://spdx.org/licenses/IPL-1.0': 'IPL-1.0',
  'https://spdx.org/licenses/IPL-1.0': 'IPL-1.0',
  'http://opensource.org/licenses/IPL-1.0': 'IPL-1.0',
  'https://opensource.org/licenses/IPL-1.0': 'IPL-1.0',
  'http://spdx.org/licenses/ISC': 'ISC',
  'https://spdx.org/licenses/ISC': 'ISC',
  'http://www.isc.org/licenses/': 'ISC',
  'https://www.isc.org/licenses/': 'ISC',
  'http://www.isc.org/downloads/software-support-policy/isc-license/': 'ISC',
  'https://www.isc.org/downloads/software-support-policy/isc-license/': 'ISC',
  'http://opensource.org/licenses/ISC': 'ISC',
  'https://opensource.org/licenses/ISC': 'ISC',
  'http://spdx.org/licenses/Ja': 'Jam',
  'https://spdx.org/licenses/Ja': 'Jam',
  'http://www.boost.org/doc/libs/1_35_0/doc/html/jam.html': 'Jam',
  'https://www.boost.org/doc/libs/1_35_0/doc/html/jam.html': 'Jam',
  'http://web.archive.org/web/20160330173339/http://swarm.workshop.perforce.com/files/guest/perforce_software/jam/src/README': 'Jam',
  'https://web.archive.org/web/20160330173339/https://swarm.workshop.perforce.com/files/guest/perforce_software/jam/src/README': 'Jam',
  'http://spdx.org/licenses/JasPer-2.0': 'JasPer-2.0',
  'https://spdx.org/licenses/JasPer-2.0': 'JasPer-2.0',
  'http://www.ece.uvic.ca/~mdadams/jasper/LICENSE': 'JasPer-2.0',
  'https://www.ece.uvic.ca/~mdadams/jasper/LICENSE': 'JasPer-2.0',
  'http://spdx.org/licenses/JPL-image': 'JPL-image',
  'https://spdx.org/licenses/JPL-image': 'JPL-image',
  'http://www.jpl.nasa.gov/jpl-image-use-policy': 'JPL-image',
  'https://www.jpl.nasa.gov/jpl-image-use-policy': 'JPL-image',
  'http://spdx.org/licenses/JPNIC': 'JPNIC',
  'https://spdx.org/licenses/JPNIC': 'JPNIC',
  'http://gitlab.isc.org/isc-projects/bind9/blob/master/COPYRIGHT#L366': 'JPNIC',
  'https://gitlab.isc.org/isc-projects/bind9/blob/master/COPYRIGHT#L366': 'JPNIC',
  'http://spdx.org/licenses/JSON': 'JSON',
  'https://spdx.org/licenses/JSON': 'JSON',
  'http://www.json.org/license.html': 'JSON',
  'https://www.json.org/license.html': 'JSON',
  'http://spdx.org/licenses/Kazlib': 'Kazlib',
  'https://spdx.org/licenses/Kazlib': 'Kazlib',
  'http://git.savannah.gnu.org/cgit/kazlib.git/tree/except.c?id=0062df360c2d17d57f6af19b0e444c51feb99036': 'Kazlib',
  'https://git.savannah.gnu.org/cgit/kazlib.git/tree/except.c?id=0062df360c2d17d57f6af19b0e444c51feb99036': 'Kazlib',
  'http://spdx.org/licenses/Knuth-CTAN': 'Knuth-CTAN',
  'https://spdx.org/licenses/Knuth-CTAN': 'Knuth-CTAN',
  'http://ctan.org/license/knuth': 'Knuth-CTAN',
  'https://ctan.org/license/knuth': 'Knuth-CTAN',
  'http://spdx.org/licenses/LAL-1.2': 'LAL-1.2',
  'https://spdx.org/licenses/LAL-1.2': 'LAL-1.2',
  'http://artlibre.org/licence/lal/licence-art-libre-12/': 'LAL-1.2',
  'https://artlibre.org/licence/lal/licence-art-libre-12/': 'LAL-1.2',
  'http://spdx.org/licenses/LAL-1.3': 'LAL-1.3',
  'https://spdx.org/licenses/LAL-1.3': 'LAL-1.3',
  'http://artlibre.org/': 'LAL-1.3',
  'https://artlibre.org/': 'LAL-1.3',
  'http://spdx.org/licenses/Latex2e': 'Latex2e',
  'https://spdx.org/licenses/Latex2e': 'Latex2e',
  'http://fedoraproject.org/wiki/Licensing/Latex2e': 'Latex2e',
  'https://fedoraproject.org/wiki/Licensing/Latex2e': 'Latex2e',
  'http://spdx.org/licenses/Leptonica': 'Leptonica',
  'https://spdx.org/licenses/Leptonica': 'Leptonica',
  'http://fedoraproject.org/wiki/Licensing/Leptonica': 'Leptonica',
  'https://fedoraproject.org/wiki/Licensing/Leptonica': 'Leptonica',
  'http://spdx.org/licenses/LGPL-2.0': 'LGPL-2.0',
  'https://spdx.org/licenses/LGPL-2.0': 'LGPL-2.0',
  'http://www.gnu.org/licenses/old-licenses/lgpl-2.0-standalone.html': 'LGPL-2.0-or-later',
  'https://www.gnu.org/licenses/old-licenses/lgpl-2.0-standalone.html': 'LGPL-2.0-or-later',
  'http://spdx.org/licenses/LGPL-2.0+': 'LGPL-2.0+',
  'https://spdx.org/licenses/LGPL-2.0+': 'LGPL-2.0+',
  'http://spdx.org/licenses/LGPL-2.0-only': 'LGPL-2.0-only',
  'https://spdx.org/licenses/LGPL-2.0-only': 'LGPL-2.0-only',
  'http://spdx.org/licenses/LGPL-2.0-or-later': 'LGPL-2.0-or-later',
  'https://spdx.org/licenses/LGPL-2.0-or-later': 'LGPL-2.0-or-later',
  'http://spdx.org/licenses/LGPL-2.1': 'LGPL-2.1',
  'https://spdx.org/licenses/LGPL-2.1': 'LGPL-2.1',
  'http://www.gnu.org/licenses/old-licenses/lgpl-2.1-standalone.html': 'LGPL-2.1-or-later',
  'https://www.gnu.org/licenses/old-licenses/lgpl-2.1-standalone.html': 'LGPL-2.1-or-later',
  'http://opensource.org/licenses/LGPL-2.1': 'LGPL-2.1-or-later',
  'https://opensource.org/licenses/LGPL-2.1': 'LGPL-2.1-or-later',
  'http://spdx.org/licenses/LGPL-2.1+': 'LGPL-2.1+',
  'https://spdx.org/licenses/LGPL-2.1+': 'LGPL-2.1+',
  'http://spdx.org/licenses/LGPL-2.1-only': 'LGPL-2.1-only',
  'https://spdx.org/licenses/LGPL-2.1-only': 'LGPL-2.1-only',
  'http://spdx.org/licenses/LGPL-2.1-or-later': 'LGPL-2.1-or-later',
  'https://spdx.org/licenses/LGPL-2.1-or-later': 'LGPL-2.1-or-later',
  'http://spdx.org/licenses/LGPL-3.0': 'LGPL-3.0',
  'https://spdx.org/licenses/LGPL-3.0': 'LGPL-3.0',
  'http://www.gnu.org/licenses/lgpl-3.0-standalone.html': 'LGPL-3.0-or-later',
  'https://www.gnu.org/licenses/lgpl-3.0-standalone.html': 'LGPL-3.0-or-later',
  'http://www.gnu.org/licenses/lgpl+gpl-3.0.txt': 'LGPL-3.0-or-later',
  'https://www.gnu.org/licenses/lgpl+gpl-3.0.txt': 'LGPL-3.0-or-later',
  'http://opensource.org/licenses/LGPL-3.0': 'LGPL-3.0-or-later',
  'https://opensource.org/licenses/LGPL-3.0': 'LGPL-3.0-or-later',
  'http://spdx.org/licenses/LGPL-3.0+': 'LGPL-3.0+',
  'https://spdx.org/licenses/LGPL-3.0+': 'LGPL-3.0+',
  'http://spdx.org/licenses/LGPL-3.0-only': 'LGPL-3.0-only',
  'https://spdx.org/licenses/LGPL-3.0-only': 'LGPL-3.0-only',
  'http://spdx.org/licenses/LGPL-3.0-or-later': 'LGPL-3.0-or-later',
  'https://spdx.org/licenses/LGPL-3.0-or-later': 'LGPL-3.0-or-later',
  'http://spdx.org/licenses/LGPLLR': 'LGPLLR',
  'https://spdx.org/licenses/LGPLLR': 'LGPLLR',
  'http://www-igm.univ-mlv.fr/~unitex/lgpllr.html': 'LGPLLR',
  'https://www-igm.univ-mlv.fr/~unitex/lgpllr.html': 'LGPLLR',
  'http://spdx.org/licenses/Libpng': 'Libpng',
  'https://spdx.org/licenses/Libpng': 'Libpng',
  'http://www.libpng.org/pub/png/src/libpng-LICENSE.txt': 'libpng-2.0',
  'https://www.libpng.org/pub/png/src/libpng-LICENSE.txt': 'libpng-2.0',
  'http://spdx.org/licenses/libpng-2.0': 'libpng-2.0',
  'https://spdx.org/licenses/libpng-2.0': 'libpng-2.0',
  'http://spdx.org/licenses/libselinux-1.0': 'libselinux-1.0',
  'https://spdx.org/licenses/libselinux-1.0': 'libselinux-1.0',
  'http://github.com/SELinuxProject/selinux/blob/master/libselinux/LICENSE': 'libselinux-1.0',
  'https://github.com/SELinuxProject/selinux/blob/master/libselinux/LICENSE': 'libselinux-1.0',
  'http://spdx.org/licenses/libtiff': 'libtiff',
  'https://spdx.org/licenses/libtiff': 'libtiff',
  'http://fedoraproject.org/wiki/Licensing/libtiff': 'libtiff',
  'https://fedoraproject.org/wiki/Licensing/libtiff': 'libtiff',
  'http://spdx.org/licenses/libutil-David-Nugen': 'libutil-David-Nugent',
  'https://spdx.org/licenses/libutil-David-Nugen': 'libutil-David-Nugent',
  'http://web.mit.edu/freebsd/head/lib/libutil/login_ok.3': 'libutil-David-Nugent',
  'https://web.mit.edu/freebsd/head/lib/libutil/login_ok.3': 'libutil-David-Nugent',
  'http://cgit.freedesktop.org/libbsd/tree/man/setproctitle.3bsd': 'libutil-David-Nugent',
  'https://cgit.freedesktop.org/libbsd/tree/man/setproctitle.3bsd': 'libutil-David-Nugent',
  'http://spdx.org/licenses/LiLiQ-P-1.1': 'LiLiQ-P-1.1',
  'https://spdx.org/licenses/LiLiQ-P-1.1': 'LiLiQ-P-1.1',
  'http://forge.gouv.qc.ca/licence/fr/liliq-v1-1/': 'LiLiQ-P-1.1',
  'https://forge.gouv.qc.ca/licence/fr/liliq-v1-1/': 'LiLiQ-P-1.1',
  'http://opensource.org/licenses/LiLiQ-P-1.1': 'LiLiQ-P-1.1',
  'https://opensource.org/licenses/LiLiQ-P-1.1': 'LiLiQ-P-1.1',
  'http://spdx.org/licenses/LiLiQ-R-1.1': 'LiLiQ-R-1.1',
  'https://spdx.org/licenses/LiLiQ-R-1.1': 'LiLiQ-R-1.1',
  'http://www.forge.gouv.qc.ca/participez/licence-logicielle/licence-libre-du-quebec-liliq-en-francais/licence-libre-du-quebec-reciprocite-liliq-r-v1-1/': 'LiLiQ-R-1.1',
  'https://www.forge.gouv.qc.ca/participez/licence-logicielle/licence-libre-du-quebec-liliq-en-francais/licence-libre-du-quebec-reciprocite-liliq-r-v1-1/': 'LiLiQ-R-1.1',
  'http://opensource.org/licenses/LiLiQ-R-1.1': 'LiLiQ-R-1.1',
  'https://opensource.org/licenses/LiLiQ-R-1.1': 'LiLiQ-R-1.1',
  'http://spdx.org/licenses/LiLiQ-Rplus-1.1': 'LiLiQ-Rplus-1.1',
  'https://spdx.org/licenses/LiLiQ-Rplus-1.1': 'LiLiQ-Rplus-1.1',
  'http://www.forge.gouv.qc.ca/participez/licence-logicielle/licence-libre-du-quebec-liliq-en-francais/licence-libre-du-quebec-reciprocite-forte-liliq-r-v1-1/': 'LiLiQ-Rplus-1.1',
  'https://www.forge.gouv.qc.ca/participez/licence-logicielle/licence-libre-du-quebec-liliq-en-francais/licence-libre-du-quebec-reciprocite-forte-liliq-r-v1-1/': 'LiLiQ-Rplus-1.1',
  'http://opensource.org/licenses/LiLiQ-Rplus-1.1': 'LiLiQ-Rplus-1.1',
  'https://opensource.org/licenses/LiLiQ-Rplus-1.1': 'LiLiQ-Rplus-1.1',
  'http://spdx.org/licenses/Linux-man-pages-copylef': 'Linux-man-pages-copyleft',
  'https://spdx.org/licenses/Linux-man-pages-copylef': 'Linux-man-pages-copyleft',
  'http://www.kernel.org/doc/man-pages/licenses.html': 'Linux-man-pages-copyleft',
  'https://www.kernel.org/doc/man-pages/licenses.html': 'Linux-man-pages-copyleft',
  'http://spdx.org/licenses/Linux-OpenIB': 'Linux-OpenIB',
  'https://spdx.org/licenses/Linux-OpenIB': 'Linux-OpenIB',
  'http://git.kernel.org/pub/scm/linux/kernel/git/torvalds/linux.git/tree/drivers/infiniband/core/sa.h': 'Linux-OpenIB',
  'https://git.kernel.org/pub/scm/linux/kernel/git/torvalds/linux.git/tree/drivers/infiniband/core/sa.h': 'Linux-OpenIB',
  'http://spdx.org/licenses/LOOP': 'LOOP',
  'https://spdx.org/licenses/LOOP': 'LOOP',
  'http://gitlab.com/embeddable-common-lisp/ecl/-/blob/develop/src/lsp/loop.lsp': 'LOOP',
  'https://gitlab.com/embeddable-common-lisp/ecl/-/blob/develop/src/lsp/loop.lsp': 'LOOP',
  'http://git.savannah.gnu.org/cgit/gcl.git/tree/gcl/lsp/gcl_loop.lsp?h=Version_2_6_13pre': 'LOOP',
  'https://git.savannah.gnu.org/cgit/gcl.git/tree/gcl/lsp/gcl_loop.lsp?h=Version_2_6_13pre': 'LOOP',
  'http://sourceforge.net/p/sbcl/sbcl/ci/master/tree/src/code/loop.lisp': 'LOOP',
  'https://sourceforge.net/p/sbcl/sbcl/ci/master/tree/src/code/loop.lisp': 'LOOP',
  'http://github.com/cl-adams/adams/blob/master/LICENSE.md': 'LOOP',
  'https://github.com/cl-adams/adams/blob/master/LICENSE.md': 'LOOP',
  'http://github.com/blakemcbride/eclipse-lisp/blob/master/lisp/loop.lisp': 'LOOP',
  'https://github.com/blakemcbride/eclipse-lisp/blob/master/lisp/loop.lisp': 'LOOP',
  'http://gitlab.common-lisp.net/cmucl/cmucl/-/blob/master/src/code/loop.lisp': 'LOOP',
  'https://gitlab.common-lisp.net/cmucl/cmucl/-/blob/master/src/code/loop.lisp': 'LOOP',
  'http://spdx.org/licenses/LPL-1.0': 'LPL-1.0',
  'https://spdx.org/licenses/LPL-1.0': 'LPL-1.0',
  'http://opensource.org/licenses/LPL-1.0': 'LPL-1.0',
  'https://opensource.org/licenses/LPL-1.0': 'LPL-1.0',
  'http://spdx.org/licenses/LPL-1.02': 'LPL-1.02',
  'https://spdx.org/licenses/LPL-1.02': 'LPL-1.02',
  'http://plan9.bell-labs.com/plan9/license.html': 'LPL-1.02',
  'https://plan9.bell-labs.com/plan9/license.html': 'LPL-1.02',
  'http://opensource.org/licenses/LPL-1.02': 'LPL-1.02',
  'https://opensource.org/licenses/LPL-1.02': 'LPL-1.02',
  'http://spdx.org/licenses/LPPL-1.0': 'LPPL-1.0',
  'https://spdx.org/licenses/LPPL-1.0': 'LPPL-1.0',
  'http://www.latex-project.org/lppl/lppl-1-0.txt': 'LPPL-1.0',
  'https://www.latex-project.org/lppl/lppl-1-0.txt': 'LPPL-1.0',
  'http://spdx.org/licenses/LPPL-1.1': 'LPPL-1.1',
  'https://spdx.org/licenses/LPPL-1.1': 'LPPL-1.1',
  'http://www.latex-project.org/lppl/lppl-1-1.txt': 'LPPL-1.1',
  'https://www.latex-project.org/lppl/lppl-1-1.txt': 'LPPL-1.1',
  'http://spdx.org/licenses/LPPL-1.2': 'LPPL-1.2',
  'https://spdx.org/licenses/LPPL-1.2': 'LPPL-1.2',
  'http://www.latex-project.org/lppl/lppl-1-2.txt': 'LPPL-1.2',
  'https://www.latex-project.org/lppl/lppl-1-2.txt': 'LPPL-1.2',
  'http://spdx.org/licenses/LPPL-1.3a': 'LPPL-1.3a',
  'https://spdx.org/licenses/LPPL-1.3a': 'LPPL-1.3a',
  'http://www.latex-project.org/lppl/lppl-1-3a.txt': 'LPPL-1.3a',
  'https://www.latex-project.org/lppl/lppl-1-3a.txt': 'LPPL-1.3a',
  'http://spdx.org/licenses/LPPL-1.3c': 'LPPL-1.3c',
  'https://spdx.org/licenses/LPPL-1.3c': 'LPPL-1.3c',
  'http://www.latex-project.org/lppl/lppl-1-3c.txt': 'LPPL-1.3c',
  'https://www.latex-project.org/lppl/lppl-1-3c.txt': 'LPPL-1.3c',
  'http://opensource.org/licenses/LPPL-1.3c': 'LPPL-1.3c',
  'https://opensource.org/licenses/LPPL-1.3c': 'LPPL-1.3c',
  'http://spdx.org/licenses/LZMA-SDK-9.11-to-9.20': 'LZMA-SDK-9.11-to-9.20',
  'https://spdx.org/licenses/LZMA-SDK-9.11-to-9.20': 'LZMA-SDK-9.11-to-9.20',
  'http://www.7-zip.org/sdk.html': 'LZMA-SDK-9.22',
  'https://www.7-zip.org/sdk.html': 'LZMA-SDK-9.22',
  'http://sourceforge.net/projects/sevenzip/files/LZMA%20SDK/': 'LZMA-SDK-9.22',
  'https://sourceforge.net/projects/sevenzip/files/LZMA%20SDK/': 'LZMA-SDK-9.22',
  'http://spdx.org/licenses/LZMA-SDK-9.22': 'LZMA-SDK-9.22',
  'https://spdx.org/licenses/LZMA-SDK-9.22': 'LZMA-SDK-9.22',
  'http://spdx.org/licenses/MakeIndex': 'MakeIndex',
  'https://spdx.org/licenses/MakeIndex': 'MakeIndex',
  'http://fedoraproject.org/wiki/Licensing/MakeIndex': 'MakeIndex',
  'https://fedoraproject.org/wiki/Licensing/MakeIndex': 'MakeIndex',
  'http://spdx.org/licenses/Martin-Birgmeier': 'Martin-Birgmeier',
  'https://spdx.org/licenses/Martin-Birgmeier': 'Martin-Birgmeier',
  'http://github.com/Perl/perl5/blob/blead/util.c#L6136': 'Martin-Birgmeier',
  'https://github.com/Perl/perl5/blob/blead/util.c#L6136': 'Martin-Birgmeier',
  'http://spdx.org/licenses/Minpack': 'Minpack',
  'https://spdx.org/licenses/Minpack': 'Minpack',
  'http://www.netlib.org/minpack/disclaimer': 'Minpack',
  'https://www.netlib.org/minpack/disclaimer': 'Minpack',
  'http://gitlab.com/libeigen/eigen/-/blob/master/COPYING.MINPACK': 'Minpack',
  'https://gitlab.com/libeigen/eigen/-/blob/master/COPYING.MINPACK': 'Minpack',
  'http://spdx.org/licenses/MirOS': 'MirOS',
  'https://spdx.org/licenses/MirOS': 'MirOS',
  'http://opensource.org/licenses/MirOS': 'MirOS',
  'https://opensource.org/licenses/MirOS': 'MirOS',
  'http://spdx.org/licenses/MIT': 'MIT',
  'https://spdx.org/licenses/MIT': 'MIT',
  'http://opensource.org/licenses/MIT': 'MIT',
  'https://opensource.org/licenses/MIT': 'MIT',
  'http://spdx.org/licenses/MIT-0': 'MIT-0',
  'https://spdx.org/licenses/MIT-0': 'MIT-0',
  'http://github.com/aws/mit-0': 'MIT-0',
  'https://github.com/aws/mit-0': 'MIT-0',
  'http://romanrm.net/mit-zero': 'MIT-0',
  'https://romanrm.net/mit-zero': 'MIT-0',
  'http://github.com/awsdocs/aws-cloud9-user-guide/blob/master/LICENSE-SAMPLECODE': 'MIT-0',
  'https://github.com/awsdocs/aws-cloud9-user-guide/blob/master/LICENSE-SAMPLECODE': 'MIT-0',
  'http://spdx.org/licenses/MIT-advertising': 'MIT-advertising',
  'https://spdx.org/licenses/MIT-advertising': 'MIT-advertising',
  'http://fedoraproject.org/wiki/Licensing/MIT_With_Advertising': 'MIT-advertising',
  'https://fedoraproject.org/wiki/Licensing/MIT_With_Advertising': 'MIT-advertising',
  'http://spdx.org/licenses/MIT-CMU': 'MIT-CMU',
  'https://spdx.org/licenses/MIT-CMU': 'MIT-CMU',
  'http://fedoraproject.org/wiki/Licensing:MIT?rd=Licensing/MIT#CMU_Style': 'MIT-CMU',
  'https://fedoraproject.org/wiki/Licensing:MIT?rd=Licensing/MIT#CMU_Style': 'MIT-CMU',
  'http://github.com/python-pillow/Pillow/blob/fffb426092c8db24a5f4b6df243a8a3c01fb63cd/LICENSE': 'MIT-CMU',
  'https://github.com/python-pillow/Pillow/blob/fffb426092c8db24a5f4b6df243a8a3c01fb63cd/LICENSE': 'MIT-CMU',
  'http://spdx.org/licenses/MIT-enna': 'MIT-enna',
  'https://spdx.org/licenses/MIT-enna': 'MIT-enna',
  'http://fedoraproject.org/wiki/Licensing/MIT#enna': 'MIT-enna',
  'https://fedoraproject.org/wiki/Licensing/MIT#enna': 'MIT-enna',
  'http://spdx.org/licenses/MIT-fe': 'MIT-feh',
  'https://spdx.org/licenses/MIT-fe': 'MIT-feh',
  'http://fedoraproject.org/wiki/Licensing/MIT#feh': 'MIT-feh',
  'https://fedoraproject.org/wiki/Licensing/MIT#feh': 'MIT-feh',
  'http://spdx.org/licenses/MIT-Modern-Varian': 'MIT-Modern-Variant',
  'https://spdx.org/licenses/MIT-Modern-Varian': 'MIT-Modern-Variant',
  'http://fedoraproject.org/wiki/Licensing:MIT#Modern_Variants': 'MIT-Modern-Variant',
  'https://fedoraproject.org/wiki/Licensing:MIT#Modern_Variants': 'MIT-Modern-Variant',
  'http://ptolemy.berkeley.edu/copyright.htm': 'MIT-Modern-Variant',
  'https://ptolemy.berkeley.edu/copyright.htm': 'MIT-Modern-Variant',
  'http://pirlwww.lpl.arizona.edu/resources/guide/software/PerlTk/Tixlic.html': 'MIT-Modern-Variant',
  'https://pirlwww.lpl.arizona.edu/resources/guide/software/PerlTk/Tixlic.html': 'MIT-Modern-Variant',
  'http://spdx.org/licenses/MIT-open-group': 'MIT-open-group',
  'https://spdx.org/licenses/MIT-open-group': 'MIT-open-group',
  'http://gitlab.freedesktop.org/xorg/app/iceauth/-/blob/master/COPYING': 'MIT-open-group',
  'https://gitlab.freedesktop.org/xorg/app/iceauth/-/blob/master/COPYING': 'MIT-open-group',
  'http://gitlab.freedesktop.org/xorg/app/xvinfo/-/blob/master/COPYING': 'MIT-open-group',
  'https://gitlab.freedesktop.org/xorg/app/xvinfo/-/blob/master/COPYING': 'MIT-open-group',
  'http://gitlab.freedesktop.org/xorg/app/xsetroot/-/blob/master/COPYING': 'MIT-open-group',
  'https://gitlab.freedesktop.org/xorg/app/xsetroot/-/blob/master/COPYING': 'MIT-open-group',
  'http://gitlab.freedesktop.org/xorg/app/xauth/-/blob/master/COPYING': 'MIT-open-group',
  'https://gitlab.freedesktop.org/xorg/app/xauth/-/blob/master/COPYING': 'MIT-open-group',
  'http://spdx.org/licenses/MIT-Wu': 'MIT-Wu',
  'https://spdx.org/licenses/MIT-Wu': 'MIT-Wu',
  'http://github.com/chromium/octane/blob/master/crypto.js': 'MIT-Wu',
  'https://github.com/chromium/octane/blob/master/crypto.js': 'MIT-Wu',
  'http://spdx.org/licenses/MITNFA': 'MITNFA',
  'https://spdx.org/licenses/MITNFA': 'MITNFA',
  'http://fedoraproject.org/wiki/Licensing/MITNFA': 'MITNFA',
  'https://fedoraproject.org/wiki/Licensing/MITNFA': 'MITNFA',
  'http://spdx.org/licenses/Motosoto': 'Motosoto',
  'https://spdx.org/licenses/Motosoto': 'Motosoto',
  'http://opensource.org/licenses/Motosoto': 'Motosoto',
  'https://opensource.org/licenses/Motosoto': 'Motosoto',
  'http://spdx.org/licenses/mpi-permissive': 'mpi-permissive',
  'https://spdx.org/licenses/mpi-permissive': 'mpi-permissive',
  'http://sources.debian.org/src/openmpi/4.1.0-10/ompi/debuggers/msgq_interface.h/?hl=19#L19': 'mpi-permissive',
  'https://sources.debian.org/src/openmpi/4.1.0-10/ompi/debuggers/msgq_interface.h/?hl=19#L19': 'mpi-permissive',
  'http://spdx.org/licenses/mpich2': 'mpich2',
  'https://spdx.org/licenses/mpich2': 'mpich2',
  'http://fedoraproject.org/wiki/Licensing/MIT': 'mpich2',
  'https://fedoraproject.org/wiki/Licensing/MIT': 'mpich2',
  'http://spdx.org/licenses/MPL-1.0': 'MPL-1.0',
  'https://spdx.org/licenses/MPL-1.0': 'MPL-1.0',
  'http://www.mozilla.org/MPL/MPL-1.0.html': 'MPL-1.0',
  'https://www.mozilla.org/MPL/MPL-1.0.html': 'MPL-1.0',
  'http://opensource.org/licenses/MPL-1.0': 'MPL-1.0',
  'https://opensource.org/licenses/MPL-1.0': 'MPL-1.0',
  'http://spdx.org/licenses/MPL-1.1': 'MPL-1.1',
  'https://spdx.org/licenses/MPL-1.1': 'MPL-1.1',
  'http://www.mozilla.org/MPL/MPL-1.1.html': 'MPL-1.1',
  'https://www.mozilla.org/MPL/MPL-1.1.html': 'MPL-1.1',
  'http://opensource.org/licenses/MPL-1.1': 'MPL-1.1',
  'https://opensource.org/licenses/MPL-1.1': 'MPL-1.1',
  'http://spdx.org/licenses/MPL-2.0': 'MPL-2.0',
  'https://spdx.org/licenses/MPL-2.0': 'MPL-2.0',
  'http://www.mozilla.org/MPL/2.0/': 'MPL-2.0-no-copyleft-exception',
  'https://www.mozilla.org/MPL/2.0/': 'MPL-2.0-no-copyleft-exception',
  'http://opensource.org/licenses/MPL-2.0': 'MPL-2.0-no-copyleft-exception',
  'https://opensource.org/licenses/MPL-2.0': 'MPL-2.0-no-copyleft-exception',
  'http://spdx.org/licenses/MPL-2.0-no-copyleft-exception': 'MPL-2.0-no-copyleft-exception',
  'https://spdx.org/licenses/MPL-2.0-no-copyleft-exception': 'MPL-2.0-no-copyleft-exception',
  'http://spdx.org/licenses/mplus': 'mplus',
  'https://spdx.org/licenses/mplus': 'mplus',
  'http://fedoraproject.org/wiki/Licensing:Mplus?rd=Licensing/mplus': 'mplus',
  'https://fedoraproject.org/wiki/Licensing:Mplus?rd=Licensing/mplus': 'mplus',
  'http://spdx.org/licenses/MS-LPL': 'MS-LPL',
  'https://spdx.org/licenses/MS-LPL': 'MS-LPL',
  'http://www.openhub.net/licenses/mslpl': 'MS-LPL',
  'https://www.openhub.net/licenses/mslpl': 'MS-LPL',
  'http://github.com/gabegundy/atlserver/blob/master/License.txt': 'MS-LPL',
  'https://github.com/gabegundy/atlserver/blob/master/License.txt': 'MS-LPL',
  'http://en.wikipedia.org/wiki/Shared_Source_Initiative#Microsoft_Limited_Public_License_(Ms-LPL)': 'MS-LPL',
  'https://en.wikipedia.org/wiki/Shared_Source_Initiative#Microsoft_Limited_Public_License_(Ms-LPL)': 'MS-LPL',
  'http://spdx.org/licenses/MS-PL': 'MS-PL',
  'https://spdx.org/licenses/MS-PL': 'MS-PL',
  'http://www.microsoft.com/opensource/licenses.mspx': 'MS-RL',
  'https://www.microsoft.com/opensource/licenses.mspx': 'MS-RL',
  'http://opensource.org/licenses/MS-PL': 'MS-PL',
  'https://opensource.org/licenses/MS-PL': 'MS-PL',
  'http://spdx.org/licenses/MS-RL': 'MS-RL',
  'https://spdx.org/licenses/MS-RL': 'MS-RL',
  'http://opensource.org/licenses/MS-RL': 'MS-RL',
  'https://opensource.org/licenses/MS-RL': 'MS-RL',
  'http://spdx.org/licenses/MTLL': 'MTLL',
  'https://spdx.org/licenses/MTLL': 'MTLL',
  'http://fedoraproject.org/wiki/Licensing/Matrix_Template_Library_License': 'MTLL',
  'https://fedoraproject.org/wiki/Licensing/Matrix_Template_Library_License': 'MTLL',
  'http://spdx.org/licenses/MulanPSL-1.0': 'MulanPSL-1.0',
  'https://spdx.org/licenses/MulanPSL-1.0': 'MulanPSL-1.0',
  'http://license.coscl.org.cn/MulanPSL/': 'MulanPSL-1.0',
  'https://license.coscl.org.cn/MulanPSL/': 'MulanPSL-1.0',
  'http://github.com/yuwenlong/longphp/blob/25dfb70cc2a466dc4bb55ba30901cbce08d164b5/LICENSE': 'MulanPSL-1.0',
  'https://github.com/yuwenlong/longphp/blob/25dfb70cc2a466dc4bb55ba30901cbce08d164b5/LICENSE': 'MulanPSL-1.0',
  'http://spdx.org/licenses/MulanPSL-2.0': 'MulanPSL-2.0',
  'https://spdx.org/licenses/MulanPSL-2.0': 'MulanPSL-2.0',
  'http://license.coscl.org.cn/MulanPSL2/': 'MulanPSL-2.0',
  'https://license.coscl.org.cn/MulanPSL2/': 'MulanPSL-2.0',
  'http://spdx.org/licenses/Multics': 'Multics',
  'https://spdx.org/licenses/Multics': 'Multics',
  'http://opensource.org/licenses/Multics': 'Multics',
  'https://opensource.org/licenses/Multics': 'Multics',
  'http://spdx.org/licenses/Mup': 'Mup',
  'https://spdx.org/licenses/Mup': 'Mup',
  'http://fedoraproject.org/wiki/Licensing/Mup': 'Mup',
  'https://fedoraproject.org/wiki/Licensing/Mup': 'Mup',
  'http://spdx.org/licenses/NAIST-2003': 'NAIST-2003',
  'https://spdx.org/licenses/NAIST-2003': 'NAIST-2003',
  'http://enterprise.dejacode.com/licenses/public/naist-2003/#license-text': 'NAIST-2003',
  'https://enterprise.dejacode.com/licenses/public/naist-2003/#license-text': 'NAIST-2003',
  'http://github.com/nodejs/node/blob/4a19cc8947b1bba2b2d27816ec3d0edf9b28e503/LICENSE#L343': 'NAIST-2003',
  'https://github.com/nodejs/node/blob/4a19cc8947b1bba2b2d27816ec3d0edf9b28e503/LICENSE#L343': 'NAIST-2003',
  'http://spdx.org/licenses/NASA-1.3': 'NASA-1.3',
  'https://spdx.org/licenses/NASA-1.3': 'NASA-1.3',
  'http://ti.arc.nasa.gov/opensource/nosa/': 'NASA-1.3',
  'https://ti.arc.nasa.gov/opensource/nosa/': 'NASA-1.3',
  'http://opensource.org/licenses/NASA-1.3': 'NASA-1.3',
  'https://opensource.org/licenses/NASA-1.3': 'NASA-1.3',
  'http://spdx.org/licenses/Naumen': 'Naumen',
  'https://spdx.org/licenses/Naumen': 'Naumen',
  'http://opensource.org/licenses/Naumen': 'Naumen',
  'https://opensource.org/licenses/Naumen': 'Naumen',
  'http://spdx.org/licenses/NBPL-1.0': 'NBPL-1.0',
  'https://spdx.org/licenses/NBPL-1.0': 'NBPL-1.0',
  'http://www.openldap.org/devel/gitweb.cgi?p=openldap.git;a=blob;f=LICENSE;hb=37b4b3f6cc4bf34e1d3dec61e69914b9819d8894': 'NBPL-1.0',
  'https://www.openldap.org/devel/gitweb.cgi?p=openldap.git;a=blob;f=LICENSE;hb=37b4b3f6cc4bf34e1d3dec61e69914b9819d8894': 'NBPL-1.0',
  'http://spdx.org/licenses/NCGL-UK-2.0': 'NCGL-UK-2.0',
  'https://spdx.org/licenses/NCGL-UK-2.0': 'NCGL-UK-2.0',
  'http://www.nationalarchives.gov.uk/doc/non-commercial-government-licence/version/2/': 'NCGL-UK-2.0',
  'https://www.nationalarchives.gov.uk/doc/non-commercial-government-licence/version/2/': 'NCGL-UK-2.0',
  'http://spdx.org/licenses/NCSA': 'NCSA',
  'https://spdx.org/licenses/NCSA': 'NCSA',
  'http://otm.illinois.edu/uiuc_openSource': 'NCSA',
  'https://otm.illinois.edu/uiuc_openSource': 'NCSA',
  'http://opensource.org/licenses/NCSA': 'NCSA',
  'https://opensource.org/licenses/NCSA': 'NCSA',
  'http://spdx.org/licenses/Net-SNMP': 'Net-SNMP',
  'https://spdx.org/licenses/Net-SNMP': 'Net-SNMP',
  'http://net-snmp.sourceforge.net/about/license.html': 'Net-SNMP',
  'https://net-snmp.sourceforge.net/about/license.html': 'Net-SNMP',
  'http://spdx.org/licenses/NetCDF': 'NetCDF',
  'https://spdx.org/licenses/NetCDF': 'NetCDF',
  'http://www.unidata.ucar.edu/software/netcdf/copyright.html': 'NetCDF',
  'https://www.unidata.ucar.edu/software/netcdf/copyright.html': 'NetCDF',
  'http://spdx.org/licenses/Newsletr': 'Newsletr',
  'https://spdx.org/licenses/Newsletr': 'Newsletr',
  'http://fedoraproject.org/wiki/Licensing/Newsletr': 'Newsletr',
  'https://fedoraproject.org/wiki/Licensing/Newsletr': 'Newsletr',
  'http://spdx.org/licenses/NGPL': 'NGPL',
  'https://spdx.org/licenses/NGPL': 'NGPL',
  'http://opensource.org/licenses/NGPL': 'NGPL',
  'https://opensource.org/licenses/NGPL': 'NGPL',
  'http://spdx.org/licenses/NICTA-1.0': 'NICTA-1.0',
  'https://spdx.org/licenses/NICTA-1.0': 'NICTA-1.0',
  'http://opensource.apple.com/source/mDNSResponder/mDNSResponder-320.10/mDNSPosix/nss_ReadMe.txt': 'NICTA-1.0',
  'https://opensource.apple.com/source/mDNSResponder/mDNSResponder-320.10/mDNSPosix/nss_ReadMe.txt': 'NICTA-1.0',
  'http://spdx.org/licenses/NIST-PD': 'NIST-PD',
  'https://spdx.org/licenses/NIST-PD': 'NIST-PD',
  'http://github.com/tcheneau/simpleRPL/blob/e645e69e38dd4e3ccfeceb2db8cba05b7c2e0cd3/LICENSE.txt': 'NIST-PD',
  'https://github.com/tcheneau/simpleRPL/blob/e645e69e38dd4e3ccfeceb2db8cba05b7c2e0cd3/LICENSE.txt': 'NIST-PD',
  'http://github.com/tcheneau/Routing/blob/f09f46fcfe636107f22f2c98348188a65a135d98/README.md': 'NIST-PD',
  'https://github.com/tcheneau/Routing/blob/f09f46fcfe636107f22f2c98348188a65a135d98/README.md': 'NIST-PD',
  'http://spdx.org/licenses/NIST-PD-fallback': 'NIST-PD-fallback',
  'https://spdx.org/licenses/NIST-PD-fallback': 'NIST-PD-fallback',
  'http://github.com/usnistgov/jsip/blob/59700e6926cbe96c5cdae897d9a7d2656b42abe3/LICENSE': 'NIST-PD-fallback',
  'https://github.com/usnistgov/jsip/blob/59700e6926cbe96c5cdae897d9a7d2656b42abe3/LICENSE': 'NIST-PD-fallback',
  'http://github.com/usnistgov/fipy/blob/86aaa5c2ba2c6f1be19593c5986071cf6568cc34/LICENSE.rst': 'NIST-PD-fallback',
  'https://github.com/usnistgov/fipy/blob/86aaa5c2ba2c6f1be19593c5986071cf6568cc34/LICENSE.rst': 'NIST-PD-fallback',
  'http://spdx.org/licenses/NLOD-1.0': 'NLOD-1.0',
  'https://spdx.org/licenses/NLOD-1.0': 'NLOD-1.0',
  'http://data.norge.no/nlod/en/1.0': 'NLOD-1.0',
  'https://data.norge.no/nlod/en/1.0': 'NLOD-1.0',
  'http://spdx.org/licenses/NLOD-2.0': 'NLOD-2.0',
  'https://spdx.org/licenses/NLOD-2.0': 'NLOD-2.0',
  'http://data.norge.no/nlod/en/2.0': 'NLOD-2.0',
  'https://data.norge.no/nlod/en/2.0': 'NLOD-2.0',
  'http://spdx.org/licenses/NLPL': 'NLPL',
  'https://spdx.org/licenses/NLPL': 'NLPL',
  'http://fedoraproject.org/wiki/Licensing/NLPL': 'NLPL',
  'https://fedoraproject.org/wiki/Licensing/NLPL': 'NLPL',
  'http://spdx.org/licenses/Nokia': 'Nokia',
  'https://spdx.org/licenses/Nokia': 'Nokia',
  'http://opensource.org/licenses/nokia': 'Nokia',
  'https://opensource.org/licenses/nokia': 'Nokia',
  'http://spdx.org/licenses/NOSL': 'NOSL',
  'https://spdx.org/licenses/NOSL': 'NOSL',
  'http://bits.netizen.com.au/licenses/NOSL/nosl.txt': 'NOSL',
  'https://bits.netizen.com.au/licenses/NOSL/nosl.txt': 'NOSL',
  'http://spdx.org/licenses/Noweb': 'Noweb',
  'https://spdx.org/licenses/Noweb': 'Noweb',
  'http://fedoraproject.org/wiki/Licensing/Noweb': 'Noweb',
  'https://fedoraproject.org/wiki/Licensing/Noweb': 'Noweb',
  'http://spdx.org/licenses/NPL-1.0': 'NPL-1.0',
  'https://spdx.org/licenses/NPL-1.0': 'NPL-1.0',
  'http://www.mozilla.org/MPL/NPL/1.0/': 'NPL-1.0',
  'https://www.mozilla.org/MPL/NPL/1.0/': 'NPL-1.0',
  'http://spdx.org/licenses/NPL-1.1': 'NPL-1.1',
  'https://spdx.org/licenses/NPL-1.1': 'NPL-1.1',
  'http://www.mozilla.org/MPL/NPL/1.1/': 'NPL-1.1',
  'https://www.mozilla.org/MPL/NPL/1.1/': 'NPL-1.1',
  'http://spdx.org/licenses/NPOSL-3.0': 'NPOSL-3.0',
  'https://spdx.org/licenses/NPOSL-3.0': 'NPOSL-3.0',
  'http://opensource.org/licenses/NOSL3.0': 'NPOSL-3.0',
  'https://opensource.org/licenses/NOSL3.0': 'NPOSL-3.0',
  'http://spdx.org/licenses/NRL': 'NRL',
  'https://spdx.org/licenses/NRL': 'NRL',
  'http://web.mit.edu/network/isakmp/nrllicense.html': 'NRL',
  'https://web.mit.edu/network/isakmp/nrllicense.html': 'NRL',
  'http://spdx.org/licenses/NTP': 'NTP',
  'https://spdx.org/licenses/NTP': 'NTP',
  'http://opensource.org/licenses/NTP': 'NTP',
  'https://opensource.org/licenses/NTP': 'NTP',
  'http://spdx.org/licenses/NTP-0': 'NTP-0',
  'https://spdx.org/licenses/NTP-0': 'NTP-0',
  'http://github.com/tytso/e2fsprogs/blob/master/lib/et/et_name.c': 'NTP-0',
  'https://github.com/tytso/e2fsprogs/blob/master/lib/et/et_name.c': 'NTP-0',
  'http://spdx.org/licenses/Nuni': 'Nunit',
  'https://spdx.org/licenses/Nuni': 'Nunit',
  'http://fedoraproject.org/wiki/Licensing/Nunit': 'Nunit',
  'https://fedoraproject.org/wiki/Licensing/Nunit': 'Nunit',
  'http://spdx.org/licenses/O-UDA-1.0': 'O-UDA-1.0',
  'https://spdx.org/licenses/O-UDA-1.0': 'O-UDA-1.0',
  'http://github.com/microsoft/Open-Use-of-Data-Agreement/blob/v1.0/O-UDA-1.0.md': 'O-UDA-1.0',
  'https://github.com/microsoft/Open-Use-of-Data-Agreement/blob/v1.0/O-UDA-1.0.md': 'O-UDA-1.0',
  'http://cdla.dev/open-use-of-data-agreement-v1-0/': 'O-UDA-1.0',
  'https://cdla.dev/open-use-of-data-agreement-v1-0/': 'O-UDA-1.0',
  'http://spdx.org/licenses/OCCT-PL': 'OCCT-PL',
  'https://spdx.org/licenses/OCCT-PL': 'OCCT-PL',
  'http://www.opencascade.com/content/occt-public-license': 'OCCT-PL',
  'https://www.opencascade.com/content/occt-public-license': 'OCCT-PL',
  'http://spdx.org/licenses/OCLC-2.0': 'OCLC-2.0',
  'https://spdx.org/licenses/OCLC-2.0': 'OCLC-2.0',
  'http://www.oclc.org/research/activities/software/license/v2final.htm': 'OCLC-2.0',
  'https://www.oclc.org/research/activities/software/license/v2final.htm': 'OCLC-2.0',
  'http://opensource.org/licenses/OCLC-2.0': 'OCLC-2.0',
  'https://opensource.org/licenses/OCLC-2.0': 'OCLC-2.0',
  'http://spdx.org/licenses/ODbL-1.0': 'ODbL-1.0',
  'https://spdx.org/licenses/ODbL-1.0': 'ODbL-1.0',
  'http://www.opendatacommons.org/licenses/odbl/1.0/': 'ODbL-1.0',
  'https://www.opendatacommons.org/licenses/odbl/1.0/': 'ODbL-1.0',
  'http://opendatacommons.org/licenses/odbl/1-0/': 'ODbL-1.0',
  'https://opendatacommons.org/licenses/odbl/1-0/': 'ODbL-1.0',
  'http://spdx.org/licenses/ODC-By-1.0': 'ODC-By-1.0',
  'https://spdx.org/licenses/ODC-By-1.0': 'ODC-By-1.0',
  'http://opendatacommons.org/licenses/by/1.0/': 'ODC-By-1.0',
  'https://opendatacommons.org/licenses/by/1.0/': 'ODC-By-1.0',
  'http://spdx.org/licenses/OFFIS': 'OFFIS',
  'https://spdx.org/licenses/OFFIS': 'OFFIS',
  'http://sourceforge.net/p/xmedcon/code/ci/master/tree/libs/dicom/README': 'OFFIS',
  'https://sourceforge.net/p/xmedcon/code/ci/master/tree/libs/dicom/README': 'OFFIS',
  'http://spdx.org/licenses/OFL-1.0': 'OFL-1.0',
  'https://spdx.org/licenses/OFL-1.0': 'OFL-1.0',
  'http://scripts.sil.org/cms/scripts/page.php?item_id=OFL10_web': 'OFL-1.0-RFN',
  'https://scripts.sil.org/cms/scripts/page.php?item_id=OFL10_web': 'OFL-1.0-RFN',
  'http://spdx.org/licenses/OFL-1.0-no-RFN': 'OFL-1.0-no-RFN',
  'https://spdx.org/licenses/OFL-1.0-no-RFN': 'OFL-1.0-no-RFN',
  'http://spdx.org/licenses/OFL-1.0-RFN': 'OFL-1.0-RFN',
  'https://spdx.org/licenses/OFL-1.0-RFN': 'OFL-1.0-RFN',
  'http://spdx.org/licenses/OFL-1.1': 'OFL-1.1',
  'https://spdx.org/licenses/OFL-1.1': 'OFL-1.1',
  'http://scripts.sil.org/cms/scripts/page.php?item_id=OFL_web': 'OFL-1.1-RFN',
  'https://scripts.sil.org/cms/scripts/page.php?item_id=OFL_web': 'OFL-1.1-RFN',
  'http://opensource.org/licenses/OFL-1.1': 'OFL-1.1-RFN',
  'https://opensource.org/licenses/OFL-1.1': 'OFL-1.1-RFN',
  'http://spdx.org/licenses/OFL-1.1-no-RFN': 'OFL-1.1-no-RFN',
  'https://spdx.org/licenses/OFL-1.1-no-RFN': 'OFL-1.1-no-RFN',
  'http://spdx.org/licenses/OFL-1.1-RFN': 'OFL-1.1-RFN',
  'https://spdx.org/licenses/OFL-1.1-RFN': 'OFL-1.1-RFN',
  'http://spdx.org/licenses/OGC-1.0': 'OGC-1.0',
  'https://spdx.org/licenses/OGC-1.0': 'OGC-1.0',
  'http://www.ogc.org/ogc/software/1.0': 'OGC-1.0',
  'https://www.ogc.org/ogc/software/1.0': 'OGC-1.0',
  'http://spdx.org/licenses/OGDL-Taiwan-1.0': 'OGDL-Taiwan-1.0',
  'https://spdx.org/licenses/OGDL-Taiwan-1.0': 'OGDL-Taiwan-1.0',
  'http://data.gov.tw/license': 'OGDL-Taiwan-1.0',
  'https://data.gov.tw/license': 'OGDL-Taiwan-1.0',
  'http://spdx.org/licenses/OGL-Canada-2.0': 'OGL-Canada-2.0',
  'https://spdx.org/licenses/OGL-Canada-2.0': 'OGL-Canada-2.0',
  'http://open.canada.ca/en/open-government-licence-canada': 'OGL-Canada-2.0',
  'https://open.canada.ca/en/open-government-licence-canada': 'OGL-Canada-2.0',
  'http://spdx.org/licenses/OGL-UK-1.0': 'OGL-UK-1.0',
  'https://spdx.org/licenses/OGL-UK-1.0': 'OGL-UK-1.0',
  'http://www.nationalarchives.gov.uk/doc/open-government-licence/version/1/': 'OGL-UK-1.0',
  'https://www.nationalarchives.gov.uk/doc/open-government-licence/version/1/': 'OGL-UK-1.0',
  'http://spdx.org/licenses/OGL-UK-2.0': 'OGL-UK-2.0',
  'https://spdx.org/licenses/OGL-UK-2.0': 'OGL-UK-2.0',
  'http://www.nationalarchives.gov.uk/doc/open-government-licence/version/2/': 'OGL-UK-2.0',
  'https://www.nationalarchives.gov.uk/doc/open-government-licence/version/2/': 'OGL-UK-2.0',
  'http://spdx.org/licenses/OGL-UK-3.0': 'OGL-UK-3.0',
  'https://spdx.org/licenses/OGL-UK-3.0': 'OGL-UK-3.0',
  'http://www.nationalarchives.gov.uk/doc/open-government-licence/version/3/': 'OGL-UK-3.0',
  'https://www.nationalarchives.gov.uk/doc/open-government-licence/version/3/': 'OGL-UK-3.0',
  'http://spdx.org/licenses/OGTSL': 'OGTSL',
  'https://spdx.org/licenses/OGTSL': 'OGTSL',
  'http://www.opengroup.org/testing/downloads/The_Open_Group_TSL.txt': 'OGTSL',
  'https://www.opengroup.org/testing/downloads/The_Open_Group_TSL.txt': 'OGTSL',
  'http://opensource.org/licenses/OGTSL': 'OGTSL',
  'https://opensource.org/licenses/OGTSL': 'OGTSL',
  'http://spdx.org/licenses/OLDAP-1.1': 'OLDAP-1.1',
  'https://spdx.org/licenses/OLDAP-1.1': 'OLDAP-1.1',
  'http://www.openldap.org/devel/gitweb.cgi?p=openldap.git;a=blob;f=LICENSE;hb=806557a5ad59804ef3a44d5abfbe91d706b0791f': 'OLDAP-1.1',
  'https://www.openldap.org/devel/gitweb.cgi?p=openldap.git;a=blob;f=LICENSE;hb=806557a5ad59804ef3a44d5abfbe91d706b0791f': 'OLDAP-1.1',
  'http://spdx.org/licenses/OLDAP-1.2': 'OLDAP-1.2',
  'https://spdx.org/licenses/OLDAP-1.2': 'OLDAP-1.2',
  'http://www.openldap.org/devel/gitweb.cgi?p=openldap.git;a=blob;f=LICENSE;hb=42b0383c50c299977b5893ee695cf4e486fb0dc7': 'OLDAP-1.2',
  'https://www.openldap.org/devel/gitweb.cgi?p=openldap.git;a=blob;f=LICENSE;hb=42b0383c50c299977b5893ee695cf4e486fb0dc7': 'OLDAP-1.2',
  'http://spdx.org/licenses/OLDAP-1.3': 'OLDAP-1.3',
  'https://spdx.org/licenses/OLDAP-1.3': 'OLDAP-1.3',
  'http://www.openldap.org/devel/gitweb.cgi?p=openldap.git;a=blob;f=LICENSE;hb=e5f8117f0ce088d0bd7a8e18ddf37eaa40eb09b1': 'OLDAP-1.3',
  'https://www.openldap.org/devel/gitweb.cgi?p=openldap.git;a=blob;f=LICENSE;hb=e5f8117f0ce088d0bd7a8e18ddf37eaa40eb09b1': 'OLDAP-1.3',
  'http://spdx.org/licenses/OLDAP-1.4': 'OLDAP-1.4',
  'https://spdx.org/licenses/OLDAP-1.4': 'OLDAP-1.4',
  'http://www.openldap.org/devel/gitweb.cgi?p=openldap.git;a=blob;f=LICENSE;hb=c9f95c2f3f2ffb5e0ae55fe7388af75547660941': 'OLDAP-1.4',
  'https://www.openldap.org/devel/gitweb.cgi?p=openldap.git;a=blob;f=LICENSE;hb=c9f95c2f3f2ffb5e0ae55fe7388af75547660941': 'OLDAP-1.4',
  'http://spdx.org/licenses/OLDAP-2.0': 'OLDAP-2.0',
  'https://spdx.org/licenses/OLDAP-2.0': 'OLDAP-2.0',
  'http://www.openldap.org/devel/gitweb.cgi?p=openldap.git;a=blob;f=LICENSE;hb=cbf50f4e1185a21abd4c0a54d3f4341fe28f36ea': 'OLDAP-2.0',
  'https://www.openldap.org/devel/gitweb.cgi?p=openldap.git;a=blob;f=LICENSE;hb=cbf50f4e1185a21abd4c0a54d3f4341fe28f36ea': 'OLDAP-2.0',
  'http://spdx.org/licenses/OLDAP-2.0.1': 'OLDAP-2.0.1',
  'https://spdx.org/licenses/OLDAP-2.0.1': 'OLDAP-2.0.1',
  'http://www.openldap.org/devel/gitweb.cgi?p=openldap.git;a=blob;f=LICENSE;hb=b6d68acd14e51ca3aab4428bf26522aa74873f0e': 'OLDAP-2.0.1',
  'https://www.openldap.org/devel/gitweb.cgi?p=openldap.git;a=blob;f=LICENSE;hb=b6d68acd14e51ca3aab4428bf26522aa74873f0e': 'OLDAP-2.0.1',
  'http://spdx.org/licenses/OLDAP-2.1': 'OLDAP-2.1',
  'https://spdx.org/licenses/OLDAP-2.1': 'OLDAP-2.1',
  'http://www.openldap.org/devel/gitweb.cgi?p=openldap.git;a=blob;f=LICENSE;hb=b0d176738e96a0d3b9f85cb51e140a86f21be715': 'OLDAP-2.1',
  'https://www.openldap.org/devel/gitweb.cgi?p=openldap.git;a=blob;f=LICENSE;hb=b0d176738e96a0d3b9f85cb51e140a86f21be715': 'OLDAP-2.1',
  'http://spdx.org/licenses/OLDAP-2.2': 'OLDAP-2.2',
  'https://spdx.org/licenses/OLDAP-2.2': 'OLDAP-2.2',
  'http://www.openldap.org/devel/gitweb.cgi?p=openldap.git;a=blob;f=LICENSE;hb=470b0c18ec67621c85881b2733057fecf4a1acc3': 'OLDAP-2.2',
  'https://www.openldap.org/devel/gitweb.cgi?p=openldap.git;a=blob;f=LICENSE;hb=470b0c18ec67621c85881b2733057fecf4a1acc3': 'OLDAP-2.2',
  'http://spdx.org/licenses/OLDAP-2.2.1': 'OLDAP-2.2.1',
  'https://spdx.org/licenses/OLDAP-2.2.1': 'OLDAP-2.2.1',
  'http://www.openldap.org/devel/gitweb.cgi?p=openldap.git;a=blob;f=LICENSE;hb=4bc786f34b50aa301be6f5600f58a980070f481e': 'OLDAP-2.2.1',
  'https://www.openldap.org/devel/gitweb.cgi?p=openldap.git;a=blob;f=LICENSE;hb=4bc786f34b50aa301be6f5600f58a980070f481e': 'OLDAP-2.2.1',
  'http://spdx.org/licenses/OLDAP-2.2.2': 'OLDAP-2.2.2',
  'https://spdx.org/licenses/OLDAP-2.2.2': 'OLDAP-2.2.2',
  'http://www.openldap.org/devel/gitweb.cgi?p=openldap.git;a=blob;f=LICENSE;hb=df2cc1e21eb7c160695f5b7cffd6296c151ba188': 'OLDAP-2.2.2',
  'https://www.openldap.org/devel/gitweb.cgi?p=openldap.git;a=blob;f=LICENSE;hb=df2cc1e21eb7c160695f5b7cffd6296c151ba188': 'OLDAP-2.2.2',
  'http://spdx.org/licenses/OLDAP-2.3': 'OLDAP-2.3',
  'https://spdx.org/licenses/OLDAP-2.3': 'OLDAP-2.3',
  'http://www.openldap.org/devel/gitweb.cgi?p=openldap.git;a=blob;f=LICENSE;hb=d32cf54a32d581ab475d23c810b0a7fbaf8d63c3': 'OLDAP-2.3',
  'https://www.openldap.org/devel/gitweb.cgi?p=openldap.git;a=blob;f=LICENSE;hb=d32cf54a32d581ab475d23c810b0a7fbaf8d63c3': 'OLDAP-2.3',
  'http://spdx.org/licenses/OLDAP-2.4': 'OLDAP-2.4',
  'https://spdx.org/licenses/OLDAP-2.4': 'OLDAP-2.4',
  'http://www.openldap.org/devel/gitweb.cgi?p=openldap.git;a=blob;f=LICENSE;hb=cd1284c4a91a8a380d904eee68d1583f989ed386': 'OLDAP-2.4',
  'https://www.openldap.org/devel/gitweb.cgi?p=openldap.git;a=blob;f=LICENSE;hb=cd1284c4a91a8a380d904eee68d1583f989ed386': 'OLDAP-2.4',
  'http://spdx.org/licenses/OLDAP-2.5': 'OLDAP-2.5',
  'https://spdx.org/licenses/OLDAP-2.5': 'OLDAP-2.5',
  'http://www.openldap.org/devel/gitweb.cgi?p=openldap.git;a=blob;f=LICENSE;hb=6852b9d90022e8593c98205413380536b1b5a7cf': 'OLDAP-2.5',
  'https://www.openldap.org/devel/gitweb.cgi?p=openldap.git;a=blob;f=LICENSE;hb=6852b9d90022e8593c98205413380536b1b5a7cf': 'OLDAP-2.5',
  'http://spdx.org/licenses/OLDAP-2.6': 'OLDAP-2.6',
  'https://spdx.org/licenses/OLDAP-2.6': 'OLDAP-2.6',
  'http://www.openldap.org/devel/gitweb.cgi?p=openldap.git;a=blob;f=LICENSE;hb=1cae062821881f41b73012ba816434897abf4205': 'OLDAP-2.6',
  'https://www.openldap.org/devel/gitweb.cgi?p=openldap.git;a=blob;f=LICENSE;hb=1cae062821881f41b73012ba816434897abf4205': 'OLDAP-2.6',
  'http://spdx.org/licenses/OLDAP-2.7': 'OLDAP-2.7',
  'https://spdx.org/licenses/OLDAP-2.7': 'OLDAP-2.7',
  'http://www.openldap.org/devel/gitweb.cgi?p=openldap.git;a=blob;f=LICENSE;hb=47c2415c1df81556eeb39be6cad458ef87c534a2': 'OLDAP-2.7',
  'https://www.openldap.org/devel/gitweb.cgi?p=openldap.git;a=blob;f=LICENSE;hb=47c2415c1df81556eeb39be6cad458ef87c534a2': 'OLDAP-2.7',
  'http://spdx.org/licenses/OLDAP-2.8': 'OLDAP-2.8',
  'https://spdx.org/licenses/OLDAP-2.8': 'OLDAP-2.8',
  'http://www.openldap.org/software/release/license.html': 'OLDAP-2.8',
  'https://www.openldap.org/software/release/license.html': 'OLDAP-2.8',
  'http://spdx.org/licenses/OML': 'OML',
  'https://spdx.org/licenses/OML': 'OML',
  'http://fedoraproject.org/wiki/Licensing/Open_Market_License': 'OML',
  'https://fedoraproject.org/wiki/Licensing/Open_Market_License': 'OML',
  'http://spdx.org/licenses/OpenPBS-2.3': 'OpenPBS-2.3',
  'https://spdx.org/licenses/OpenPBS-2.3': 'OpenPBS-2.3',
  'http://github.com/adaptivecomputing/torque/blob/master/PBS_License.txt': 'OpenPBS-2.3',
  'https://github.com/adaptivecomputing/torque/blob/master/PBS_License.txt': 'OpenPBS-2.3',
  'http://www.mcs.anl.gov/research/projects/openpbs/PBS_License.txt': 'OpenPBS-2.3',
  'https://www.mcs.anl.gov/research/projects/openpbs/PBS_License.txt': 'OpenPBS-2.3',
  'http://spdx.org/licenses/OpenSSL': 'OpenSSL',
  'https://spdx.org/licenses/OpenSSL': 'OpenSSL',
  'http://www.openssl.org/source/license.html': 'OpenSSL',
  'https://www.openssl.org/source/license.html': 'OpenSSL',
  'http://spdx.org/licenses/OPL-1.0': 'OPL-1.0',
  'https://spdx.org/licenses/OPL-1.0': 'OPL-1.0',
  'http://old.koalateam.com/jackaroo/OPL_1_0.TXT': 'OPL-1.0',
  'https://old.koalateam.com/jackaroo/OPL_1_0.TXT': 'OPL-1.0',
  'http://fedoraproject.org/wiki/Licensing/Open_Public_License': 'OPL-1.0',
  'https://fedoraproject.org/wiki/Licensing/Open_Public_License': 'OPL-1.0',
  'http://spdx.org/licenses/OPUBL-1.0': 'OPUBL-1.0',
  'https://spdx.org/licenses/OPUBL-1.0': 'OPUBL-1.0',
  'http://opencontent.org/openpub/': 'OPUBL-1.0',
  'https://opencontent.org/openpub/': 'OPUBL-1.0',
  'http://www.debian.org/opl': 'OPUBL-1.0',
  'https://www.debian.org/opl': 'OPUBL-1.0',
  'http://www.ctan.org/license/opl': 'OPUBL-1.0',
  'https://www.ctan.org/license/opl': 'OPUBL-1.0',
  'http://spdx.org/licenses/OSET-PL-2.1': 'OSET-PL-2.1',
  'https://spdx.org/licenses/OSET-PL-2.1': 'OSET-PL-2.1',
  'http://www.osetfoundation.org/public-license': 'OSET-PL-2.1',
  'https://www.osetfoundation.org/public-license': 'OSET-PL-2.1',
  'http://opensource.org/licenses/OPL-2.1': 'OSET-PL-2.1',
  'https://opensource.org/licenses/OPL-2.1': 'OSET-PL-2.1',
  'http://spdx.org/licenses/OSL-1.0': 'OSL-1.0',
  'https://spdx.org/licenses/OSL-1.0': 'OSL-1.0',
  'http://opensource.org/licenses/OSL-1.0': 'OSL-1.0',
  'https://opensource.org/licenses/OSL-1.0': 'OSL-1.0',
  'http://spdx.org/licenses/OSL-1.1': 'OSL-1.1',
  'https://spdx.org/licenses/OSL-1.1': 'OSL-1.1',
  'http://fedoraproject.org/wiki/Licensing/OSL1.1': 'OSL-1.1',
  'https://fedoraproject.org/wiki/Licensing/OSL1.1': 'OSL-1.1',
  'http://spdx.org/licenses/OSL-2.0': 'OSL-2.0',
  'https://spdx.org/licenses/OSL-2.0': 'OSL-2.0',
  'http://web.archive.org/web/20041020171434/http://www.rosenlaw.com/osl2.0.html': 'OSL-2.0',
  'https://web.archive.org/web/20041020171434/https://www.rosenlaw.com/osl2.0.html': 'OSL-2.0',
  'http://spdx.org/licenses/OSL-2.1': 'OSL-2.1',
  'https://spdx.org/licenses/OSL-2.1': 'OSL-2.1',
  'http://web.archive.org/web/20050212003940/http://www.rosenlaw.com/osl21.htm': 'OSL-2.1',
  'https://web.archive.org/web/20050212003940/https://www.rosenlaw.com/osl21.htm': 'OSL-2.1',
  'http://opensource.org/licenses/OSL-2.1': 'OSL-2.1',
  'https://opensource.org/licenses/OSL-2.1': 'OSL-2.1',
  'http://spdx.org/licenses/OSL-3.0': 'OSL-3.0',
  'https://spdx.org/licenses/OSL-3.0': 'OSL-3.0',
  'http://web.archive.org/web/20120101081418/http://rosenlaw.com:80/OSL3.0.htm': 'OSL-3.0',
  'https://web.archive.org/web/20120101081418/https://rosenlaw.com:80/OSL3.0.htm': 'OSL-3.0',
  'http://opensource.org/licenses/OSL-3.0': 'OSL-3.0',
  'https://opensource.org/licenses/OSL-3.0': 'OSL-3.0',
  'http://spdx.org/licenses/Parity-6.0.0': 'Parity-6.0.0',
  'https://spdx.org/licenses/Parity-6.0.0': 'Parity-6.0.0',
  'http://paritylicense.com/versions/6.0.0.html': 'Parity-6.0.0',
  'https://paritylicense.com/versions/6.0.0.html': 'Parity-6.0.0',
  'http://spdx.org/licenses/Parity-7.0.0': 'Parity-7.0.0',
  'https://spdx.org/licenses/Parity-7.0.0': 'Parity-7.0.0',
  'http://paritylicense.com/versions/7.0.0.html': 'Parity-7.0.0',
  'https://paritylicense.com/versions/7.0.0.html': 'Parity-7.0.0',
  'http://spdx.org/licenses/PDDL-1.0': 'PDDL-1.0',
  'https://spdx.org/licenses/PDDL-1.0': 'PDDL-1.0',
  'http://opendatacommons.org/licenses/pddl/1.0/': 'PDDL-1.0',
  'https://opendatacommons.org/licenses/pddl/1.0/': 'PDDL-1.0',
  'http://opendatacommons.org/licenses/pddl/': 'PDDL-1.0',
  'https://opendatacommons.org/licenses/pddl/': 'PDDL-1.0',
  'http://spdx.org/licenses/PHP-3.0': 'PHP-3.0',
  'https://spdx.org/licenses/PHP-3.0': 'PHP-3.0',
  'http://www.php.net/license/3_0.txt': 'PHP-3.0',
  'https://www.php.net/license/3_0.txt': 'PHP-3.0',
  'http://opensource.org/licenses/PHP-3.0': 'PHP-3.0',
  'https://opensource.org/licenses/PHP-3.0': 'PHP-3.0',
  'http://spdx.org/licenses/PHP-3.01': 'PHP-3.01',
  'https://spdx.org/licenses/PHP-3.01': 'PHP-3.01',
  'http://www.php.net/license/3_01.txt': 'PHP-3.01',
  'https://www.php.net/license/3_01.txt': 'PHP-3.01',
  'http://spdx.org/licenses/Plexus': 'Plexus',
  'https://spdx.org/licenses/Plexus': 'Plexus',
  'http://fedoraproject.org/wiki/Licensing/Plexus_Classworlds_License': 'Plexus',
  'https://fedoraproject.org/wiki/Licensing/Plexus_Classworlds_License': 'Plexus',
  'http://spdx.org/licenses/PolyForm-Noncommercial-1.0.0': 'PolyForm-Noncommercial-1.0.0',
  'https://spdx.org/licenses/PolyForm-Noncommercial-1.0.0': 'PolyForm-Noncommercial-1.0.0',
  'http://polyformproject.org/licenses/noncommercial/1.0.0': 'PolyForm-Noncommercial-1.0.0',
  'https://polyformproject.org/licenses/noncommercial/1.0.0': 'PolyForm-Noncommercial-1.0.0',
  'http://spdx.org/licenses/PolyForm-Small-Business-1.0.0': 'PolyForm-Small-Business-1.0.0',
  'https://spdx.org/licenses/PolyForm-Small-Business-1.0.0': 'PolyForm-Small-Business-1.0.0',
  'http://polyformproject.org/licenses/small-business/1.0.0': 'PolyForm-Small-Business-1.0.0',
  'https://polyformproject.org/licenses/small-business/1.0.0': 'PolyForm-Small-Business-1.0.0',
  'http://spdx.org/licenses/PostgreSQL': 'PostgreSQL',
  'https://spdx.org/licenses/PostgreSQL': 'PostgreSQL',
  'http://www.postgresql.org/about/licence': 'PostgreSQL',
  'https://www.postgresql.org/about/licence': 'PostgreSQL',
  'http://opensource.org/licenses/PostgreSQL': 'PostgreSQL',
  'https://opensource.org/licenses/PostgreSQL': 'PostgreSQL',
  'http://spdx.org/licenses/PSF-2.0': 'PSF-2.0',
  'https://spdx.org/licenses/PSF-2.0': 'PSF-2.0',
  'http://opensource.org/licenses/Python-2.0': 'Python-2.0',
  'https://opensource.org/licenses/Python-2.0': 'Python-2.0',
  'http://spdx.org/licenses/psfrag': 'psfrag',
  'https://spdx.org/licenses/psfrag': 'psfrag',
  'http://fedoraproject.org/wiki/Licensing/psfrag': 'psfrag',
  'https://fedoraproject.org/wiki/Licensing/psfrag': 'psfrag',
  'http://spdx.org/licenses/psutils': 'psutils',
  'https://spdx.org/licenses/psutils': 'psutils',
  'http://fedoraproject.org/wiki/Licensing/psutils': 'psutils',
  'https://fedoraproject.org/wiki/Licensing/psutils': 'psutils',
  'http://spdx.org/licenses/Python-2.0': 'Python-2.0',
  'https://spdx.org/licenses/Python-2.0': 'Python-2.0',
  'http://spdx.org/licenses/Python-2.0.1': 'Python-2.0.1',
  'https://spdx.org/licenses/Python-2.0.1': 'Python-2.0.1',
  'http://www.python.org/download/releases/2.0.1/license/': 'Python-2.0.1',
  'https://www.python.org/download/releases/2.0.1/license/': 'Python-2.0.1',
  'http://docs.python.org/3/license.html': 'Python-2.0.1',
  'https://docs.python.org/3/license.html': 'Python-2.0.1',
  'http://github.com/python/cpython/blob/main/LICENSE': 'Python-2.0.1',
  'https://github.com/python/cpython/blob/main/LICENSE': 'Python-2.0.1',
  'http://spdx.org/licenses/Qhu': 'Qhull',
  'https://spdx.org/licenses/Qhu': 'Qhull',
  'http://fedoraproject.org/wiki/Licensing/Qhull': 'Qhull',
  'https://fedoraproject.org/wiki/Licensing/Qhull': 'Qhull',
  'http://spdx.org/licenses/QPL-1.0': 'QPL-1.0',
  'https://spdx.org/licenses/QPL-1.0': 'QPL-1.0',
  'http://doc.qt.nokia.com/3.3/license.html': 'QPL-1.0',
  'https://doc.qt.nokia.com/3.3/license.html': 'QPL-1.0',
  'http://opensource.org/licenses/QPL-1.0': 'QPL-1.0',
  'https://opensource.org/licenses/QPL-1.0': 'QPL-1.0',
  'http://doc.qt.io/archives/3.3/license.html': 'QPL-1.0',
  'https://doc.qt.io/archives/3.3/license.html': 'QPL-1.0',
  'http://spdx.org/licenses/QPL-1.0-INRIA-2004': 'QPL-1.0-INRIA-2004',
  'https://spdx.org/licenses/QPL-1.0-INRIA-2004': 'QPL-1.0-INRIA-2004',
  'http://github.com/maranget/hevea/blob/master/LICENSE': 'QPL-1.0-INRIA-2004',
  'https://github.com/maranget/hevea/blob/master/LICENSE': 'QPL-1.0-INRIA-2004',
  'http://spdx.org/licenses/Rdisc': 'Rdisc',
  'https://spdx.org/licenses/Rdisc': 'Rdisc',
  'http://fedoraproject.org/wiki/Licensing/Rdisc_License': 'Rdisc',
  'https://fedoraproject.org/wiki/Licensing/Rdisc_License': 'Rdisc',
  'http://spdx.org/licenses/RHeCos-1.1': 'RHeCos-1.1',
  'https://spdx.org/licenses/RHeCos-1.1': 'RHeCos-1.1',
  'http://ecos.sourceware.org/old-license.html': 'RHeCos-1.1',
  'https://ecos.sourceware.org/old-license.html': 'RHeCos-1.1',
  'http://spdx.org/licenses/RPL-1.1': 'RPL-1.1',
  'https://spdx.org/licenses/RPL-1.1': 'RPL-1.1',
  'http://opensource.org/licenses/RPL-1.1': 'RPL-1.1',
  'https://opensource.org/licenses/RPL-1.1': 'RPL-1.1',
  'http://spdx.org/licenses/RPL-1.5': 'RPL-1.5',
  'https://spdx.org/licenses/RPL-1.5': 'RPL-1.5',
  'http://opensource.org/licenses/RPL-1.5': 'RPL-1.5',
  'https://opensource.org/licenses/RPL-1.5': 'RPL-1.5',
  'http://spdx.org/licenses/RPSL-1.0': 'RPSL-1.0',
  'https://spdx.org/licenses/RPSL-1.0': 'RPSL-1.0',
  'http://helixcommunity.org/content/rpsl': 'RPSL-1.0',
  'https://helixcommunity.org/content/rpsl': 'RPSL-1.0',
  'http://opensource.org/licenses/RPSL-1.0': 'RPSL-1.0',
  'https://opensource.org/licenses/RPSL-1.0': 'RPSL-1.0',
  'http://spdx.org/licenses/RSA-MD': 'RSA-MD',
  'https://spdx.org/licenses/RSA-MD': 'RSA-MD',
  'http://www.faqs.org/rfcs/rfc1321.html': 'RSA-MD',
  'https://www.faqs.org/rfcs/rfc1321.html': 'RSA-MD',
  'http://spdx.org/licenses/RSCPL': 'RSCPL',
  'https://spdx.org/licenses/RSCPL': 'RSCPL',
  'http://wayback.archive.org/web/20060715140826/http://www.risource.org/RPL/RPL-1.0A.shtml': 'RSCPL',
  'https://wayback.archive.org/web/20060715140826/https://www.risource.org/RPL/RPL-1.0A.shtml': 'RSCPL',
  'http://opensource.org/licenses/RSCPL': 'RSCPL',
  'https://opensource.org/licenses/RSCPL': 'RSCPL',
  'http://spdx.org/licenses/Ruby': 'Ruby',
  'https://spdx.org/licenses/Ruby': 'Ruby',
  'http://www.ruby-lang.org/en/LICENSE.txt': 'Ruby',
  'https://www.ruby-lang.org/en/LICENSE.txt': 'Ruby',
  'http://spdx.org/licenses/SAX-PD': 'SAX-PD',
  'https://spdx.org/licenses/SAX-PD': 'SAX-PD',
  'http://www.saxproject.org/copying.html': 'SAX-PD',
  'https://www.saxproject.org/copying.html': 'SAX-PD',
  'http://spdx.org/licenses/Saxpa': 'Saxpath',
  'https://spdx.org/licenses/Saxpa': 'Saxpath',
  'http://fedoraproject.org/wiki/Licensing/Saxpath_License': 'Saxpath',
  'https://fedoraproject.org/wiki/Licensing/Saxpath_License': 'Saxpath',
  'http://spdx.org/licenses/SCEA': 'SCEA',
  'https://spdx.org/licenses/SCEA': 'SCEA',
  'http://research.scea.com/scea_shared_source_license.html': 'SCEA',
  'https://research.scea.com/scea_shared_source_license.html': 'SCEA',
  'http://spdx.org/licenses/SchemeRepor': 'SchemeReport',
  'https://spdx.org/licenses/SchemeRepor': 'SchemeReport',
  'http://spdx.org/licenses/Sendmai': 'Sendmail',
  'https://spdx.org/licenses/Sendmai': 'Sendmail',
  'http://www.sendmail.com/pdfs/open_source/sendmail_license.pdf': 'Sendmail',
  'https://www.sendmail.com/pdfs/open_source/sendmail_license.pdf': 'Sendmail',
  'http://web.archive.org/web/20160322142305/http://www.sendmail.com/pdfs/open_source/sendmail_license.pdf': 'Sendmail',
  'https://web.archive.org/web/20160322142305/https://www.sendmail.com/pdfs/open_source/sendmail_license.pdf': 'Sendmail',
  'http://spdx.org/licenses/Sendmail-8.23': 'Sendmail-8.23',
  'https://spdx.org/licenses/Sendmail-8.23': 'Sendmail-8.23',
  'http://www.proofpoint.com/sites/default/files/sendmail-license.pdf': 'Sendmail-8.23',
  'https://www.proofpoint.com/sites/default/files/sendmail-license.pdf': 'Sendmail-8.23',
  'http://web.archive.org/web/20181003101040/http://www.proofpoint.com/sites/default/files/sendmail-license.pdf': 'Sendmail-8.23',
  'https://web.archive.org/web/20181003101040/https://www.proofpoint.com/sites/default/files/sendmail-license.pdf': 'Sendmail-8.23',
  'http://spdx.org/licenses/SGI-B-1.0': 'SGI-B-1.0',
  'https://spdx.org/licenses/SGI-B-1.0': 'SGI-B-1.0',
  'http://oss.sgi.com/projects/FreeB/SGIFreeSWLicB.1.0.html': 'SGI-B-1.0',
  'https://oss.sgi.com/projects/FreeB/SGIFreeSWLicB.1.0.html': 'SGI-B-1.0',
  'http://spdx.org/licenses/SGI-B-1.1': 'SGI-B-1.1',
  'https://spdx.org/licenses/SGI-B-1.1': 'SGI-B-1.1',
  'http://oss.sgi.com/projects/FreeB/': 'SGI-B-1.1',
  'https://oss.sgi.com/projects/FreeB/': 'SGI-B-1.1',
  'http://spdx.org/licenses/SGI-B-2.0': 'SGI-B-2.0',
  'https://spdx.org/licenses/SGI-B-2.0': 'SGI-B-2.0',
  'http://oss.sgi.com/projects/FreeB/SGIFreeSWLicB.2.0.pdf': 'SGI-B-2.0',
  'https://oss.sgi.com/projects/FreeB/SGIFreeSWLicB.2.0.pdf': 'SGI-B-2.0',
  'http://spdx.org/licenses/SHL-0.5': 'SHL-0.5',
  'https://spdx.org/licenses/SHL-0.5': 'SHL-0.5',
  'http://solderpad.org/licenses/SHL-0.5/': 'SHL-0.5',
  'https://solderpad.org/licenses/SHL-0.5/': 'SHL-0.5',
  'http://spdx.org/licenses/SHL-0.51': 'SHL-0.51',
  'https://spdx.org/licenses/SHL-0.51': 'SHL-0.51',
  'http://solderpad.org/licenses/SHL-0.51/': 'SHL-0.51',
  'https://solderpad.org/licenses/SHL-0.51/': 'SHL-0.51',
  'http://spdx.org/licenses/SimPL-2.0': 'SimPL-2.0',
  'https://spdx.org/licenses/SimPL-2.0': 'SimPL-2.0',
  'http://opensource.org/licenses/SimPL-2.0': 'SimPL-2.0',
  'https://opensource.org/licenses/SimPL-2.0': 'SimPL-2.0',
  'http://spdx.org/licenses/SISSL': 'SISSL',
  'https://spdx.org/licenses/SISSL': 'SISSL',
  'http://www.openoffice.org/licenses/sissl_license.html': 'SISSL',
  'https://www.openoffice.org/licenses/sissl_license.html': 'SISSL',
  'http://opensource.org/licenses/SISSL': 'SISSL',
  'https://opensource.org/licenses/SISSL': 'SISSL',
  'http://spdx.org/licenses/SISSL-1.2': 'SISSL-1.2',
  'https://spdx.org/licenses/SISSL-1.2': 'SISSL-1.2',
  'http://gridscheduler.sourceforge.net/Gridengine_SISSL_license.html': 'SISSL-1.2',
  'https://gridscheduler.sourceforge.net/Gridengine_SISSL_license.html': 'SISSL-1.2',
  'http://spdx.org/licenses/Sleepyca': 'Sleepycat',
  'https://spdx.org/licenses/Sleepyca': 'Sleepycat',
  'http://opensource.org/licenses/Sleepycat': 'Sleepycat',
  'https://opensource.org/licenses/Sleepycat': 'Sleepycat',
  'http://spdx.org/licenses/SMLNJ': 'SMLNJ',
  'https://spdx.org/licenses/SMLNJ': 'SMLNJ',
  'http://www.smlnj.org/license.html': 'StandardML-NJ',
  'https://www.smlnj.org/license.html': 'StandardML-NJ',
  'http://spdx.org/licenses/SMPPL': 'SMPPL',
  'https://spdx.org/licenses/SMPPL': 'SMPPL',
  'http://github.com/dcblake/SMP/blob/master/Documentation/License.txt': 'SMPPL',
  'https://github.com/dcblake/SMP/blob/master/Documentation/License.txt': 'SMPPL',
  'http://spdx.org/licenses/SNIA': 'SNIA',
  'https://spdx.org/licenses/SNIA': 'SNIA',
  'http://fedoraproject.org/wiki/Licensing/SNIA_Public_License': 'SNIA',
  'https://fedoraproject.org/wiki/Licensing/SNIA_Public_License': 'SNIA',
  'http://spdx.org/licenses/snprintf': 'snprintf',
  'https://spdx.org/licenses/snprintf': 'snprintf',
  'http://github.com/openssh/openssh-portable/blob/master/openbsd-compat/bsd-snprintf.c#L2': 'snprintf',
  'https://github.com/openssh/openssh-portable/blob/master/openbsd-compat/bsd-snprintf.c#L2': 'snprintf',
  'http://spdx.org/licenses/Spencer-86': 'Spencer-86',
  'https://spdx.org/licenses/Spencer-86': 'Spencer-86',
  'http://fedoraproject.org/wiki/Licensing/Henry_Spencer_Reg-Ex_Library_License': 'Spencer-94',
  'https://fedoraproject.org/wiki/Licensing/Henry_Spencer_Reg-Ex_Library_License': 'Spencer-94',
  'http://spdx.org/licenses/Spencer-94': 'Spencer-94',
  'https://spdx.org/licenses/Spencer-94': 'Spencer-94',
  'http://spdx.org/licenses/Spencer-99': 'Spencer-99',
  'https://spdx.org/licenses/Spencer-99': 'Spencer-99',
  'http://www.opensource.apple.com/source/tcl/tcl-5/tcl/generic/regfronts.c': 'Spencer-99',
  'https://www.opensource.apple.com/source/tcl/tcl-5/tcl/generic/regfronts.c': 'Spencer-99',
  'http://spdx.org/licenses/SPL-1.0': 'SPL-1.0',
  'https://spdx.org/licenses/SPL-1.0': 'SPL-1.0',
  'http://opensource.org/licenses/SPL-1.0': 'SPL-1.0',
  'https://opensource.org/licenses/SPL-1.0': 'SPL-1.0',
  'http://spdx.org/licenses/SSH-OpenSSH': 'SSH-OpenSSH',
  'https://spdx.org/licenses/SSH-OpenSSH': 'SSH-OpenSSH',
  'http://github.com/openssh/openssh-portable/blob/1b11ea7c58cd5c59838b5fa574cd456d6047b2d4/LICENCE#L10': 'SSH-OpenSSH',
  'https://github.com/openssh/openssh-portable/blob/1b11ea7c58cd5c59838b5fa574cd456d6047b2d4/LICENCE#L10': 'SSH-OpenSSH',
  'http://spdx.org/licenses/SSH-shor': 'SSH-short',
  'https://spdx.org/licenses/SSH-shor': 'SSH-short',
  'http://github.com/openssh/openssh-portable/blob/1b11ea7c58cd5c59838b5fa574cd456d6047b2d4/pathnames.h': 'SSH-short',
  'https://github.com/openssh/openssh-portable/blob/1b11ea7c58cd5c59838b5fa574cd456d6047b2d4/pathnames.h': 'SSH-short',
  'http://web.mit.edu/kolya/.f/root/athena.mit.edu/sipb.mit.edu/project/openssh/OldFiles/src/openssh-2.9.9p2/ssh-add.1': 'SSH-short',
  'https://web.mit.edu/kolya/.f/root/athena.mit.edu/sipb.mit.edu/project/openssh/OldFiles/src/openssh-2.9.9p2/ssh-add.1': 'SSH-short',
  'http://joinup.ec.europa.eu/svn/lesoll/trunk/italc/lib/src/dsa_key.cpp': 'SSH-short',
  'https://joinup.ec.europa.eu/svn/lesoll/trunk/italc/lib/src/dsa_key.cpp': 'SSH-short',
  'http://spdx.org/licenses/SSPL-1.0': 'SSPL-1.0',
  'https://spdx.org/licenses/SSPL-1.0': 'SSPL-1.0',
  'http://www.mongodb.com/licensing/server-side-public-license': 'SSPL-1.0',
  'https://www.mongodb.com/licensing/server-side-public-license': 'SSPL-1.0',
  'http://spdx.org/licenses/StandardML-NJ': 'StandardML-NJ',
  'https://spdx.org/licenses/StandardML-NJ': 'StandardML-NJ',
  'http://spdx.org/licenses/SugarCRM-1.1.3': 'SugarCRM-1.1.3',
  'https://spdx.org/licenses/SugarCRM-1.1.3': 'SugarCRM-1.1.3',
  'http://www.sugarcrm.com/crm/SPL': 'SugarCRM-1.1.3',
  'https://www.sugarcrm.com/crm/SPL': 'SugarCRM-1.1.3',
  'http://spdx.org/licenses/SunPro': 'SunPro',
  'https://spdx.org/licenses/SunPro': 'SunPro',
  'http://github.com/freebsd/freebsd-src/blob/main/lib/msun/src/e_acosh.c': 'SunPro',
  'https://github.com/freebsd/freebsd-src/blob/main/lib/msun/src/e_acosh.c': 'SunPro',
  'http://github.com/freebsd/freebsd-src/blob/main/lib/msun/src/e_lgammal.c': 'SunPro',
  'https://github.com/freebsd/freebsd-src/blob/main/lib/msun/src/e_lgammal.c': 'SunPro',
  'http://spdx.org/licenses/SWL': 'SWL',
  'https://spdx.org/licenses/SWL': 'SWL',
  'http://fedoraproject.org/wiki/Licensing/SWL': 'SWL',
  'https://fedoraproject.org/wiki/Licensing/SWL': 'SWL',
  'http://spdx.org/licenses/Symlinks': 'Symlinks',
  'https://spdx.org/licenses/Symlinks': 'Symlinks',
  'http://www.mail-archive.com/debian-bugs-rc@lists.debian.org/msg11494.html': 'Symlinks',
  'https://www.mail-archive.com/debian-bugs-rc@lists.debian.org/msg11494.html': 'Symlinks',
  'http://spdx.org/licenses/TAPR-OHL-1.0': 'TAPR-OHL-1.0',
  'https://spdx.org/licenses/TAPR-OHL-1.0': 'TAPR-OHL-1.0',
  'http://www.tapr.org/OHL': 'TAPR-OHL-1.0',
  'https://www.tapr.org/OHL': 'TAPR-OHL-1.0',
  'http://spdx.org/licenses/TCL': 'TCL',
  'https://spdx.org/licenses/TCL': 'TCL',
  'http://www.tcl.tk/software/tcltk/license.html': 'TCL',
  'https://www.tcl.tk/software/tcltk/license.html': 'TCL',
  'http://fedoraproject.org/wiki/Licensing/TCL': 'TCL',
  'https://fedoraproject.org/wiki/Licensing/TCL': 'TCL',
  'http://spdx.org/licenses/TCP-wrappers': 'TCP-wrappers',
  'https://spdx.org/licenses/TCP-wrappers': 'TCP-wrappers',
  'http://rc.quest.com/topics/openssh/license.php#tcpwrappers': 'TCP-wrappers',
  'https://rc.quest.com/topics/openssh/license.php#tcpwrappers': 'TCP-wrappers',
  'http://spdx.org/licenses/TMate': 'TMate',
  'https://spdx.org/licenses/TMate': 'TMate',
  'http://svnkit.com/license.html': 'TMate',
  'https://svnkit.com/license.html': 'TMate',
  'http://spdx.org/licenses/TORQUE-1.1': 'TORQUE-1.1',
  'https://spdx.org/licenses/TORQUE-1.1': 'TORQUE-1.1',
  'http://fedoraproject.org/wiki/Licensing/TORQUEv1.1': 'TORQUE-1.1',
  'https://fedoraproject.org/wiki/Licensing/TORQUEv1.1': 'TORQUE-1.1',
  'http://spdx.org/licenses/TOSL': 'TOSL',
  'https://spdx.org/licenses/TOSL': 'TOSL',
  'http://fedoraproject.org/wiki/Licensing/TOSL': 'TOSL',
  'https://fedoraproject.org/wiki/Licensing/TOSL': 'TOSL',
  'http://spdx.org/licenses/TPDL': 'TPDL',
  'https://spdx.org/licenses/TPDL': 'TPDL',
  'http://metacpan.org/pod/Time::ParseDate#LICENSE': 'TPDL',
  'https://metacpan.org/pod/Time::ParseDate#LICENSE': 'TPDL',
  'http://spdx.org/licenses/TPL-1.0': 'TPL-1.0',
  'https://spdx.org/licenses/TPL-1.0': 'TPL-1.0',
  'http://fedoraproject.org/wiki/Licensing:ThorPublicLicense': 'TPL-1.0',
  'https://fedoraproject.org/wiki/Licensing:ThorPublicLicense': 'TPL-1.0',
  'http://spdx.org/licenses/TTWL': 'TTWL',
  'https://spdx.org/licenses/TTWL': 'TTWL',
  'http://fedoraproject.org/wiki/Licensing/TTWL': 'TTWL',
  'https://fedoraproject.org/wiki/Licensing/TTWL': 'TTWL',
  'http://github.com/ap/Text-Tabs/blob/master/lib.modern/Text/Tabs.pm#L148': 'TTWL',
  'https://github.com/ap/Text-Tabs/blob/master/lib.modern/Text/Tabs.pm#L148': 'TTWL',
  'http://spdx.org/licenses/TU-Berlin-1.0': 'TU-Berlin-1.0',
  'https://spdx.org/licenses/TU-Berlin-1.0': 'TU-Berlin-1.0',
  'http://github.com/swh/ladspa/blob/7bf6f3799fdba70fda297c2d8fd9f526803d9680/gsm/COPYRIGHT': 'TU-Berlin-1.0',
  'https://github.com/swh/ladspa/blob/7bf6f3799fdba70fda297c2d8fd9f526803d9680/gsm/COPYRIGHT': 'TU-Berlin-1.0',
  'http://spdx.org/licenses/TU-Berlin-2.0': 'TU-Berlin-2.0',
  'https://spdx.org/licenses/TU-Berlin-2.0': 'TU-Berlin-2.0',
  'http://github.com/CorsixTH/deps/blob/fd339a9f526d1d9c9f01ccf39e438a015da50035/licences/libgsm.txt': 'TU-Berlin-2.0',
  'https://github.com/CorsixTH/deps/blob/fd339a9f526d1d9c9f01ccf39e438a015da50035/licences/libgsm.txt': 'TU-Berlin-2.0',
  'http://spdx.org/licenses/UCAR': 'UCAR',
  'https://spdx.org/licenses/UCAR': 'UCAR',
  'http://github.com/Unidata/UDUNITS-2/blob/master/COPYRIGHT': 'UCAR',
  'https://github.com/Unidata/UDUNITS-2/blob/master/COPYRIGHT': 'UCAR',
  'http://spdx.org/licenses/UCL-1.0': 'UCL-1.0',
  'https://spdx.org/licenses/UCL-1.0': 'UCL-1.0',
  'http://opensource.org/licenses/UCL-1.0': 'UCL-1.0',
  'https://opensource.org/licenses/UCL-1.0': 'UCL-1.0',
  'http://spdx.org/licenses/Unicode-DFS-2015': 'Unicode-DFS-2015',
  'https://spdx.org/licenses/Unicode-DFS-2015': 'Unicode-DFS-2015',
  'http://web.archive.org/web/20151224134844/http://unicode.org/copyright.html': 'Unicode-DFS-2015',
  'https://web.archive.org/web/20151224134844/https://unicode.org/copyright.html': 'Unicode-DFS-2015',
  'http://spdx.org/licenses/Unicode-DFS-2016': 'Unicode-DFS-2016',
  'https://spdx.org/licenses/Unicode-DFS-2016': 'Unicode-DFS-2016',
  'http://www.unicode.org/copyright.html': 'Unicode-TOU',
  'https://www.unicode.org/copyright.html': 'Unicode-TOU',
  'http://spdx.org/licenses/Unicode-TOU': 'Unicode-TOU',
  'https://spdx.org/licenses/Unicode-TOU': 'Unicode-TOU',
  'http://spdx.org/licenses/Unlicense': 'Unlicense',
  'https://spdx.org/licenses/Unlicense': 'Unlicense',
  'http://unlicense.org/': 'Unlicense',
  'https://unlicense.org/': 'Unlicense',
  'http://spdx.org/licenses/UPL-1.0': 'UPL-1.0',
  'https://spdx.org/licenses/UPL-1.0': 'UPL-1.0',
  'http://opensource.org/licenses/UPL': 'UPL-1.0',
  'https://opensource.org/licenses/UPL': 'UPL-1.0',
  'http://spdx.org/licenses/Vi': 'Vim',
  'https://spdx.org/licenses/Vi': 'Vim',
  'http://vimdoc.sourceforge.net/htmldoc/uganda.html': 'Vim',
  'https://vimdoc.sourceforge.net/htmldoc/uganda.html': 'Vim',
  'http://spdx.org/licenses/VOSTROM': 'VOSTROM',
  'https://spdx.org/licenses/VOSTROM': 'VOSTROM',
  'http://fedoraproject.org/wiki/Licensing/VOSTROM': 'VOSTROM',
  'https://fedoraproject.org/wiki/Licensing/VOSTROM': 'VOSTROM',
  'http://spdx.org/licenses/VSL-1.0': 'VSL-1.0',
  'https://spdx.org/licenses/VSL-1.0': 'VSL-1.0',
  'http://opensource.org/licenses/VSL-1.0': 'VSL-1.0',
  'https://opensource.org/licenses/VSL-1.0': 'VSL-1.0',
  'http://spdx.org/licenses/W3C': 'W3C',
  'https://spdx.org/licenses/W3C': 'W3C',
  'http://www.w3.org/Consortium/Legal/2002/copyright-software-20021231.html': 'W3C',
  'https://www.w3.org/Consortium/Legal/2002/copyright-software-20021231.html': 'W3C',
  'http://opensource.org/licenses/W3C': 'W3C',
  'https://opensource.org/licenses/W3C': 'W3C',
  'http://spdx.org/licenses/W3C-19980720': 'W3C-19980720',
  'https://spdx.org/licenses/W3C-19980720': 'W3C-19980720',
  'http://www.w3.org/Consortium/Legal/copyright-software-19980720.html': 'W3C-19980720',
  'https://www.w3.org/Consortium/Legal/copyright-software-19980720.html': 'W3C-19980720',
  'http://spdx.org/licenses/W3C-20150513': 'W3C-20150513',
  'https://spdx.org/licenses/W3C-20150513': 'W3C-20150513',
  'http://www.w3.org/Consortium/Legal/2015/copyright-software-and-document': 'W3C-20150513',
  'https://www.w3.org/Consortium/Legal/2015/copyright-software-and-document': 'W3C-20150513',
  'http://spdx.org/licenses/w3': 'w3m',
  'https://spdx.org/licenses/w3': 'w3m',
  'http://github.com/tats/w3m/blob/master/COPYING': 'w3m',
  'https://github.com/tats/w3m/blob/master/COPYING': 'w3m',
  'http://spdx.org/licenses/Watcom-1.0': 'Watcom-1.0',
  'https://spdx.org/licenses/Watcom-1.0': 'Watcom-1.0',
  'http://opensource.org/licenses/Watcom-1.0': 'Watcom-1.0',
  'https://opensource.org/licenses/Watcom-1.0': 'Watcom-1.0',
  'http://spdx.org/licenses/Wsuipa': 'Wsuipa',
  'https://spdx.org/licenses/Wsuipa': 'Wsuipa',
  'http://fedoraproject.org/wiki/Licensing/Wsuipa': 'Wsuipa',
  'https://fedoraproject.org/wiki/Licensing/Wsuipa': 'Wsuipa',
  'http://spdx.org/licenses/WTFPL': 'WTFPL',
  'https://spdx.org/licenses/WTFPL': 'WTFPL',
  'http://www.wtfpl.net/about/': 'WTFPL',
  'https://www.wtfpl.net/about/': 'WTFPL',
  'http://sam.zoy.org/wtfpl/COPYING': 'WTFPL',
  'https://sam.zoy.org/wtfpl/COPYING': 'WTFPL',
  'http://spdx.org/licenses/wxWindows': 'wxWindows',
  'https://spdx.org/licenses/wxWindows': 'wxWindows',
  'http://opensource.org/licenses/WXwindows': 'wxWindows',
  'https://opensource.org/licenses/WXwindows': 'wxWindows',
  'http://spdx.org/licenses/X11': 'X11',
  'https://spdx.org/licenses/X11': 'X11',
  'http://www.xfree86.org/3.3.6/COPYRIGHT2.html#3': 'X11',
  'https://www.xfree86.org/3.3.6/COPYRIGHT2.html#3': 'X11',
  'http://spdx.org/licenses/X11-distribute-modifications-varian': 'X11-distribute-modifications-variant',
  'https://spdx.org/licenses/X11-distribute-modifications-varian': 'X11-distribute-modifications-variant',
  'http://github.com/mirror/ncurses/blob/master/COPYING': 'X11-distribute-modifications-variant',
  'https://github.com/mirror/ncurses/blob/master/COPYING': 'X11-distribute-modifications-variant',
  'http://spdx.org/licenses/Xerox': 'Xerox',
  'https://spdx.org/licenses/Xerox': 'Xerox',
  'http://fedoraproject.org/wiki/Licensing/Xerox': 'Xerox',
  'https://fedoraproject.org/wiki/Licensing/Xerox': 'Xerox',
  'http://spdx.org/licenses/XFree86-1.1': 'XFree86-1.1',
  'https://spdx.org/licenses/XFree86-1.1': 'XFree86-1.1',
  'http://www.xfree86.org/current/LICENSE4.html': 'XFree86-1.1',
  'https://www.xfree86.org/current/LICENSE4.html': 'XFree86-1.1',
  'http://spdx.org/licenses/xinetd': 'xinetd',
  'https://spdx.org/licenses/xinetd': 'xinetd',
  'http://fedoraproject.org/wiki/Licensing/Xinetd_License': 'xinetd',
  'https://fedoraproject.org/wiki/Licensing/Xinetd_License': 'xinetd',
  'http://spdx.org/licenses/xlock': 'xlock',
  'https://spdx.org/licenses/xlock': 'xlock',
  'http://fossies.org/linux/tiff/contrib/ras/ras2tif.c': 'xlock',
  'https://fossies.org/linux/tiff/contrib/ras/ras2tif.c': 'xlock',
  'http://spdx.org/licenses/Xne': 'Xnet',
  'https://spdx.org/licenses/Xne': 'Xnet',
  'http://opensource.org/licenses/Xnet': 'Xnet',
  'https://opensource.org/licenses/Xnet': 'Xnet',
  'http://spdx.org/licenses/xpp': 'xpp',
  'https://spdx.org/licenses/xpp': 'xpp',
  'http://fedoraproject.org/wiki/Licensing/xpp': 'xpp',
  'https://fedoraproject.org/wiki/Licensing/xpp': 'xpp',
  'http://spdx.org/licenses/XSka': 'XSkat',
  'https://spdx.org/licenses/XSka': 'XSkat',
  'http://fedoraproject.org/wiki/Licensing/XSkat_License': 'XSkat',
  'https://fedoraproject.org/wiki/Licensing/XSkat_License': 'XSkat',
  'http://spdx.org/licenses/YPL-1.0': 'YPL-1.0',
  'https://spdx.org/licenses/YPL-1.0': 'YPL-1.0',
  'http://www.zimbra.com/license/yahoo_public_license_1.0.html': 'YPL-1.0',
  'https://www.zimbra.com/license/yahoo_public_license_1.0.html': 'YPL-1.0',
  'http://spdx.org/licenses/YPL-1.1': 'YPL-1.1',
  'https://spdx.org/licenses/YPL-1.1': 'YPL-1.1',
  'http://www.zimbra.com/license/yahoo_public_license_1.1.html': 'YPL-1.1',
  'https://www.zimbra.com/license/yahoo_public_license_1.1.html': 'YPL-1.1',
  'http://spdx.org/licenses/Zed': 'Zed',
  'https://spdx.org/licenses/Zed': 'Zed',
  'http://fedoraproject.org/wiki/Licensing/Zed': 'Zed',
  'https://fedoraproject.org/wiki/Licensing/Zed': 'Zed',
  'http://spdx.org/licenses/Zend-2.0': 'Zend-2.0',
  'https://spdx.org/licenses/Zend-2.0': 'Zend-2.0',
  'http://web.archive.org/web/20130517195954/http://www.zend.com/license/2_00.txt': 'Zend-2.0',
  'https://web.archive.org/web/20130517195954/https://www.zend.com/license/2_00.txt': 'Zend-2.0',
  'http://spdx.org/licenses/Zimbra-1.3': 'Zimbra-1.3',
  'https://spdx.org/licenses/Zimbra-1.3': 'Zimbra-1.3',
  'http://web.archive.org/web/20100302225219/http://www.zimbra.com/license/zimbra-public-license-1-3.html': 'Zimbra-1.3',
  'https://web.archive.org/web/20100302225219/https://www.zimbra.com/license/zimbra-public-license-1-3.html': 'Zimbra-1.3',
  'http://spdx.org/licenses/Zimbra-1.4': 'Zimbra-1.4',
  'https://spdx.org/licenses/Zimbra-1.4': 'Zimbra-1.4',
  'http://www.zimbra.com/legal/zimbra-public-license-1-4': 'Zimbra-1.4',
  'https://www.zimbra.com/legal/zimbra-public-license-1-4': 'Zimbra-1.4',
  'http://spdx.org/licenses/Zlib': 'Zlib',
  'https://spdx.org/licenses/Zlib': 'Zlib',
  'http://www.zlib.net/zlib_license.html': 'Zlib',
  'https://www.zlib.net/zlib_license.html': 'Zlib',
  'http://opensource.org/licenses/Zlib': 'Zlib',
  'https://opensource.org/licenses/Zlib': 'Zlib',
  'http://spdx.org/licenses/zlib-acknowledgemen': 'zlib-acknowledgement',
  'https://spdx.org/licenses/zlib-acknowledgemen': 'zlib-acknowledgement',
  'http://fedoraproject.org/wiki/Licensing/ZlibWithAcknowledgement': 'zlib-acknowledgement',
  'https://fedoraproject.org/wiki/Licensing/ZlibWithAcknowledgement': 'zlib-acknowledgement',
  'http://spdx.org/licenses/ZPL-1.1': 'ZPL-1.1',
  'https://spdx.org/licenses/ZPL-1.1': 'ZPL-1.1',
  'http://old.zope.org/Resources/License/ZPL-1.1': 'ZPL-1.1',
  'https://old.zope.org/Resources/License/ZPL-1.1': 'ZPL-1.1',
  'http://spdx.org/licenses/ZPL-2.0': 'ZPL-2.0',
  'https://spdx.org/licenses/ZPL-2.0': 'ZPL-2.0',
  'http://old.zope.org/Resources/License/ZPL-2.0': 'ZPL-2.0',
  'https://old.zope.org/Resources/License/ZPL-2.0': 'ZPL-2.0',
  'http://opensource.org/licenses/ZPL-2.0': 'ZPL-2.0',
  'https://opensource.org/licenses/ZPL-2.0': 'ZPL-2.0',
  'http://spdx.org/licenses/ZPL-2.1': 'ZPL-2.1',
  'https://spdx.org/licenses/ZPL-2.1': 'ZPL-2.1',
  'http://old.zope.org/Resources/ZPL/': 'ZPL-2.1',
  'https://old.zope.org/Resources/ZPL/': 'ZPL-2.1',
})
